/**
 * 
 */
package com.wincornixdorf.vnc.global;

import java.nio.charset.Charset;
import java.util.Observable;
import java.util.Set;
import java.util.SortedMap;

import javax.swing.ComboBoxModel;
import javax.swing.event.ListDataListener;

import com.wincornixdorf.vnc.rfb.RFB;

/**
 * @author dic
 *
 */
public class CharsetModel extends Observable
    implements ComboBoxModel<Charset> {
    private String[] charsets = new String[ 0];
    
    private Charset  selected = RFB.DEFAULT_CHARSET;
    
    public CharsetModel() {
        SortedMap<String,Charset> smAvail = Charset.availableCharsets();
        Set<String> csAliases = smAvail.keySet();
        charsets = csAliases.toArray( new String[0]);
    }

    /* (non-Javadoc)
     * @see javax.swing.ComboBoxModel#getSelectedItem()
     */
    @Override
    public Object getSelectedItem() { return selected; }

    /* (non-Javadoc)
     * Allows setting the charset to be used when decoding RFB.ServerCutText
     * messages in the RFB protocol. By default US-ASCII will be assumed, as
     * is stipulated by the protocol. Some servers however seem to be using
     * their platforms native encoding.
     * @see javax.swing.ComboBoxModel#setSelectedItem(java.lang.Object)
     */
    @Override
    public void setSelectedItem( Object anItem) {
        if ( anItem == null || !(anItem instanceof Charset)) return;
        Charset cs = (Charset)anItem;

        if ( cs == selected) return;

        selected = cs;
        notifyObservers( selected);
    }

    public Charset getSelectedCharset() {
        return selected;
    }

    /* (non-Javadoc)
     * @see javax.swing.ListModel#addListDataListener(javax.swing.event.ListDataListener)
     */
    @Override
    public void addListDataListener( ListDataListener l) { }

    /* (non-Javadoc)
     * @see javax.swing.ListModel#removeListDataListener(javax.swing.event.ListDataListener)
     */
    @Override
    public void removeListDataListener( ListDataListener l) { }

    /* (non-Javadoc)
     * @see javax.swing.ListModel#getElementAt(int)
     */
    @Override
    public Charset getElementAt( int index) {
        if ( index >= charsets.length) return RFB.DEFAULT_CHARSET;
        return Charset.availableCharsets().get( charsets[ index]);
    }


    /* (non-Javadoc)
     * @see javax.swing.ListModel#getSize()
     */
    @Override
    public int getSize() { return charsets.length; }

}
