package com.wincornixdorf.vnc.global;

import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JComboBox;
import javax.swing.KeyStroke;

import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;

public class VNCAction extends AbstractAction {
    private static final long serialVersionUID = 8558772865154189466L;

    private static final String SMALL = "/16x16/";
    private static final String LARGE = "/64x64/";

    public static String GLASSPANE_ICON = "MY_PRIVATE_GLASSPANE_ICON";

    protected Goal goal = Goal.QUIT;
    protected VNCViewer v = null;

    public enum Goal {
        QUIT, ABOUT, GLASS, QUICK_KEYS, LIST_SESSIONS,
        SHOW_FB, SHOW_FT, SHOW_LG,
        SHOW_CONNECTION_DOCTOR,
        CREATE_SESSION, DELETE_SESSION, CLOSE_FB_CONNECTION, RESET_FB_CONNECTION, CLOSE_FT_CONNECTION, RESET_FT_CONNECTION,
        OPEN_BOTH_CONNECTIONS, CLOSE_BOTH_CONNECTIONS, RESET_BOTH_CONNECTIONS,
        CLIPBOARD, REFRESH, RSPOWERUP, RSPOWERDOWN, OPTIONS, RECORD,
        TOGGLE_FULLSCREEN, TOGGLE_CONN_STATUS_BAR,
        CTRLALTDEL, META, ALTTAB, BATTLA, ALTF4,
        TOGGLE_GP_UI_ENABLED,
        DEVELOP_1,
        CONTROL_NAVIGATOR,
        ALTGR_EUR, ALTGR_AT,
        DISABLE_CONTROLS,
        CHANGE_KEYBOARD_LAYOUT
    }

    public VNCAction(VNCViewer v, Goal g) {
        super();
        this.v = v;
        this.goal = g;

        putValue(Action.NAME, getDefaultName(goal));
        putValue(Action.SMALL_ICON, getDefaultIcon(goal, true));
        putValue(VNCAction.GLASSPANE_ICON, getDefaultIcon(goal, false));
        putValue(Action.SHORT_DESCRIPTION, getDefaultDescription(goal));
        putValue(Action.ACCELERATOR_KEY, getDefaultKeyStroke(goal));
        putValue(Action.SELECTED_KEY, getDefaultSelectedKey(goal));

        setEnabled(false);
    }

    public Goal getGoal() {
        return goal;
    }

    private Boolean getDefaultSelectedKey(Goal g) {
        switch (g) {
            case CLIPBOARD:
                return (Boolean.FALSE);
            case TOGGLE_FULLSCREEN:
                return (Boolean.FALSE);
            case OPTIONS:
                return (Boolean.FALSE);
            case TOGGLE_GP_UI_ENABLED:
                return (VNCViewer.isDivine() ? Boolean.TRUE : Boolean.FALSE);
            case CONTROL_NAVIGATOR:
                return (Boolean.FALSE);
            default:
                return (Boolean.TRUE);
        }
    }

    // TODO: This is hilarious! This code belongs into the Viewer. This class should have no knowledge of the Viewer!
    @Override
    public void actionPerformed(ActionEvent e) {


        switch (goal) {
            case QUIT:
                v.quit();
                break;
            case ABOUT:
                v.toggleAboutVisiblity();
                break;
            case GLASS:
                v.toggleGlassPane();
                break;
            case QUICK_KEYS:
                v.toggleQuickKeys();
                break;
            case TOGGLE_FULLSCREEN:
                v.toggleFullScreen();
                break;
            case LIST_SESSIONS:
                v.toggleSessionPicker();
                break;

            case SHOW_FB:
                v.showCurrentSessionsFB();
                break;
            case SHOW_FT:
                v.showCurrentSessionsFT();
                break;
            case SHOW_LG:
                v.showCurrentSessionsLG();
                break;

            case SHOW_CONNECTION_DOCTOR:
                v.showConnectionDoctor();
                break;

            case CREATE_SESSION:
                v.createNewSession();
                break;
            case DELETE_SESSION:
                v.disposeOfCurrentSession();
                break;
            case CLOSE_FB_CONNECTION:
                v.closeConnection(true);
                break;
            case RESET_FB_CONNECTION:
                v.resetConnection(true);
                break;
            case CLOSE_FT_CONNECTION:
                v.closeConnection(false);
                break;
            case RESET_FT_CONNECTION:
                v.resetConnection(false);
                break;

            case OPEN_BOTH_CONNECTIONS:
                v.openConnections();
                break;
            case CLOSE_BOTH_CONNECTIONS:
                v.closeConnections();
                break;
            case RESET_BOTH_CONNECTIONS:
                v.resetConnections();
                break;

            case CLIPBOARD:
                v.toggleClipboardVisiblity();
                break;
            case CTRLALTDEL:
                v.requestCtrlAltDeleteKeypressBeSent();
                break;
            case ALTGR_EUR:
                v.requestAltGrEurKeypressBeSent();
                break;
            case ALTGR_AT:
                v.requestAltGrAtKeypressBeSent();
                break;
            case REFRESH:
                v.requestScreenRefresh();
                break;
            case ALTTAB:
                v.requestAltTabKeypressBeSent(false);
                break;
            case BATTLA:
                v.requestAltTabKeypressBeSent(true);
                break;
            case ALTF4:
                v.requestAltF4KeypressBeSent();
                break;
            case META:
                v.requestMetaKeypressBeSent();
                break;
            case RSPOWERDOWN:
                v.requestRemoteScreenBeTurned(true);
                break;
            case RSPOWERUP:
                v.requestRemoteScreenBeTurned(false);
                break;
            case OPTIONS:
                v.toggleOptionsVisiblity();
                break;
            case RECORD:
                v.rec.setVisible(!v.rec.isVisible());
                break;
            case TOGGLE_CONN_STATUS_BAR:
                v.toggleConnectionStatusBar();
                break;
            case TOGGLE_GP_UI_ENABLED:
                v.toggleUIMode();
                break;
            case DEVELOP_1:
                v.developAction1();
                break;
            case CONTROL_NAVIGATOR:
                v.showControlNavigator();
                break;
            case DISABLE_CONTROLS:
                v.disableRemoteControl();
                break;
            case CHANGE_KEYBOARD_LAYOUT:
                JComboBox cb = (JComboBox) e.getSource();
                String keyboardLayout = (String) cb.getSelectedItem();
                v.getActiveSession().getConfiguration().setUsedKeyboardLayout(VNCSessionCfg.KeyboardLayout.valueOf(keyboardLayout));
                break;
        }
    }

    private ImageIcon getDefaultIcon(Goal g, boolean small) {
        String folder = small ? SMALL : LARGE;
        String fn = "bug.png";

        switch (g) {
            case QUIT:
                fn = "exit.png";
                break;
            case ABOUT:
                fn = "documentinfo.png";
                break;
            case GLASS:
                fn = "kllckety.png";
                break;
            case QUICK_KEYS:
                fn = "kllckety.png";
                break;
            case LIST_SESSIONS:
                fn = "background.png";
                break;

            case SHOW_FB:
                fn = "desktop22.png";
                break;
            case SHOW_FT:
                fn = "file-manager.png";
                break;
            case SHOW_LG:
                fn = "kedit.png";
                break;

            case SHOW_CONNECTION_DOCTOR:
                fn = "bug.png";
                break;

            case CREATE_SESSION:
                fn = "background_plus.png";
                break;
            case DELETE_SESSION:
                fn = "background_minus.png";
                break;

            case CLOSE_FB_CONNECTION:
                fn = "desktop_stop.png";
                break;
            case RESET_FB_CONNECTION:
                fn = "desktop_restart.png";
                break;
            case CLOSE_FT_CONNECTION:
                fn = "file-manager_stop.png";
                break;
            case RESET_FT_CONNECTION:
                fn = "file-manager_restart.png";
                break;

            case OPEN_BOTH_CONNECTIONS:
                folder = SMALL;
                fn = "connect.png";
                break;
            case CLOSE_BOTH_CONNECTIONS:
                folder = SMALL;
                fn = "disconnect.png";
                break;
            case RESET_BOTH_CONNECTIONS:
                folder = SMALL;
                fn = "reconnect.png";
                break;

            case CLIPBOARD:
                fn = "editpaste.png";
                break;
            case CTRLALTDEL:
            case ALTGR_EUR:
            case ALTGR_AT:
            case DISABLE_CONTROLS:
                fn = "advanced.png";
                break;
            case REFRESH:
                fn = "reload.png";
                break;
            case ALTTAB:
                folder = LARGE;
                fn = "cycle-right.png";
                break;
            case BATTLA:
                folder = LARGE;
                fn = "cycle-left.png";
                break;
            case ALTF4:
                folder = LARGE;
                fn = "application-exit.png";
                break;
            case META:
                fn = "go_kde.png";
                break;
            case RSPOWERDOWN:
                fn = "displayconfig.png";
                break;
            case RSPOWERUP:
                fn = "display.png";
                break;
            case OPTIONS:
                fn = "configure.png";
                break;
            case RECORD:
                fn = "record.png";
                break;

            case TOGGLE_FULLSCREEN:
                folder = SMALL;
                fn = "view-fullscreen.png";
                break;
            case TOGGLE_CONN_STATUS_BAR:
                folder = LARGE;
                fn = "kmag.png";
                break;
            case TOGGLE_GP_UI_ENABLED:
                folder = SMALL;
                fn = "bookmark-toolbar.png";
                break;
            case DEVELOP_1:
                folder = LARGE;
                fn = "bug.png";
                break;
            case CONTROL_NAVIGATOR:
                folder = SMALL;
                fn = "bookmark-toolbar.png";
                break;

        }

        String dbg = folder + fn;
        return new ImageIcon(this.getClass().getResource(dbg));
    }

    private static String getDefaultName(Goal g) {
        switch (g) {
            case QUIT:
                return "Quit";
            case ABOUT:
                return "About";
            case GLASS:
                return "Action toolbox";
            case QUICK_KEYS:
                return "QuickKeys";
            case LIST_SESSIONS:
                return "Sessions";

            case SHOW_FB:
                return "Desktop";
            case SHOW_FT:
                return "Files";
            case SHOW_LG:
                return "Log";

            case SHOW_CONNECTION_DOCTOR:
                return "Connection Doctor";

            case CREATE_SESSION:
                return "New Session";
            case DELETE_SESSION:
                return "Discard Session";
            case CLOSE_FB_CONNECTION:
                return "Disconnect";
            case RESET_FB_CONNECTION:
                return "Reset remote desktop connection";
            case CLOSE_FT_CONNECTION:
                return "Disconnect";
            case RESET_FT_CONNECTION:
                return "Reset remote filesystem connection";

            case CLIPBOARD:
                return "Show Clipboard";
            case CTRLALTDEL:
                return "Send Ctrl Alt Delete";
            case REFRESH:
                return "Refresh";
            case ALTTAB:
                return "Send Alt + Tab";
            case BATTLA:
                return "Send Shift + Alt + Tab";
            case ALTF4:
                return "Send Alt + F4";
            case META:
                return "Send Windows Key";
            case RSPOWERDOWN:
                return "Power down console";
            case RSPOWERUP:
                return "Power up console";
            case OPTIONS:
                return "Options";
            case RECORD:
                return "Record";

            case TOGGLE_FULLSCREEN:
                return "Fullscreen";
            case TOGGLE_CONN_STATUS_BAR:
                return "Status Bar";
            case TOGGLE_GP_UI_ENABLED:
                return "Toggle control's visibility";
            case DEVELOP_1:
                return "Develop1";
            case CONTROL_NAVIGATOR:
                return "Toggle control's navigator";
            case ALTGR_EUR:
                return "Send € symbol";
            case ALTGR_AT:
                return "Send @ symbol";
            case DISABLE_CONTROLS:
                return "Disable/Enable Controls";
            case CHANGE_KEYBOARD_LAYOUT:
                return "Keyboard Layout";
            default:
                return "Error";
        }
    }

    private static String getDefaultDescription(Goal g) {
        switch (g) {
            case QUIT:
                return "Terminate the program";
            case ABOUT:
                return "Vendor information";
            case GLASS:
                return "Speed access popular actions";
            case QUICK_KEYS:
                return "Send a privileged key stroke to the server";
            case LIST_SESSIONS:
                return "Show/hide the list of sessions";

            case SHOW_FB:
                return "Show the remote desktop";
            case SHOW_FT:
                return "Show the remote filesystem";
            case SHOW_LG:
                return "Show the sessions log";

            case SHOW_CONNECTION_DOCTOR:
                return "Helps find the reason for connection failures.";

            case CREATE_SESSION:
                return "Start a new Session with a VNC Server";
            case DELETE_SESSION:
                return "Close the current Session";
            case CLOSE_FB_CONNECTION:
                return "Closes the connection that offers access to a remote computers desktop.";
            case RESET_FB_CONNECTION:
                return "Resets the connection that offers access to a remote computers desktop.";
            case CLOSE_FT_CONNECTION:
                return "Closes the connection for exchanging files with a remote computer.";
            case RESET_FT_CONNECTION:
                return "Resets the connection for exchanging files with a remote computer.";

            case OPEN_BOTH_CONNECTIONS:
                return "Connect to both desktop and filesystem";
            case CLOSE_BOTH_CONNECTIONS:
                return "Disconnect from both desktop and filesystem";
            case RESET_BOTH_CONNECTIONS:
                return "Reset both desktop and filesystem connection";

            case CLIPBOARD:
                return "Show the contents of the clipboard";
            case CTRLALTDEL:
                return "Send the Ctrl-Alt-Del key combination to the server";
            case REFRESH:
                return "Redraw the entire destop";
            case ALTTAB:
                return "Send the key sequence Alt + Tab to the server";
            case BATTLA:
                return "Send the key sequence Shift + Alt + Tab to the server";
            case ALTF4:
                return "Send the key sequence Alt + F4 to the server";
            case META:
                return "Send a \"Windows\" key event to the server";
            case RSPOWERDOWN:
                return "Power down the server's monitor";
            case RSPOWERUP:
                return "Power up the server's monitor";
            case OPTIONS:
                return "Options for fine-tuning the VNC connection";
            case RECORD:
                return "Record a video";

            case TOGGLE_FULLSCREEN:
                return "Toggle fullscreen";
            case TOGGLE_CONN_STATUS_BAR:
                return "Toggle connection status bar";
            case TOGGLE_GP_UI_ENABLED:
                return "Hides / Shows the Action buttons and Drop Down menues.";
            case DEVELOP_1:
                return "Develop1 does what it wants!!!";
            case CONTROL_NAVIGATOR:
                return "Hides / Shows the control navigator";
            case ALTGR_EUR:
                return "Send € symbol to the remote desktop.";
            case ALTGR_AT:
                return "Send @ symbol to the remote desktop.";
            case DISABLE_CONTROLS:
                return "Disable / Enable remote remote controlling.";
            default:
                return "Error";
        }
    }

    public static KeyStroke getDefaultKeyStroke(Goal g) {
        switch (g) {
            case QUIT:
                return KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_MASK);
            case ABOUT:
                return KeyStroke.getKeyStroke(KeyEvent.VK_A, InputEvent.CTRL_MASK);
            case GLASS:
                return KeyStroke.getKeyStroke(KeyEvent.VK_1, InputEvent.CTRL_MASK);
            case QUICK_KEYS:
                return KeyStroke.getKeyStroke(KeyEvent.VK_2, InputEvent.CTRL_MASK);
            case LIST_SESSIONS:
                return KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_MASK);

            case SHOW_FB:
                return KeyStroke.getKeyStroke(KeyEvent.VK_1, InputEvent.ALT_MASK);
            case SHOW_FT:
                return KeyStroke.getKeyStroke(KeyEvent.VK_2, InputEvent.ALT_MASK);
            case SHOW_LG:
                return KeyStroke.getKeyStroke(KeyEvent.VK_3, InputEvent.ALT_MASK);

            case SHOW_CONNECTION_DOCTOR:
                return KeyStroke.getKeyStroke(KeyEvent.VK_B, InputEvent.CTRL_MASK);

            case CLIPBOARD:
                return KeyStroke.getKeyStroke(KeyEvent.VK_C, InputEvent.CTRL_MASK);
            case CTRLALTDEL:
                return KeyStroke.getKeyStroke(KeyEvent.VK_X, InputEvent.CTRL_MASK);
            case ALTGR_EUR:
                return KeyStroke.getKeyStroke(KeyEvent.VK_E, InputEvent.ALT_GRAPH_MASK);
            case ALTGR_AT:
                return KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.ALT_GRAPH_MASK);
            case REFRESH:
                return KeyStroke.getKeyStroke(KeyEvent.VK_F5, InputEvent.ALT_MASK);
            case ALTF4:
                return KeyStroke.getKeyStroke(KeyEvent.VK_Y, InputEvent.CTRL_MASK);
            case META:
                return KeyStroke.getKeyStroke(KeyEvent.VK_WINDOWS, 0);
            case RSPOWERDOWN:
                return KeyStroke.getKeyStroke(KeyEvent.VK_P, InputEvent.CTRL_MASK);
            case RSPOWERUP:
                return KeyStroke.getKeyStroke(KeyEvent.VK_P, InputEvent.CTRL_MASK | InputEvent.SHIFT_DOWN_MASK);
            case OPTIONS:
                return KeyStroke.getKeyStroke(KeyEvent.VK_O, InputEvent.CTRL_MASK);
            case RECORD:
                return KeyStroke.getKeyStroke(KeyEvent.VK_E, InputEvent.CTRL_MASK);

            //case TOGGLE_FULLSCREEN		: return KeyStroke.getKeyStroke( KeyEvent.VK_S, InputEvent.ALT_GRAPH_MASK|InputEvent.SHIFT_DOWN_MASK);
            case TOGGLE_CONN_STATUS_BAR:
                return KeyStroke.getKeyStroke(KeyEvent.VK_I, InputEvent.CTRL_MASK);
            case TOGGLE_GP_UI_ENABLED:
                return KeyStroke.getKeyStroke(KeyEvent.VK_ADD, InputEvent.CTRL_MASK);
            case DEVELOP_1:
                return KeyStroke.getKeyStroke(KeyEvent.VK_D, InputEvent.CTRL_MASK);
            case CONTROL_NAVIGATOR:
                return KeyStroke.getKeyStroke(KeyEvent.VK_K, InputEvent.CTRL_MASK);
            default:
                return null;
        }
    }

    public static String getDefaultKeyStrokeAsString(Goal g) {
        KeyStroke scut = getDefaultKeyStroke(g);
        if (scut == null) return "";

        String modi = KeyEvent.getKeyModifiersText(scut.getModifiers());
        String code = KeyEvent.getKeyText(scut.getKeyCode());
        return modi.isEmpty() ? code : modi + " + " + code;
    }
}
