package com.wincornixdorf.vnc.global;

//
//  Copyright (C) 2001-2004 HorizonLive.com, Inc.  All Rights Reserved.
//  Copyright (C) 2002 Constantin Kaplinsky.  All Rights Reserved.
//  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
//
//  This is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This software is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this software; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
//  USA.
//

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.KeyEventDispatcher;
import java.awt.KeyboardFocusManager;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.Date;
import java.util.Observable;
import java.util.Observer;
import java.util.logging.ConsoleHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.X509TrustManager;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.JApplet;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JSeparator;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

import com.wincornixdorf.vnc.global.VNCAction.Goal;
import com.wincornixdorf.vnc.gui.ConnectionDoctor;
import com.wincornixdorf.vnc.gui.GlassPane;
import com.wincornixdorf.vnc.gui.JAboutDialog;
import com.wincornixdorf.vnc.gui.JButtonPanel;
import com.wincornixdorf.vnc.gui.JRecordingFrame;
import com.wincornixdorf.vnc.gui.RemoteDesktopPanel;
import com.wincornixdorf.vnc.gui.VNCSessionPanel;
import com.wincornixdorf.vnc.gui.VNCSessionPoolPanel;
import com.wincornixdorf.vnc.rfb.RFBConnJob;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent;
import com.wincornixdorf.vnc.rfb.VNCSession;
import com.wincornixdorf.vnc.rfb.VNCSessionPool;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;

public class VNCViewer extends JApplet
        implements WindowListener, Observer, ListDataListener {
    private static final String app_Name = "Viewer Applet";

    private static final long serialVersionUID = -2140907750194877719L;

    // TODO: This shared logger was a bad idea from the start. Using it to collect messages for the built in log view was even worse. 
    public static final Logger pen = Logger.getLogger("com.wincornixdorf.vnc");

    public static final String ASK_THE_DOCTOR = "Please consult the 'Connection Doctor' from the 'Help' menu for additional information";
    public static final String DONT_ASK_THE_DOCTOR = "Please inform vendor, if problem persists.";

    private final VNCViewerOptions options = new VNCViewerOptions();

    private JMenuBar menuBar = new JMenuBar();

    private static GlassPane glasspane = null;
    private static LogToWebServiceAppender wslogger = null;

    // TODO: This is a legacy sin that deserves extra redemption!
    public static SSLContext sslContext = null;

    /**
     * This variable indicates whether the user is has "God mode" turned on.
     * While the "God mode" won't make you omnipotent, it will give the user
     * access to functions and methods not intended for mere mortals.
     * Use it at your own peril.
     * Before you ask: It used to be a boolean variable, but to prevent the
     * application from spamming the HDD with requests for the file expected
     * to only reside in the users home directory of users who are supposed
     * to have these privileges, this change had to be made.
     * -1 = Unknown | 0 = definitely Not divine | 1 = clearly divine
     */
    private static byte divinityStatus = -1;

    private JFrame mainFrame = null;

    private VNCSessionPoolPanel pp = null;
    private VNCSessionPool sessions = new VNCSessionPool();
    public VNCActionsPool actions = new VNCActionsPool(this);

    private Container bgw = null;
    private JAboutDialog aboutDlg = null;

    private JButtonPanel buttonPanel = null;
    public JRecordingFrame rec = null;

    private ConsoleHandler console = null;

    private static String versionStr = "UNKNOWN-VERSION";
    private static String buildDateStr = "UNKNOWN-BUILD-DATE";

    /**
     * This method is called when run as a java program from the command line.
     * It simply runs the Applet inside a newly created frame.
     *
     * @param argv CMDLN arguments. Nothing to say really.
     */
    public static void main(String[] argv) {
        VNCViewer v = new VNCViewer();

        try {
            VNCViewerOptionsAndPreferencesAssisstant.initialize(v.options, argv);
        } catch (MissingParameterException mpe) {
            String cause = mpe.getCause() != null ? mpe.getCause().getMessage() : "Root cause unknown.";
            v.fatalError(mpe.getMessage() + "\n" + cause);
        }

        v.init();
        v.start();
    }

    public static String appBaseline(boolean extended) {
        String v = extended ? " Version " : " V ";
        StringBuilder sb = new StringBuilder();

        sb.append(app_Name).append(v).append(versionStr);
        if (extended) sb.append(" Built on (").append(buildDateStr).append(")");

        return sb.toString();
    }

    public static boolean isDivine() {
        // divinityStatus -1 = Unknown | 0 = definitely Not divine | 1 = clearly divine
        if (divinityStatus != -1) return divinityStatus == 1;
        try {
            File f = new File(System.getProperty("user.home") + File.separatorChar + ".iddqd_DIC");
            if (f != null && f.exists()) divinityStatus = 1;
            else divinityStatus = 0;
        } catch (Exception e) {
            divinityStatus = 0;
        }
        return divinityStatus == 1;
    }

    public static boolean isGlasspaneVisible() {
        return glasspane.isVisible();
    }

    public static boolean isSnapshot() {
        return versionStr.endsWith("SNAPSHOT");
    }

    @Override
    public void init() {
        tabKeyWorkaround();
        // The options get "sealed", which means the settings therein can't be
        // overwritten accidentally by invoking initialize a second time as may
        // be the case when executing the code as an application. 
        try {
            VNCViewerOptionsAndPreferencesAssisstant.initialize(options, this);
        } catch (MissingParameterException mpe) {
            fatalError(mpe.getMessage());
        }

        // Read the version Number and Generation Date from the Maven POM properties file (prerequisite 4 setUpLogging!)
        initializeVersion();

        // Set up basic logging
        setUpLogging();

        // If encryption is desired, this this will create the required SSLContext.
        setUpSSLContext();

        // Create the GUI in a thread of its own as recommended by SUN.
        try {
            SwingUtilities.invokeAndWait(new Runnable() {
                public void run() {
                    createGUI();
                    pen.log(Level.FINEST, "The GUI was created.");
                }
            });
        } catch (Exception e) {
            System.err.println("Failed to create the GUI.");
            e.printStackTrace();
        }

        sessions.addListDataListener(this);
        pen.log(Level.FINEST, "The applet is done initializing!");
    }

    private void initializeVersion() {
        // TODO: Make finding the Maven POM Props file independent of the class structure used
        URL mavenPomProps = getClass().getResource("/META-INF/maven/com.wincornixdorf.vnc/Viewer/pom.properties");

        if (mavenPomProps == null) {
            pen.log(Level.WARNING, "Failed to read the applications version!");
            return;
        }

        try {
            InputStream is = mavenPomProps.openStream();
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);

            // get me past the line expected to read "Generated by Maven"
            String hs = br.readLine();

            // second line is expected to read the date the jar was generated preceded by a '#'
            hs = br.readLine();
            //System.out.println( "BuildDateStr: " + hs);
            if (hs != null) buildDateStr = hs.substring(1);

            // third line is expected to read "version='number'"
            hs = br.readLine();
            //System.out.println( "versionStr  : " + hs);
            if (hs != null) versionStr = hs.substring(8);

            // You open it ... You close it!
            is.close();
        } catch (IOException e) {
            e.printStackTrace();
            pen.log(Level.WARNING, "Failed to extract the applications version from the pom file!");
        }
    }

    /**
     * Implementation of the start method in JApplet. Will either be invoked
     * by the browser if run as an Applet or statically from the main method if
     * run as a standalone application.
     */
    @Override
    public void start() {
        //        try {
        //            SwingUtilities.invokeLater( this);
        //        } catch ( Exception e) {
        //            System.err.println( "Failed to create the GUI.");
        //            e.printStackTrace();
        //        }

        // If a session has been configured trough command-line/applet
        // parameters, a session is being created
        createSession(new VNCSessionCfg(options));

        //pen.log( Level.FINEST, "The initial thread has been terminated.");
    }

    /**
     * Implementation of the stop method in JApplet.
     */
    @Override
    public void stop() {
        pen.log(Level.FINEST, "The Applet is about to be stopped.");
        super.stop();
        //sessions.closeAll();
        pen.log(Level.FINEST, "The Applet is done stopping.");
    }

    /**
     * This method is called before the Applet is destroyed.
     */
    @Override
    public void destroy() {
        pen.log(Level.FINEST, "The Applet is about to be destroyed.");
        onExit();
        super.destroy();
        pen.log(Level.FINEST, "The Applet has been destroyed.");
    }


    /**
     * Writes the argument to the log and kills the application using bailOut.
     *
     * @param str An error message.
     */
    public void fatalError(String str) {
        pen.log(Level.SEVERE, str);
        doTerminate(1);
    }

    private void doTerminate(int errCode) {
        if (!options.isRunAsApplet()) {
            System.exit(errCode);
        } else {
            stop();
            System.exit(errCode);
        }

    }

    /**
     * Will clean up whatever resources have been allocated in the past.
     * Developers are strongly encouraged to call this method before closing
     * the application/Applet for good. Developers interested in closing the
     * present connection should call disconnect.
     */
    private void onExit() {
        pen.log(Level.FINEST, "On Exit got called!");

        disconnect();
        pen.log(Level.FINEST, "Disconnected from Server!");

        bgw = null;

        if (mainFrame != null) {
            mainFrame.dispose();
            mainFrame = null;
        }
        if (aboutDlg != null) {
            aboutDlg.dispose();
            aboutDlg = null;
        }
        if (rec != null) {
            rec.dispose();
            rec = null;
        }

        // This is a clean exit ...
        doTerminate(0);
    }

    /**
     * Will try its best to close the application down as cleanly as possible.
     */
    public void quit() {
        onExit();
    }

    /**
     * Implementation of the WindowListener interface. Does some basic clean up.
     */
    @Override
    public void windowClosing(WindowEvent evt) {
        onExit();
    }

    @Override
    public void windowActivated(WindowEvent evt) {
    }

    @Override
    public void windowDeactivated(WindowEvent evt) {
    }

    @Override
    public void windowOpened(WindowEvent evt) {
    }

    @Override
    public void windowClosed(WindowEvent evt) {
    }

    @Override
    public void windowIconified(WindowEvent evt) {
    }

    @Override
    public void windowDeiconified(WindowEvent evt) {
    }

    /**
     * Closes the connection to the server.
     */
    public void disconnect() {
        //if ( SwingUtilities.isEventDispatchThread()) VncViewer.pen.log( Level.SEVERE, ">>>> This is the EDT!!!!!!!!!!");
        //else VncViewer.pen.log( Level.SEVERE, ">>>> This is some other thread");

        VNCSession act = getActiveSession();
        if (act == null || pp.active() == null) return;

        // Save the present configuration as a set of preferences
        pp.active().extract(act.getConfiguration()); // TODO: Get rid of this JUNK!
        VNCViewerOptionsAndPreferencesAssisstant.persist(act.getConfiguration());

        sessions.destroySession(act);
    }

    private VNCSession createSession(VNCSessionCfg cfg) {
        if (cfg == null || cfg.vncserver == null) return null;
        if (!cfg.vncserver.isConfigured()) {
            String msg = "Unable to open a session using pseudo URL\n[" +
                    cfg.vncserver.getPseudoURL() + "]";
            JOptionPane.showMessageDialog(bgw, msg, "Error", JOptionPane.ERROR_MESSAGE);
            return null;
        }

        // Restore the session-preferences from wherever they where stored
        VNCViewerOptionsAndPreferencesAssisstant.restore(cfg);
//        stealth.setStealthMode( cfg.gb.getStealthMode());
        return sessions.spawnSession(cfg);
    }

    public void disposeOfCurrentSession() {
        VNCSession act = getActiveSession();
        if (act == null || pp.active() == null) return;

        String msg = "<html>Would you like to <font color=red>close and discard</font> the current session ?</html>";
        int n = JOptionPane.showConfirmDialog(bgw, msg,
                "Close current session", JOptionPane.YES_NO_OPTION);

        if (n == JOptionPane.CLOSED_OPTION || n == JOptionPane.NO_OPTION) return;
        sessions.destroySession(act);
    }

    public VNCSession createNewSession() {
        String msg = "<html>The createNewSessionMethod has been disabled for security reasons.<br>" +
                "If you see this dialog by chance, you are probably the sort of hacker<br>" +
                "my mother kept warning me about! <font color=red>GO AWAY!!!!!!</font><html>";
        JOptionPane.showMessageDialog(bgw, msg, "Error", JOptionPane.ERROR_MESSAGE);
        return null;
        //    SessionSetupPanel ssp = new SessionSetupPanel( isDevine);
        //
        //    JOptionPane op = new JOptionPane( ssp, JOptionPane.QUESTION_MESSAGE,
        //         JOptionPane.OK_CANCEL_OPTION);
        //
        //    JDialog dialog = op.createDialog( bgw, "Create new Session");
        //    dialog.show();
        //    Object selectedValue = op.getValue();
        //
        //    if ( selectedValue == null || ! ( selectedValue instanceof Integer)) return null;
        //    Integer btn = (Integer)selectedValue;
        //
        //    if ( btn == JOptionPane.CANCEL_OPTION) return null;
        //
        //    VNCSessionCfg cfg = new VNCSessionCfg();
        //    cfg.vncserver = new VNCSrvCfg( ssp.tfURL.getText());
        //    cfg.vncserver.pEnd.password = ssp.pfPWD.getText();
        //    return createSession( cfg);
    }

    public void closeConnection(boolean fb) {
        VNCSession act = getActiveSession();
        if (act == null || pp.active() == null) return;
        RFBHandler conn2close = fb ? act.getFB() : act.getFT();
        if (conn2close != null) conn2close.queue(RFBConnJob.disconnect(conn2close));
    }

    public void openConnection(boolean fb) {
        VNCSession act = getActiveSession();
        if (act == null || pp.active() == null) return;
        RFBHandler conn2open = fb ? act.getFB() : act.getFT();
        if (conn2open != null) conn2open.queue(RFBConnJob.connect(conn2open));
    }

    public void resetConnection(boolean fb) {
        VNCSession act = getActiveSession();
        if (act == null || pp.active() == null) return;
        RFBHandler conn2reset = fb ? act.getFB() : act.getFT();
        conn2reset.queue(RFBConnJob.disconnect(conn2reset));
        conn2reset.queue(RFBConnJob.connect(conn2reset));
    }

    public void openConnections() {
        openConnection(true);
        openConnection(false);
    }

    public void closeConnections() {
        closeConnection(true);
        closeConnection(false);
    }

    public void resetConnections() {
        // not implemented yet
    }

    /**
     * @return A reference to the {@link VNCSession} currently active in
     * the {@link VNCSessionPool} or null if there is no active Session.
     */
    public VNCSession getActiveSession() {
        return pp.getActiveSession();
    }

    /**
     * Will examine all VNCActions and set them en-/disabled depending on the
     * present general status of the application. Call it whenever there is reason
     * to suspect an Action may have gained or forfeit its enabled status. This
     * method must be invoked on the EDT.
     */
    public void checkActionsAvailability() {

        VNCSession sess = getActiveSession();

        VNCSessionPanel sp = pp.active();

        boolean hasSession = sess != null;
        //boolean isConnected         = false;
        //boolean lostAConnection     = true;
        boolean hasFBConnection = false;
        boolean hasFTConnection = false;
        boolean isFBOnly = options.isFBOnly();
        boolean isFTOnly = options.isFTOnly();
        boolean isFBNDFT = options.isBothFBAndFT();
        boolean isShowingFB = sp != null ? sp.isShowingFB() : false;
        //boolean isShowingFT         = sp != null ? sp.isShowingFT() : false;
        boolean isShowingLG = sp != null ? sp.isShowingLG() : false;

        if (hasSession) {
            //isConnected         = sess.hasOpenConnection();
            //lostAConnection     = sess.hasBeenCrippled();
            hasFBConnection = sess.hasOpenConnection(true);
            hasFTConnection = sess.hasOpenConnection(false);
        }

//        actions.options.setEnabled( !isShowingLG && ( hasFBConnection || hasFTConnection));
        actions.setEnabled(VNCAction.Goal.QUIT, !options.isRunAsApplet());
        actions.setEnabled(VNCAction.Goal.OPTIONS, !isShowingLG);
        actions.setEnabled(VNCAction.Goal.CLIPBOARD, isShowingFB && !isFTOnly && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.META, isShowingFB && !isFTOnly && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.REFRESH, isShowingFB && !isFTOnly && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.RECORD, false);


        //ctrl + alt + delete is disabled on remote support, because it has not supported yet
        if (!sess.getHost().isRemoteSupport()) {
            actions.setEnabled(VNCAction.Goal.CTRLALTDEL, isShowingFB && !isFTOnly && hasFBConnection);
        } else {
            actions.setEnabled(VNCAction.Goal.CTRLALTDEL, false);
        }

        actions.setEnabled(Goal.CHANGE_KEYBOARD_LAYOUT, true);
        actions.setEnabled(Goal.DISABLE_CONTROLS, isShowingFB && !isFTOnly && hasFBConnection);

        actions.setEnabled(VNCAction.Goal.RSPOWERDOWN, isShowingFB && !isFTOnly && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.RSPOWERUP, isShowingFB && !isFTOnly && hasFBConnection);

        actions.setEnabled(VNCAction.Goal.LIST_SESSIONS, isDivine());
        actions.setEnabled(VNCAction.Goal.GLASS, true);

        actions.setEnabled(VNCAction.Goal.SHOW_FT, !isFBOnly && hasSession);
        actions.setEnabled(VNCAction.Goal.SHOW_FB, !isFTOnly && hasSession);
        actions.setEnabled(VNCAction.Goal.SHOW_LG, hasSession);
        actions.setEnabled(VNCAction.Goal.SHOW_CONNECTION_DOCTOR, hasSession && (!hasFBConnection || !hasFTConnection));

        actions.setEnabled(VNCAction.Goal.CLOSE_FB_CONNECTION, !isFTOnly && hasSession && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.RESET_FB_CONNECTION, !isFTOnly && hasSession && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.CLOSE_FT_CONNECTION, !isFBOnly && hasSession && hasFTConnection);
        actions.setEnabled(VNCAction.Goal.RESET_FT_CONNECTION, !isFBOnly && hasSession && hasFTConnection);

        actions.setEnabled(VNCAction.Goal.OPEN_BOTH_CONNECTIONS, isFBNDFT && (!hasFBConnection || !hasFTConnection));
        actions.setEnabled(VNCAction.Goal.CLOSE_BOTH_CONNECTIONS, isFBNDFT && (hasFBConnection || hasFTConnection));

        actions.setEnabled(VNCAction.Goal.CREATE_SESSION, isDivine() && !sessions.isFull());
        actions.setEnabled(VNCAction.Goal.DELETE_SESSION, isDivine() && !sessions.isEmpty());

        actions.setEnabled(VNCAction.Goal.QUICK_KEYS, isShowingFB && !isFTOnly);
        actions.setEnabled(VNCAction.Goal.TOGGLE_FULLSCREEN, VNCViewer.isDivine());
        actions.setEnabled(VNCAction.Goal.TOGGLE_CONN_STATUS_BAR, pp.isConnectionStatusBarToggable());
        actions.setEnabled(VNCAction.Goal.ALTTAB, isShowingFB && !isFTOnly && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.BATTLA, isShowingFB && !isFTOnly && hasFBConnection);
        actions.setEnabled(VNCAction.Goal.ALTF4, isShowingFB && !isFTOnly && hasFBConnection);
    }

    /**
     * Forces the VNCCanvas to accept the keyboard input focus.
     */
    private void moveFocusToDesktop() {
        if (pp.active().getInputComponent() == null) return;
        pp.active().getInputComponent().requestFocusInWindow();
    }

    /**
     * Switches the new GlasspaneUI on/off, depending on the current setting of the actions selected value.
     */
    public void toggleUIMode() {
        Boolean useGPUI = (Boolean) (actions.getActionByGoal(Goal.TOGGLE_GP_UI_ENABLED).getValue(Action.SELECTED_KEY));
        //VNCViewer.pen.log( Level.WARNING, "Toggling GLASSPANE UI Mode [" + useGPUI + "]!");
        menuBar.setVisible(!useGPUI);
        buttonPanel.setVisible(!useGPUI);
        actions.getActionByGoal(Goal.TOGGLE_GP_UI_ENABLED).putValue(Action.SELECTED_KEY, new Boolean(!useGPUI));
    }

    /**
     * Will hide or show the JFrame with the display options for the present
     * Session.
     */
    public void toggleOptionsVisiblity() {
        // Read the current "visible" status held in the action itself
        //Boolean b = (Boolean)(actions.options.getValue( Action.SELECTED_KEY));
        pp.setOptionsVisible(!pp.areOptionsVisible());
        RemoteDesktopPanel rdp = pp.active().fbWidget;
        VNCViewer.pen.severe("Here we go ... " + rdp.getSize() + " " + rdp.getPreferredSize());
    }

    /**
     * Will hide or show the JFrame with the About widget.
     */
    public void toggleAboutVisiblity() {
        if (aboutDlg == null) {
            aboutDlg = new JAboutDialog(options);
        }

        aboutDlg.setLocationRelativeTo(pp);
        aboutDlg.setVisible(!aboutDlg.isVisible());
    }

    /**
     * Will hide or show the JFrame with the present Clipboard contents.
     */
    public void toggleClipboardVisiblity() {
        //Boolean b = (Boolean)(actions.clipboard.getValue( Action.SELECTED_KEY));
//        actions.clipboard.putValue( Action.SELECTED_KEY, !b);
        pp.setClipboardsVisible(!pp.areClipboardsVisible());
    }

    public void toggleGlassPane() {
        glasspane.activateActionSet(GlassPane.DEF_ACTIONSET, 4, 5);
        glasspane.setVisible(!glasspane.isVisible());
    }

    public void toggleQuickKeys() {
        glasspane.activateActionSet(GlassPane.KEYS_ACTIONSET, 2, 3);
        glasspane.setVisible(!glasspane.isVisible());
    }

    public void toggleFullScreen() {
        Boolean b = (Boolean) (actions.getActionByGoal(Goal.TOGGLE_FULLSCREEN).getValue(Action.SELECTED_KEY));

        if (!options.isRunInSeparateFrame()) return;
        JFrame f = ((JFrame) bgw);
        f.setVisible(false);

        GraphicsEnvironment genv = GraphicsEnvironment.getLocalGraphicsEnvironment();
        GraphicsDevice gd = genv.getDefaultScreenDevice();

        if (b) mainScreenTurnOn(f, gd);
        else mainScreenTurnOff(f, gd);
    }

    public void toggleSessionPicker() {
        Boolean futureSessionPanelVisibility = new Boolean(!pp.panelSessLst.isVisible());
        pp.panelSessLst.setVisible(futureSessionPanelVisibility);
        // Make sure ALL toggle buttons offering this functionality are set properly
        actions.getActionByGoal(Goal.LIST_SESSIONS).putValue(Action.SELECTED_KEY, futureSessionPanelVisibility);
    }

    public void toggleConnectionStatusBar() {
        VNCViewer.pen.log(Level.WARNING, "Toggling Status Bar!");
        pp.active().toggleStatusBar();
    }

    public void showCurrentSessionsFB() {
        VNCSessionPanel sp = pp.active();
        if (sp != null) sp.showFB();
        else VNCViewer.pen.log(Level.WARNING, "request showCurrentSessionsFB failed!");
        checkActionsAvailability();
    }

    public void showCurrentSessionsFT() {
        VNCSessionPanel sp = pp.active();
        if (sp != null) sp.showFT();
        else VNCViewer.pen.log(Level.WARNING, "request showCurrentSessionsFT failed!");
        checkActionsAvailability();
    }

    public void showCurrentSessionsLG() {
        VNCSessionPanel sp = pp.active();
        if (sp != null) sp.showLG();
        checkActionsAvailability();
    }

    public void showControlNavigator() {
        VNCSessionPanel vncSessionPanel = pp.active();
        if (vncSessionPanel != null) {
            Boolean showCtrlNav = (Boolean) (actions.getActionByGoal(Goal.CONTROL_NAVIGATOR).getValue(Action.SELECTED_KEY));
            vncSessionPanel.fbWidget.setShowControlNavigator(!showCtrlNav);
            actions.getActionByGoal(Goal.CONTROL_NAVIGATOR).putValue(Action.SELECTED_KEY, !showCtrlNav);
        }
    }

    public void developAction1() {
        if (!isDivine()) return;
        System.err.println("You are home!");

        VNCSession s = getActiveSession();
        if (s == null) return;

        Date d = new Date();
        Component c = pp.active().getInputComponent();

        moveFocusToDesktop();
        //KeyEvent(Component source, int id, long when, int modifiers, int keyCode, char keyChar)
        KeyEvent ke = new KeyEvent(c, KeyEvent.KEY_PRESSED, d.getTime(), 0, 127, (char) KeyEvent.VK_DELETE);
//        KeyEvent ke = new KeyEvent( c, KeyEvent.KEY_PRESSED, d.getTime(), 0, 8, (char)KeyEvent.VK_BACK_SPACE);
        s.getFB().queue(RFBConnJob.pressKey(s.getFB(), ke));

//        KeyStroke ksDelete = KeyStroke.getKeyStroke( 8, 0);       
//        s.getFB().queue( RFBConnJob.pressKeyStroke( s.getFB(), ksDelete, pp.active().getInputComponent()));
    }

    /**
     * Will send a request to the server asking it to send the ENTIRE screen
     * as if it had been modified. On receiving the response, the entire Canvas
     * will be redrawn. This code is presently not in use.
     */
    public void requestScreenRefresh() {
        VNCSession s = getActiveSession();
        if (s == null) return;

        s.getFB().queue(RFBConnJob.refresh(s.getFB()));
        moveFocusToDesktop();
    }

    /**
     * Sends an event to the server simulating the Ctrl+Alt+Delete keys had been pressed.
     */
    public void requestCtrlAltDeleteKeypressBeSent() {
        VNCSession s = getActiveSession();
        if (s == null) return;

        moveFocusToDesktop();
        s.getFB().queue(RFBConnJob.pressMonkeyKey(s.getFB(), pp.active().getInputComponent()));
        pen.log(Level.FINEST, "Ctrl + Alt + Del keystroke sent. ");

    }

    public void requestAltTabKeypressBeSent(boolean shift) {
        VNCSession s = getActiveSession();
        if (s == null) return;

        moveFocusToDesktop();
        KeyStroke KS_ALT_TAB = KeyStroke.getKeyStroke("alt TAB");
        KeyStroke KS_SHIFT_ALT_TAB = KeyStroke.getKeyStroke("shift alt TAB");
        KeyStroke e = shift ? KS_SHIFT_ALT_TAB : KS_ALT_TAB;

        s.getFB().queue(RFBConnJob.pressKeyStroke(s.getFB(), e, pp.active().getInputComponent()));
        pen.log(Level.FINEST, "Alt + TAB keystroke sent. " + e);

    }

    public void requestAltF4KeypressBeSent() {
        VNCSession s = getActiveSession();
        if (s == null) return;

        moveFocusToDesktop();
        KeyStroke altF4 = KeyStroke.getKeyStroke("alt F4");

        s.getFB().queue(RFBConnJob.pressKeyStroke(s.getFB(), altF4, pp.active().getInputComponent()));
        pen.log(Level.FINEST, "Alt + F4 keystroke sent. " + altF4);

    }

    public void requestAltGrEurKeypressBeSent() {
        VNCSession s = getActiveSession();
        if (s == null) return;

        moveFocusToDesktop();
        // we do not need to release button, because event will be evaluated as shortcut
        KeyEvent ke = new KeyEvent(pp.active().getInputComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), 0, KeyEvent.VK_EURO_SIGN, '\u20ac');
        s.getFB().queue(RFBConnJob.pressKey(s.getFB(), ke));
    }

    public void requestAltGrAtKeypressBeSent() {
        VNCSession s = getActiveSession();
        if (s == null) return;

        moveFocusToDesktop();
        // we do not need to release button, because event will be evaluated as shortcut
        KeyEvent ke = new KeyEvent(pp.active().getInputComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), 0, KeyEvent.VK_AT, '\u0040');
        s.getFB().queue(RFBConnJob.pressKey(s.getFB(), ke));
    }

    /**
     * Sends an event to the server simulating the Windows Start key had been pressed.
     */
    public void requestMetaKeypressBeSent() {
        VNCSession s = getActiveSession();
        if (s == null) return;

        moveFocusToDesktop();
        s.getFB().queue(RFBConnJob.pressMetaKey(s.getFB(), pp.active().getInputComponent()));
        pen.log(Level.FINEST, "Windows key sent. ");

    }

    /**
     * Sends a request to the VNC server, asking it to power down its displays
     * so no one can see what's being done. Be advised this is only a request. If
     * the server does not support the operation the remote display will not be
     * blanked. The server does not give any feedback as to whether it got the
     * request in the first place, nor whether it actually shut down the display.
     *
     * @param off Set true, if the screen is to be turned off and false if it is to be switched on.
     */
    public void requestRemoteScreenBeTurned(boolean off) {
        VNCSession s = getActiveSession();
        if (s == null) return;

        if (off) s.getFB().queue(RFBConnJob.turnConsoleOff(s.getFB()));
        else s.getFB().queue(RFBConnJob.turnConsoleOn(s.getFB()));
    }

    private void createGUI() {
        if (options.isRunInSeparateFrame()) {
            mainFrame = new JFrame(appBaseline(false));
            // mainFrame.setUndecorated( true);
            mainFrame.setExtendedState(JFrame.MAXIMIZED_BOTH);
            mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
            mainFrame.addWindowListener(this);
            mainFrame.setJMenuBar(menuBar);
            mainFrame.setSize(1280, 1024);
            mainFrame.setVisible(true);
            /*if ( !options.isRunAsApplet())*/
            mainFrame.add("Center", this);
            bgw = mainFrame;
            bgw.setName("MainFrame");
        } else {
            bgw = this.getContentPane();
            bgw.setName("AppletContentPane");
            this.setJMenuBar(menuBar);
        }

        pp = new VNCSessionPoolPanel(this, sessions);

        glasspane = new GlassPane(pp);
        glasspane.addActionSet(GlassPane.DEF_ACTIONSET, actions.getMainGlasspaneActions(options.isRunAsApplet()));
        glasspane.addActionSet(GlassPane.KEYS_ACTIONSET, actions.getKeyRequestActions());
        glasspane.activateActionSet("", 4, 4);

        if (options.isRunInSeparateFrame()) mainFrame.setGlassPane(glasspane);
        else setGlassPane(glasspane);

        bgw.setLayout(new BorderLayout());
        bgw.setPreferredSize(Toolkit.getDefaultToolkit().getScreenSize());

        JMenu mFile = new JMenu("File");
        mFile.add(actions.getActionByGoal(VNCAction.Goal.RECORD));
        mFile.addSeparator();
        mFile.add(actions.getActionByGoal(VNCAction.Goal.TOGGLE_GP_UI_ENABLED));
        mFile.add(actions.getActionByGoal(VNCAction.Goal.GLASS));
        mFile.add(actions.getActionByGoal(VNCAction.Goal.QUIT));
        menuBar.add(mFile);

        JMenu mViews = new JMenu("Views");
        JCheckBoxMenuItem cbmiFB = new JCheckBoxMenuItem(actions.getActionByGoal(VNCAction.Goal.SHOW_FB));
        JCheckBoxMenuItem cbmiFT = new JCheckBoxMenuItem(actions.getActionByGoal(VNCAction.Goal.SHOW_FT));
        JCheckBoxMenuItem cbmiLG = new JCheckBoxMenuItem(actions.getActionByGoal(VNCAction.Goal.SHOW_LG));
        ButtonGroup bg = new ButtonGroup();
        mViews.add(cbmiFB);
        bg.add(cbmiFB);
        mViews.add(cbmiFT);
        bg.add(cbmiFT);
        mViews.add(cbmiLG);
        bg.add(cbmiLG);
        if (isDivine()) {
            mViews.addSeparator();
            mViews.add(new JCheckBoxMenuItem(actions.getActionByGoal(VNCAction.Goal.TOGGLE_FULLSCREEN)));
        }
        //mViews.add( new JCheckBoxMenuItem( actions.toggleMenu));
        menuBar.add(mViews);

        if (isDivine()) {
            JMenu mSessions = new JMenu("Sessions");
            mSessions.add(new JCheckBoxMenuItem(actions.getActionByGoal(VNCAction.Goal.LIST_SESSIONS)));
            mSessions.addSeparator();
            mSessions.add(actions.getActionByGoal(VNCAction.Goal.CREATE_SESSION));
            mSessions.add(actions.getActionByGoal(VNCAction.Goal.DELETE_SESSION));
            menuBar.add(mSessions);
        }

        JMenu mConnections = new JMenu("Connection");

        if (options.isFBOnly() || options.isBothFBAndFT()) {
            mConnections.add(actions.getActionByGoal(VNCAction.Goal.CLOSE_FB_CONNECTION));
        } else {
            mConnections.add(actions.getActionByGoal(VNCAction.Goal.CLOSE_FT_CONNECTION));
        }
        menuBar.add(mConnections);

        JMenu mActions = new JMenu("Actions");
        mActions.add(actions.getActionByGoal(VNCAction.Goal.REFRESH));
        mActions.add(new JCheckBoxMenuItem(actions.getActionByGoal(VNCAction.Goal.CLIPBOARD)));
        mActions.add(new JCheckBoxMenuItem(actions.getActionByGoal(VNCAction.Goal.OPTIONS)));
        mActions.addSeparator();
        mActions.add(actions.getActionByGoal(VNCAction.Goal.RSPOWERDOWN));
        mActions.add(actions.getActionByGoal(VNCAction.Goal.RSPOWERUP));
        mActions.add(actions.getActionByGoal(Goal.DISABLE_CONTROLS));
        menuBar.add(mActions);

        JMenu mKeys = new JMenu("Keys");
        mKeys.add(actions.getActionByGoal(VNCAction.Goal.META));
        mKeys.add(actions.getActionByGoal(VNCAction.Goal.CTRLALTDEL));
        mKeys.add(actions.getActionByGoal(Goal.ALTGR_EUR));
        mKeys.add(actions.getActionByGoal(Goal.ALTGR_AT));
        mKeys.add(actions.getActionByGoal(VNCAction.Goal.QUICK_KEYS));
        menuBar.add(mKeys);

        JMenu mHelp = new JMenu("Help");
        // For now I don't want that to be seen yet ... Thats still work in progress
        if (isDivine()) {
            mHelp.add(actions.getActionByGoal(VNCAction.Goal.SHOW_CONNECTION_DOCTOR));
            mHelp.addSeparator();
        }
        mHelp.add(actions.getActionByGoal(VNCAction.Goal.CONTROL_NAVIGATOR));
        mHelp.add(actions.getActionByGoal(VNCAction.Goal.ABOUT));
        menuBar.add(mHelp);

        JSeparator jSeparator = new JSeparator(JSeparator.VERTICAL);
        jSeparator.setMaximumSize(new Dimension(10, 50));
        menuBar.add(jSeparator);
        String hostName = options.getVNCServerConnectionParameters().getLabel().split("\\(")[0];
        String hostIp = options.getVNCServerConnectionParameters().getHostParams().getOriginalHostname();
        String hostLabel = hostName + " - " + hostIp;
        JLabel lHost = new JLabel(hostLabel);
        menuBar.add(lHost);
        mainFrame.setTitle(hostLabel);


//        menuBar.setVisible( false);	// Yep ... I'm hiding the MenuBar by default.

        buttonPanel = new JButtonPanel(this/*, vsp*/);
//        stealth = new StealthJPanel( buttonPanel, BorderLayout.NORTH);
        bgw.add(buttonPanel, BorderLayout.NORTH);
        //buttonPanel.setVisible( options.wants2ShowControls()/* && !options.isFTOnly()*/);

        bgw.add(pp, BorderLayout.CENTER);

        // This is to ensure the Glasspane UI is enabled if we are divine.
        toggleUIMode();

        if (JRecordingFrame.checkSecurity()) {
            rec = new JRecordingFrame(this);
            rec.setLocationRelativeTo(bgw);
        }

        if (!isDivine() && isSnapshot()) {
            String msg = "<html>This Applet is reporting [<font color=blue>" + versionStr + "</font>] as it's<br>" +
                    "version number, which suggests this is a <font color=blue>SNAPSHOT</font>.<br>" +
                    "You may be using an <font color=red>unstable development build</font>, which<br>" +
                    "could have serious bugs.<br><br>" +
                    "<center><b>Do not use in production!</b><center></html>";
            JOptionPane.showMessageDialog(bgw, msg, "Warning", JOptionPane.WARNING_MESSAGE);
        }
    }

    /**
     * Internal method used to load a Keystore file. The method expects this
     * class to be contained in a jar File and will look for \a filename INSIDE
     * that jar.
     */
    private KeyStore loadCryptStore(boolean keystore) {
        KeyStore ks = null;
        String cryptstoreType = keystore ? "keystore" : "truststore";

        // The filename is guaranteed to be: non null, not empty, existing and not TrustAll for a keystore!
        String filename = options.getFilenameCryptstore(keystore);
        char[] pw = options.getPassword(keystore);

        pen.log(Level.FINEST, "Trying to load [" + cryptstoreType + "] from [" + filename + "]");

        if (VNCViewerOptions.isTrustAllID(filename)) {
            pen.log(Level.CONFIG, "A Truststore will not be used as the [" + VNCViewerOptions.ID_TRUST_ALL_CERTIFICATES + "] option has been set!");
            pen.log(Level.WARNING, "Without a Truststore the viewer is vulnerable to man in the middle attacks!");
            return ks;
        }

        try {
            ks = KeyStore.getInstance(VNCViewerOptions.isLocalFile(filename) ? "JCEKS" : "JKS");

            InputStream isz = null;

            if (VNCViewerOptions.isLocalFile(filename)) {
                pen.log(Level.FINE, "Trying to load [" + cryptstoreType + "] [" + filename + "] from local file.");
                File f = new File(filename.substring(7));
                isz = new FileInputStream(f);
            } else {
                pen.log(Level.FINE, "Trying to load [" + cryptstoreType + "] [" + filename + "] from jar file.");
                isz = this.getClass().getClassLoader().getResourceAsStream(filename);
            }

            ks.load(isz, pw);
            isz.close();
            pen.log(Level.CONFIG, "Cryptstore [" + filename + "] has been loaded.");
        } catch (Exception e) {
            e.printStackTrace();
            String em = "Failed to read [" + cryptstoreType + "] [" + filename + "]. Bailing out";
            System.err.println("Error: " + em);
            pen.log(Level.SEVERE, em);
            doTerminate(45);
        }
        return ks;
    }

    /**
     * Will create and make available the SSLContext object, which is required
     * for both TLS/SSL and HTTPS type connections. The initialization will be
     * skipped, if the viewer does not use any such connection mode.
     */
    private final void setUpSSLContext() {
        // WARNING!
        // ========
        // This one looks like it might become a problem someday. At the moment
        // it depends on the Servers "ConMode" if SSL is to be used or not. What if we want
        // to set up a connection via a repeater, that does not offer SSL we are stuck.
        // The Applet will NOT initiate the SSL context! (I can't imagine why one might
        // want to set up a forwarding chain like that though)
        // Instead of vncServer.ConMode check, if exists, the repeaterCfg.ConMode.
        VNCSrvCfg.ConMode cm = options.getVNCServerConnectionParameters().getConMode();
        if (cm != VNCSrvCfg.ConMode.TLS &&
                cm != VNCSrvCfg.ConMode.SSL &&
                cm != VNCSrvCfg.ConMode.HTTPS) return;

        TrustManager[] trustManagers = null;
        KeyManager[] keyManagers = null;
        KeyManagerFactory kmfSecureVNC = null;
        TrustManagerFactory tmfSecureVNC = null;

        pen.log(Level.FINEST, "Setting up the keystores ...");

        try {
            KeyStore keystorePrvKeys = loadCryptStore(true);
            KeyStore keystoreTrusted = loadCryptStore(false);

            kmfSecureVNC = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
            kmfSecureVNC.init(keystorePrvKeys, options.getPassword(true));
            keyManagers = kmfSecureVNC.getKeyManagers();


            if (keystoreTrusted != null) {
                pen.log(Level.CONFIG, "Using [" + options.getFilenameCryptstore(false) + "] as Truststore!");
                tmfSecureVNC = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
                tmfSecureVNC.init(keystoreTrusted);
                trustManagers = tmfSecureVNC.getTrustManagers();
            } else {
                pen.log(Level.WARNING, "No Truststore was defined! ANY CERTIFICATE WILL BE ACCEPTED!");
                trustManagers = new TrustManager[]{
                        new X509TrustManager() {
                            public java.security.cert.X509Certificate[] getAcceptedIssuers() {
                                return null;
                            }

                            public void checkClientTrusted(java.security.cert.X509Certificate[] certs, String authType) {
                            }

                            public void checkServerTrusted(java.security.cert.X509Certificate[] certs, String authType) {
                            }
                        }
                };
            }
        } catch (Exception e) {
            e.printStackTrace();
            String em = "Failed to set up Keystore. Bailing out";
            System.err.println("Error: " + em);
            pen.log(Level.SEVERE, em);
            doTerminate(46);
        }

        // Initialize and set up the SSL Context.
        try {
            sslContext = SSLContext.getInstance("TLS");
            sslContext.init(keyManagers, trustManagers, new SecureRandom());
        } catch (NoSuchAlgorithmException nsa) {
            String em = "Could not set up the SSL context!";
            System.err.println("Error: " + em);
            pen.log(Level.SEVERE, em, nsa);
            nsa.printStackTrace();
        } catch (KeyManagementException kme) {
            String em = "Could not set up the Key Manager!";
            System.err.println("Error: " + em);
            pen.log(Level.SEVERE, em, kme);
            kme.printStackTrace();
        }
    }

    /**
     * Checks if the web logger has set up. The method will only check if the
     * appropriate objects are in place, but it will NOT attempt to verify if
     * messages written to the logger actually arrive at their destination.
     * <p>
     * This "security feature" of course is absolute rubbish. It's like setting
     * the fox to keep the geese. The idea is to only start a session, if the
     * application can write a message to the log indicating a session is
     * initiated. Since the log server can be specified on the cmdline, an
     * attacker trying to hide their tracks only has to have the log message
     * sent to a server of their own.
     *
     * @return True if something read the first log request. Hacker, a Web Server, dev null. WHATEVER.
     */
    public static boolean canLog2Web() {
        // If in developer mode, logging does not have to work.  
        return isDivine() || (wslogger != null && wslogger.isReadyToLog());
    }

    private final void setUpLogging() {
        if (console != null) return;

        pen.setLevel(Level.ALL);

        console = new ConsoleHandler();
        console.setFormatter(new VNCLogFormatter());
        //console.setFormatter( new SimpleFormatter());
        console.setLevel(Level.FINEST);
        pen.addHandler(console);

        pen.log(Level.INFO, appBaseline(true));

        wslogger = new LogToWebServiceAppender(
                options.getLogServerParameters(),
                options.getVNCServerConnectionParameters());
        if (wslogger.isReadyToLog()) pen.addHandler(wslogger);
    }

    /**
     * This callback will be invoked, whenever the present Session has opened
     * a new connection.
     */
    private void onSessionOpenedAConnection(VNCSession s, RFBHandlerEvent e) {
        checkActionsAvailability();
        if (s == null || e == null) return;

        if (s.getConfiguration().lazyOpenFTConnection) {
            s.getConfiguration().lazyOpenFTConnection = false;
            RFBHandler connWhichJustGotOpened = (RFBHandler) e.getSource();
            RFBHandler conn2open = s.getFT();
            // Trying to lazy open the (FT)connection which just got opened
            // is pointless. In fact downright devastating.
            if (conn2open != null && conn2open != connWhichJustGotOpened)
                conn2open.queue(RFBConnJob.connect(conn2open));
        }
    }

    /**
     * This callback will be invoked, whenever the present Session has closed
     * an existing connection.
     */
    private void onSessionClosedAConnection(VNCSession s) {
        checkActionsAvailability();
    }

    @Override
    public void contentsChanged(ListDataEvent e) {
    }

    @Override
    public void intervalRemoved(ListDataEvent e) {
    }

    /**
     * The viewer is registered as a ListDataListener with the session pool.
     * This method will register the viewer as an observer with EVERY Session
     * being added to the sessions pool. Otherwise the viewer will not know when
     * a sessions connection is opened/closed. Additionally it will open ONE
     * of the new sessions connections. The other one (where required) gets
     * opened when the other is up to prevent server crashes.
     */
    @Override
    public void intervalAdded(ListDataEvent e) {
        VNCSession s = (VNCSession) (sessions.get(e.getIndex0()));
        if (s == null) return;
        s.addObserver(this);

        //if ( SwingUtilities.isEventDispatchThread()) VncViewer.pen.log( Level.SEVERE, "This is the EDT!!!!!!!!!! intervalAdded");
        //else VncViewer.pen.log( Level.SEVERE, "This is some other thread intervalAdded");

        if (options.isBothFBAndFT() || options.isFBOnly()) {
            RFBHandler conn2open = s.getFB();
            if (conn2open != null) conn2open.queue(RFBConnJob.connect(conn2open));
            return;
        }
        if (options.isFTOnly()) {
            RFBHandler conn2open = s.getFT();
            if (conn2open != null) conn2open.queue(RFBConnJob.connect(conn2open));
        }
    }

    /**
     * Implementation of the @link Observer interface. At the time of writing the
     * viewer is observing the "active" @link VNCSession. This is may change in the future.
     */
    @Override
    public void update(Observable o, Object arg) {
        if (o instanceof VNCSession) {
            VNCSession sess = (VNCSession) o;

            if (arg != null) {
                RFBHandlerEvent e = (RFBHandlerEvent) arg;
                // if ( e.operation == VNCConnectionEvent.Category.PHASECHANGED) {
                //     if ( e.nyu == VNCConnection.Phase.CONNECTING)    onSessionOpenedAConnection( sess);
                //     if ( e.nyu == VNCConnection.Phase.DISCONNECTING) onSessionClosedAConnection( sess);
                // }

                switch (e.operation) {
                    case CONNECT:
                        onSessionOpenedAConnection(sess, e);
                        break;
                    case DISCONNECT:
                        onSessionClosedAConnection(sess);
                        break;
                    default:
                        break;
                }
            }
        }
    }

    public void showConnectionDoctor() {
        VNCSession act = getActiveSession();
        if (act == null || pp.active() == null) return;

        Frame f = null;
        Container c = this;

        while (f == null) {
            c = c.getParent();
            f = (c instanceof Frame) ? (Frame) c : null;
        }

        ConnectionDoctor ci = new ConnectionDoctor(f, act, options);
        ci.setLocationRelativeTo(f);

        ci.setVisible(true);
    }

    private void mainScreenTurnOn(JFrame frame, GraphicsDevice gd) {
        //frame.getJMenuBar().setVisible( false);
        frame.dispose();
        frame.setUndecorated(true);
        gd.setFullScreenWindow(frame);
    }

    private void mainScreenTurnOff(JFrame frame, GraphicsDevice gd) {
        //frame.getJMenuBar().setVisible( true);
        gd.setFullScreenWindow(null);
        frame.dispose();
        frame.setUndecorated(false);
        frame.setVisible(true);
    }

    /**
     * Removes default KeyEventDispatcher from KeyboardFocusManager to enable sending TAB key to VNC server.
     */
    private void tabKeyWorkaround() {
        Container topParent = null;
        Container parent = this;

        for (int k = 0; k < 10; k++) {
            topParent = parent;
            parent = parent.getParent();
            if (parent == null) break;
        }

        try {
            if (topParent instanceof KeyEventDispatcher) {
                KeyEventDispatcher ked = (KeyEventDispatcher) topParent;
                KeyboardFocusManager kfm = KeyboardFocusManager.getCurrentKeyboardFocusManager();

                kfm.removeKeyEventDispatcher(ked);
                kfm.removeKeyEventDispatcher(ked);
            } else {
                VNCViewer.pen.log(Level.INFO, "Using Java plugin - does not need TAB key workaround.");
            }
        } catch (Exception e) {
            VNCViewer.pen.log(Level.WARNING, "Error while removing KeyEventDispatcher (TAB key workaround): " + e.getMessage());
        }
    }

    public void disableRemoteControl() {
        VNCSession activeSession = getActiveSession();
        if (activeSession != null) {
            activeSession.getFB().setDisabledRemoteControlling(!activeSession.getFB().isDisabledRemoteControlling());
        }
    }
}
