package com.wincornixdorf.vnc.global;

import java.awt.Dimension;
import java.io.InputStream;

import com.wincornixdorf.vnc.rfb.conf.LoginParameters;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;
import com.wincornixdorf.vnc.rfb.fb.PixelFormat;
import com.wincornixdorf.vnc.rfb.fb.VNCEncoding;

/**The VNCViewerOptions class is contains the options which got passed into the
Application either through the command line or as Applet parameters. These
parameters are best set using the VNCViewerOptionsAndPreferencesAssisstant,
which provides a static member for initializing this structure. The sole reason
the members in this class are all private and accessible through getters and
setters is that some of them may eventually become mutable during runtime and
this class will have to extend the Observable class.*/
public class VNCViewerOptions {
    public enum VNCMode { FB_ONLY, FBANDFT, FT_ONLY};
    
    private static final    String          SFB_ONLY                    = "VNC_ONLY";
    private static final    String          SFBANDFT                    = "VNCANDFT";
    private static final    String          SFT_ONLY                    = "FT__ONLY";

    public  static final    String          ID_TRUST_ALL_CERTIFICATES   = "TRUST_ALL";

    public  static final    String          DEFAULT_KEYSTORE            = "viewerKeys";
    public  static final    String          DEFAULT_TRUSTSTORE          = "viewerTrust";
    public  static final    String          DEFAULT_TRUSTSTORE_RS       = "viewerTrustRS";

    public  static final    VNCMode         DEFAULT_APPMODE             = VNCMode.FBANDFT;


    private                 VNCMode         appMode                     = DEFAULT_APPMODE;

    private                 VNCEncoding     fbEncoding                  = VNCEncoding.DEFAULT_ENC;

    private                 VNCSrvCfg       vncServer                   = null;
    private                 LoginParameters logServer                   = null;

    private                 boolean         blankConsole                = false;
    private                 boolean         showControls                = false;

    private                 String          fnKeystore                  = DEFAULT_KEYSTORE;
    private                 String          fnTruststore                = DEFAULT_TRUSTSTORE;

    private                 char[]          pwdKeystore                 = null;
    private                 char[]          pwdTruststore               = null;
    
    private                 String          fnRecording                 = "";

    private                 String          socketFactory               = "";
    private                 int             deferScreenUpdates          = 0;
    private                 int             deferCursorUpdates          = 0;
    private                 int             deferUpdateRequests         = 0;

    private                 boolean         inAnApplet                  = true;
    private                 boolean         inSeparateFrame             = false;

    //private                 boolean         offerRelogin                = false;
    //private                 boolean         showOfflineDesktop          = false;

    private                 Dimension       dCanvas                     = new Dimension( 1280, 1024);

    private                 boolean         sealed                      = false;

    private PixelFormat.PixelFormatQuality  pixelFormatQuality          = PixelFormat.PixelFormatQuality.EMPTY;

    /**The default constructor. Its sole purpose is to allow initiating this
    class as a static identifier. Instances of this class can be considered
    worthless unless one of the setUp methods has been invoked on them.*/
    public VNCViewerOptions() {
    }

    public void setUpForStandaloneApplicationUse() {
        inAnApplet      = false;
        inSeparateFrame = true;
    }

    public void setUpForAppletUse() {
        inAnApplet      = true;
        inSeparateFrame = false;
    }

    /**@return True if the object has been sealed. This property will be
    set automatically by the VNCViewerOptionsAndPreferencesAssisstant having
    initialized the object to prevent an undesired second initialization.*/
    public boolean isSealed() { return sealed; }
    /**Will be invoked by VNCViewerOptionsAndPreferencesAssisstant to mark
    the object as having been initialized.*/
    public void seal() { sealed = true; }

    public VNCMode getAppMode() { return appMode; }
    public void setAppMode( String s) {
        appMode = DEFAULT_APPMODE;
        if ( s == null) return;
        if ( s.equalsIgnoreCase( SFB_ONLY)) appMode = VNCMode.FB_ONLY;
        if ( s.equalsIgnoreCase( SFBANDFT)) appMode = VNCMode.FBANDFT;
        if ( s.equalsIgnoreCase( SFT_ONLY)) appMode = VNCMode.FT_ONLY;

        if ( isFTOnly()) setBlankConsole( false);
    }

    public boolean isFTPossible() { return appMode != VNCMode.FB_ONLY; }

    public boolean isFTOnly() { return appMode == VNCMode.FT_ONLY; }

    public boolean isFBOnly() { return appMode == VNCMode.FB_ONLY; }

    public boolean isBothFBAndFT() {
        return ( appMode != VNCMode.FB_ONLY && appMode != VNCMode.FT_ONLY);
    }

    public boolean isRunAsApplet() { return inAnApplet; }

    public boolean wantsConsoleBlanked() { return blankConsole; }
    public void setBlankConsole( boolean b) {
        if ( b == blankConsole || ( b && isFTOnly())) return;
        blankConsole = b;
    }

    public boolean wants2ShowControls() { return showControls; }
    public void setShowControls( boolean b) {
        if ( b == showControls) return;
        //showControls = b;
    }

//     public boolean doesOfferRelogin() { return offerRelogin; }
//     public void setOfferRelogin( boolean b) {
//         offerRelogin = b;
//     }

//     public boolean wants2ShowOfflineDesktop() { return showOfflineDesktop; }
//     public void setShowOfflineDesktop( boolean b) {
//         showOfflineDesktop = b;
//     }

    public boolean isRunInSeparateFrame() { return inSeparateFrame; }
    public void setRunInSeparateFrame( boolean b) {
        inSeparateFrame = b;
    }

    public Dimension getCanvasDim() { return dCanvas; }
    public void setCanvasDim( Dimension d) {
        if ( d == dCanvas) return;
        dCanvas = d;
    }

    public void setCanvasDim( String sW, String sH) {
        if ( sW == null || sH == null) return;

        try {
            int w = Integer.parseInt( sW);
            int h = Integer.parseInt( sH);
            setCanvasDim( new Dimension( w, h));
        } catch ( NumberFormatException e) {
            e.printStackTrace();
            setCanvasDim( new Dimension( 0, 0));
        }
    }

    /**Convenience method for testing if a string is the same as the identifier used
     * internally to signify peer authentication is to be disabled!
     * @param s The string which is to be compared to said Identifier.
     * @return True if it is a match. False if not. Rather straight forward.*/
    public static boolean isTrustAllID( String s) { return ID_TRUST_ALL_CERTIFICATES.compareTo( s) == 0; }

    public static boolean isLocalFile( String s) { return s != null && s.startsWith( "file://"); } 
    
    /**Tests if the string is the default filename for the keystore/truststore parameter.
     * @param fn The String to test.
     * @param keystore If True, fn will be compared with the default keystore-filename
     *        or the default-truststore parameter if false.
     * @return True if they match. False if not. Again rather straight forward.*/
    public static boolean isDefault( String fn, boolean keystore) {
        String defaultFN = keystore ? DEFAULT_KEYSTORE : DEFAULT_TRUSTSTORE;
        return defaultFN.compareTo( fn) == 0;
    }

    public boolean isTrustingAllCertificates() {
        return isTrustAllID( fnTruststore);
    }

    public String getFilenameCryptstore( boolean keystore) {
        return keystore ? fnKeystore : fnTruststore;
    }

    public boolean setFilenameCryptstore( boolean keystore, String fn)
        throws MissingParameterException
    {
    	String cstype = keystore ? "Keystore" : "Truststore";

        if ( fn == null || fn.isEmpty())
            throw new MissingParameterException( "The [" + cstype + "] can't be null or empty!");

        // No changes ... No worries
        if ( fn.compareTo( getFilenameCryptstore( keystore)) == 0)
        	return true;

        setFilenameCryptstoreInternal( keystore, fn);

        return true;
    }

    private void setFilenameCryptstoreInternal( boolean keystore, String fn)
        throws MissingParameterException
    {
        if ( isTrustAllID( fn)) {
            if ( keystore)
                throw new MissingParameterException( "[" + ID_TRUST_ALL_CERTIFICATES + "] is not a valid KEYstore (only a valid TRUSTstore)!");
        } else {
        	// Not checking 4 null here as private method AND parameter checked previously!
        	if ( !fn.startsWith( "file://")) {
        		InputStream isz = this.getClass().getClassLoader().getResourceAsStream( fn);
        		if ( isz == null) throw new MissingParameterException( "The Applet doesn't contain a file named [" + fn + "]!");
        	} // Not checking if File exists.
        }

        if ( keystore) fnKeystore = fn;
        else fnTruststore = fn;
    }

    public void clearPassword() { pwdKeystore = null; pwdTruststore = null; }
    public void clearPassword( boolean keystore) { if ( keystore) pwdKeystore = null; else pwdTruststore = null; }
    public char[] getPassword( boolean keystore) { return keystore ? pwdKeystore : pwdTruststore; }
    public void setPassword( boolean keystore, char[] pwd) {
        if ( keystore) pwdKeystore = pwd;
        else pwdTruststore = pwd;
    }

    public String getFilenameRecording() { return fnRecording; }
    public boolean setFilenameRecording( String s) {
        if ( s == null || s.isEmpty()) return false;
        fnRecording = s;
        return true;
    }

    public String getSocketFactoryName() { return socketFactory; }
    public boolean setSocketFactoryName( String s) {
        if ( s == null || s.isEmpty()) return false;
        socketFactory = s;
        return true;
    }

    public int getDeferScreen() { return deferScreenUpdates; };
    public int getDeferCursor() { return deferCursorUpdates; };
    public int getDeferUpdate() { return deferUpdateRequests; };
    public void setDeferrers( int screen, int cursor, int request) {
        deferScreenUpdates  = screen;
        deferCursorUpdates  = cursor;
        deferUpdateRequests = request;
    }


    public VNCEncoding getDefaultEncoding() { return fbEncoding; }
    public void setDefaultEncoding( String s) {
        if ( s == null || s.length() == 0) return;
        VNCEncoding e = VNCEncoding.getEncoding( s, true);
        if ( e == VNCEncoding.Auto) return;
        fbEncoding = e;
    }

    public VNCSrvCfg getVNCServerConnectionParameters() { return vncServer; }
    public void setVNCServerConnectionParameters( VNCSrvCfg c) {
    	vncServer = c;
    	seal();
    }

    public LoginParameters getLogServerParameters() { return logServer; }
    public void setLogServerParameters( LoginParameters lsp) {
    	logServer = lsp;
    }

    public PixelFormat.PixelFormatQuality getPixelFormatQuality() {
        return pixelFormatQuality;
    }

    public void setPixelFormatQuality(PixelFormat.PixelFormatQuality pixelFormatQuality) {
        this.pixelFormatQuality = pixelFormatQuality;
    }

    public boolean equals(VNCViewerOptions o) {
        if ( o == null) return false;

        if ( o.appMode != appMode) return false;
        if ( o.fbEncoding != fbEncoding) return false;
        if ( o.blankConsole!= blankConsole) return false;
        if ( o.showControls != showControls) return false;
        if ( o.deferScreenUpdates != deferScreenUpdates) return false;
        if ( o.deferCursorUpdates != deferCursorUpdates) return false;
        if ( o.deferUpdateRequests != deferUpdateRequests) return false;
        if ( o.inAnApplet != inAnApplet) return false;
        if ( o.inSeparateFrame != inSeparateFrame) return false;
//         if ( o.offerRelogin != offerRelogin) return false;
//         if ( o.showOfflineDesktop != showOfflineDesktop) return false;
        if ( o.dCanvas != dCanvas) return false;

        return true;
    }
}
