package com.wincornixdorf.vnc.global;

import javax.crypto.BadPaddingException;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.Cipher;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.DESKeySpec;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import javax.swing.JApplet;

import com.wincornixdorf.vnc.rfb.RFB;
import com.wincornixdorf.vnc.rfb.conf.LoginParameters;
import com.wincornixdorf.vnc.rfb.conf.SettingsFB;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg.ConMode;
import com.wincornixdorf.vnc.rfb.fb.PixelFormat;
import com.wincornixdorf.vnc.rfb.fb.VNCEncoding;

import java.io.ByteArrayOutputStream;
import java.net.MalformedURLException;
import java.nio.charset.Charset;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.prefs.Preferences;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;


/**
 * The primary purpose of the VNCViewerOptionsAndPreferencesAssisstant class
 * is to source out code from the VNCViewer class which threatened to clog up
 * the code. As a convenient side effect various methods can be shared between
 * the methods written to cover the two primary use cases of this class:
 * <p>
 * 1. Parsing the applications parameters (regardless if they
 * came in as command line options or as Applet parameters)
 * 2. Reading and writing user specific session preferences files.
 */
public class VNCViewerOptionsAndPreferencesAssisstant {
    private static JApplet applet = null;
    private static String[] mainArgs = null;

    private static String PREF_NODE_APP = "WNXVNCClient";
    //private static String   PREF_NODE_GUI               = "Global"; // TODO: rename Global to UI (User Interface)
    private static String PREF_NODE_FT = "FT";
    private static String PREF_NODE_FB = "FB";
    private static String PREF_NODE_CN = "CN";

    private static String PREF_SERVERTXTENCODING = "serverTxtEncoding";

    private static String PREF_AUTOZOOM = "autoZoom";
    private static String PREF_MATCHZOOMFACTORS = "matchZF";
    private static String PREF_ZF_RDESKTOP = "zfDesktop";
    private static String PREF_ZF_MOUSEPTR = "zfMousePtr";
    private static String PREF_DM_MOUSEPTR = "dmMousePtr";

    private static String PREF_CLIENT_DIR = "clientDir";
    private static String PREF_SERVER_DIR = "serverDir";
    private static String PREF_CLIENT_BOOKMARKS = "clientBookmarks";
    private static String PREF_SERVER_BOOKMARKS = "serverBookmarks";
    private static String PREF_CUSTOM_DTFORMAT = "customDTFormat";
    private static String PREF_DTFORMAT = "DTFormat";

    private static String SEP_DES = "-";
    private static String SEP_AES = "/";
    private static String SEP_DBG = ";";

    private static final byte[] DEF_AES_KEY = {(byte) -51, (byte) 123, (byte) 86, (byte) 119, (byte) 12, (byte) 69, (byte) -99, (byte) 35,
            (byte) 88, (byte) 36, (byte) 71, (byte) -43, (byte) -85, (byte) 5, (byte) 37, (byte) 121};

    private static final byte[] DEF_DES_KEY = {(byte) 23, (byte) 82, (byte) 107, (byte) 6, (byte) 35, (byte) 78, (byte) 88, (byte) 7};

    private static String YOKEL = "To the yokel trying to code here: ";

    /**
     * Fills the {@link VNCViewerOptions} with values parsed from the args,
     * unless the VNCViewerOptions object has been "sealed". Once the method is
     * done, it will mark the VNCViewerOptions object as sealed to prevent the
     * object from being initialized again using one of the initialize methods.
     * This behavior is necessary to cover up an implementation flaw.
     *
     * @param o    {@link VNCViewerOptions} The container to place discovered CMDLN parameters into.
     * @param args The actual CMDLN arguments as passed into the main method.
     * @throws MissingParameterException if vital parameters haven't been found in the options.
     */
    public static void initialize(VNCViewerOptions o, String[] args)
            throws MissingParameterException {
        if (o == null || o.isSealed()) return;
        VNCViewer.pen.log(Level.FINEST, "The Application is being initialized!");

        mainArgs = args;
        o.setUpForStandaloneApplicationUse();
        parseArguments(o);
    }

    /**
     * Fill the {@link VNCViewerOptions} with values parsed from the list
     * of parameters passed into the Applet \a app, unless the VNCViewerOptions
     * object has been "sealed". Once the method is done, it will mark the
     * VNCViewerOptions object as sealed to prevent the object from being
     * initialized again using one of the initialize methods. This behavior is
     * necessary to cover up an implementation flaw.
     *
     * @param o   {@link VNCViewerOptions} The container to place discovered Applet parameters into.
     * @param app The {@link JApplet} itself.
     * @throws MissingParameterException if vital parameters haven't been found in the Applet's options.
     */
    public static void initialize(VNCViewerOptions o, JApplet app)
            throws MissingParameterException {
        if (o == null || o.isSealed() || app == null) return;
        VNCViewer.pen.log(Level.FINEST, "The Applet is being initialized!");

        applet = app;
        o.setUpForAppletUse();
        parseArguments(o);
    }

    /**
     * Read parameters from the Applet's tag or from the command line. On the command line, the
     * arguments are just a sequence of param_name/param_value pairs where the names and values
     * correspond to those expected in the Applet tag.
     *
     * @param The {@link VNCViewerOptions} to parse.
     * @throws MissingParameterException if vital parameters haven't been found in the options.
     */
    private static void parseArguments(VNCViewerOptions o)
            throws MissingParameterException {
        // REPMSG will be sent to the repeater instead of the desired destination.
        // An empty String indicates the default behavior (Sending the Version) is desired!!!
        // A RepMsg beginning with a "$" will be treated as a filename the actual
        // message is then to be read from. This is a debug feature.
        String repmsg = readParameter("REPMSG", "", false, false);

        String logHost = readParameter("LOGHOST", "", false, false);
        String logUser = readParameter("LOGUSER", "", false, false);
        String logPass = readParameter("LOGAUTH", "", false, false);
        String logPath = readParameter("LOGPATH", "", false, false);

        // Read "ENCPASSWORD/PASSWORD" parameter.
        String thePassword = readPasswordParameters();

        // The inSeparateFrame will only be honored if running as an Applet.
        o.setRunInSeparateFrame(o.isRunAsApplet() ? readBooleanParameter("OpenNewWindow", false) : true);

        o.setDefaultEncoding(readParameter("ENCODING", VNCEncoding.DEFAULT_ENC.toString(), false, false));
        o.setBlankConsole(readBooleanParameter("BLANKCONSOLE", false));
        o.setAppMode(readParameter("APPMODE", null, false, false));
        o.setShowControls(readBooleanParameter("SHOWCONTROLS", true));

        o.setFilenameRecording(readParameter("RECORDTO", "", false, false));

        // The canvas dimension parameters are read but never used.
        o.setCanvasDim(readParameter("width", "800", false, false),
                readParameter("height", "600", false, false));

        // Fine tuning sessions.options. At the moment they are unused.
        o.setDeferrers(readIntParameter("DeferScreenUpdates", 20),
                readIntParameter("DeferCursorUpdates", 10),
                readIntParameter("DeferUpdateRequests", 50));

        // The socket factory parameter is read too, but never used.
        o.setSocketFactoryName(readParameter("SocketFactory", "", false, false));

        // Tea party in Wonderland "mkoijnbhuzgv"
        o.setPassword(true, readParameter("pwdKS", "v3tt3d", false, true).toCharArray());
        o.setPassword(false, readParameter("pwdTS", "yxcvbn", false, true).toCharArray());

        o.setPixelFormatQuality(PixelFormat.PixelFormatQuality.getTypeByCode(readIntParameter("PIXELFORMAT", PixelFormat.PixelFormatQuality.EMPTY.getCode())));


        VNCSrvCfg vncServer = null;
        String url = readParameter("URL", null, false, true);
        if (url != null) {
            try {
                vncServer = new VNCSrvCfg(url, repmsg, thePassword);
            } catch (MalformedURLException e) {
                e.printStackTrace();
                throw new MissingParameterException("Failed to parse the URL.", e);
            }

            setKeystoreFileNames(o, vncServer);
            o.setLogServerParameters(new LoginParameters(true, logHost, logUser, logPass, logPath, vncServer.isRemoteSupport()));

            String destip = readParameter("DESTIP", "", false, false);
            if (!StringUtils.isEmpty(destip) && !destip.equalsIgnoreCase("null")) {
                vncServer.setDestIP(destip);
            }

            o.setVNCServerConnectionParameters(vncServer);
            return;
        }

        // A stern reminder:
        // =================
        // The HOSTNAME means nothing to the viewer. It is but a label which can
        // be set by users as they see fit. The HOSTNAME is NEVER used to identify
        // the destination. The HOSTIP is!!!
        //
        // The Label was introduced to mend these woes. Turns out it made things
        // worse still. The LABEL is used in by the Web logger, which doesn't take
        // well to a "NULL" pointer. Had to learn this the hard way. :(
        String hostName = readParameter("HOSTNAME", "UNKNOWN", false, false);
        String hostAddr = readParameter("HOSTIP", null, true, false);
        String hostLabel = readParameter("LABEL", hostAddr, false, false);
        String hostNameRep = readParameter("REPEATERHOST", "", false, false);
        String username = readParameter("VNCUSER", "", false, false);
        int portNo = readIntParameter("PORT", 5270);
        int portNoRep = readIntParameter("REPEATERPORT", 5271);

        // If hostname was not set, it will default to HOSTIP or
        // LABEL, with the LABEL superseding the HOSTIP parameter.
        if (hostName == null || hostName.isEmpty())
            hostName = hostLabel != null ? hostLabel : hostAddr;

        vncServer = new VNCSrvCfg(hostAddr, portNo, hostNameRep, portNoRep, username, repmsg, thePassword);
        vncServer.setLabel(hostLabel);

        // The SSL Parameter used 2b optional ... For legacy reasons.
        boolean wantsSSL = readBooleanParameter("SSL", true);
        vncServer.setConMode(wantsSSL ? ConMode.TLS : ConMode.PLAIN);
        setKeystoreFileNames(o, vncServer);
        o.setVNCServerConnectionParameters(vncServer);
        o.setLogServerParameters(new LoginParameters(true, logHost, logUser, logPass, logPath, vncServer.isRemoteSupport()));
    }

    /**
     * Sets proper filename for keystore and truststore - depends on running mode (RC X RS)
     *
     * @param options
     * @param vncServer
     * @throws MissingParameterException
     */
    private static void setKeystoreFileNames(VNCViewerOptions options, VNCSrvCfg vncServer) throws MissingParameterException {
        // TODO: Key- and Truststore parameters should accept a URI rather than a filename!
        options.setFilenameCryptstore(true, readParameter("ks", VNCViewerOptions.DEFAULT_KEYSTORE, false, false));
        options.setFilenameCryptstore(false, readParameter("ts", vncServer.isRemoteSupport() ? VNCViewerOptions.DEFAULT_TRUSTSTORE_RS : VNCViewerOptions.DEFAULT_TRUSTSTORE, false, false));

    }

    private static String getNodeName(VNCSrvCfg cfg) {
        return cfg.getLabel() == null ? cfg.getHostParams().getOriginalHostname() : cfg.getLabel();
    }

    public static boolean persist(VNCSessionCfg cfg) {
        if (cfg == null || cfg.vncserver == null) return false;

        // Get the hostname of the server
        String hostname = getNodeName(cfg.vncserver);

        if (hostname == null) {
            VNCViewer.pen.log(Level.WARNING, "Failed to persist any preferences: Could not establish the Hostname.");
            return false;
        }

        try {
            Preferences pusr = Preferences.userRoot();
            Preferences papp = pusr.node(PREF_NODE_APP);
            //Preferences pglb = pusr.node( prefNodeGlobal);

            if (cfg.gb != null) {
                //Preferences pfgb = papp.node( PREF_NODE_GUI);
                //                pfgb.put( PREF_STEALTH_MODE, cfg.gb.getStealthMode().toString());
            }

            // For the time being this is the only setting required by both
            // FB and FT connections being persisted here.
            if (cfg.cn != null) {
                // Do NOT create the node until we know we have got something
                // to write there, or we risk loosing the users settings!
                Preferences pfcn = papp.node(PREF_NODE_CN);

                // This will save the encoding used when sending and reading Bytes
                // from / to the server on a "per host basis.
                Preferences connectionSettingsThisHost = pfcn.node(hostname);
                Charset cs = (Charset) (cfg.cn.presumedServerEncoding.getSelectedItem());
                connectionSettingsThisHost.put(PREF_SERVERTXTENCODING, cs.name());
            }

            // Visualization Settings. They are maintained on a global basis.
            if (cfg.fb != null) {
                // Do NOT create the node until we know we have got something
                // to write there, or we risk loosing the users settings!
                Preferences pfbf = papp.node(PREF_NODE_FB);

                pfbf.putBoolean(PREF_AUTOZOOM, cfg.fb.wants2AutoScale());
                pfbf.putBoolean(PREF_MATCHZOOMFACTORS, cfg.fb.wants2MatchZoomFactors());
                pfbf.putInt(PREF_ZF_RDESKTOP, cfg.fb.getScalingFactor());
                pfbf.putInt(PREF_ZF_MOUSEPTR, cfg.fb.getCursorScalingFactor());
                pfbf.put(PREF_DM_MOUSEPTR, cfg.fb.getCursorDisplayMode().toString());
            }

            if (cfg.ft != null) {
                // Do NOT create the node until we know we have got something
                // to write there, or we risk loosing the users settings!
                Preferences pftr = papp.node(PREF_NODE_FT);

                String dbg1 = doConcat(cfg.ft.getClientBookmarks());
                String dbg2 = doConcat(cfg.ft.getServerBookmarks());
                //VncViewer.pen.log( Level.CONFIG, "persist1 [" + p.ft.getPreferredClientDir() + "]");
                //VncViewer.pen.log( Level.CONFIG, "persist2 [" + p.ft.getPreferredServerDir() + "]");
                //VncViewer.pen.log( Level.CONFIG, "persist3 [" + dbg1 + "]");
                //VncViewer.pen.log( Level.CONFIG, "persist4 [" + dbg2 + "]");

                Preferences fileTransferPreferencesThisHost = pftr.node(hostname);
                fileTransferPreferencesThisHost.put(PREF_CLIENT_DIR, cfg.ft.getPreferredClientDir());
                fileTransferPreferencesThisHost.put(PREF_SERVER_DIR, cfg.ft.getPreferredServerDir());
                fileTransferPreferencesThisHost.put(PREF_CLIENT_BOOKMARKS, dbg1);
                fileTransferPreferencesThisHost.put(PREF_SERVER_BOOKMARKS, dbg2);
                fileTransferPreferencesThisHost.put(PREF_CUSTOM_DTFORMAT, cfg.ft.getCustomDTFormatString());
                fileTransferPreferencesThisHost.put(PREF_DTFORMAT, cfg.ft.getDTFormatById());
            }
        } catch (SecurityException e) {
            VNCViewer.pen.log(Level.SEVERE, "Error while persiting preferences", e);
            return false;
        }

        return true;
    }

    public static boolean restore(VNCSessionCfg cfg) {
        // TODO: This method requires a rewrite, as the API has changed significantly.
        if (cfg == null || cfg.vncserver == null) return false;

        // Get the hostname of the server
        String hostname = getNodeName(cfg.vncserver);

        if (hostname == null) {
            VNCViewer.pen.log(Level.WARNING, "Failed to restore any preferences: Could not establish the Hostname.");
            return false;
        }

        try {
            Preferences pusr = Preferences.userRoot();
            Preferences papp = pusr.node(PREF_NODE_APP);

            if (cfg.gb != null) {
                // Preferences pfgb = papp.node( PREF_NODE_GUI);
                // cfg.gb.setStealthMode( pfgb.get( PREF_STEALTH_MODE, ""));
            }

            if (cfg.cn != null) {
                // Do NOT create the node until we know we have got something
                // to write there, or we risk loosing the users settings!
                Preferences pfcn = papp.node(PREF_NODE_CN);

                // This will save the encoding used when sending and reading Bytes
                // from / to the server on a "per host basis.
                Preferences connectionSettingsThisHost = pfcn.node(hostname);
                String dbg1 = connectionSettingsThisHost.get(PREF_SERVERTXTENCODING, RFB.DEFAULT_CHARSET.name());
                String err = "Failed to restore the TXT encoding to be used with host [" +
                        hostname + "]: [" + dbg1 + "] is not a valid Charset!";
                try {
                    Charset cs = Charset.forName(dbg1);
                    cfg.cn.presumedServerEncoding.setSelectedItem(cs);
                } catch (Exception e) {
                    VNCViewer.pen.log(Level.WARNING, err, e);
                }
            }

            // Framebuffer Settings are maintained on a global basis.
            if (cfg.fb != null) {
                Preferences pfbf = papp.node(PREF_NODE_FB);

                cfg.fb.setAutoScale(pfbf.getBoolean(PREF_AUTOZOOM, false));
                cfg.fb.setMatchZoomFactors(pfbf.getBoolean(PREF_MATCHZOOMFACTORS, true));
                cfg.fb.setScalingFactor(pfbf.getInt(PREF_ZF_RDESKTOP, SettingsFB.DEFZOOMFACTOR));
                cfg.fb.setCursorScalingFactor(pfbf.getInt(PREF_ZF_MOUSEPTR, SettingsFB.DEFZOOMFACTOR));
                String cdmStr = pfbf.get(PREF_DM_MOUSEPTR, SettingsFB.CursorDisplayMode.SYSTEM.toString());
                cfg.fb.setCursorDisplayMode(SettingsFB.CursorDisplayMode.valueOf(cdmStr));
            }

            // Filetransfer Preferences are kept on a per host basis.
            if (cfg.ft != null && cfg.vncserver.getConMode() != VNCSrvCfg.ConMode.FILE) {
                Preferences pftr = papp.node(PREF_NODE_FT);
                Preferences fileTransferPreferencesThisHost = pftr.node(hostname);
                String dbg1 = fileTransferPreferencesThisHost.get(PREF_CLIENT_BOOKMARKS, "");
                String dbg2 = fileTransferPreferencesThisHost.get(PREF_SERVER_BOOKMARKS, "");

                cfg.ft.setPreferredClientDir(fileTransferPreferencesThisHost.get(PREF_CLIENT_DIR, ""));
                cfg.ft.setPreferredServerDir(fileTransferPreferencesThisHost.get(PREF_SERVER_DIR, ""));
                cfg.ft.setClientBookmarks(splitSafely(dbg1));
                cfg.ft.setServerBookmarks(splitSafely(dbg2));
                cfg.ft.setCustomDTFormatString(fileTransferPreferencesThisHost.get(PREF_CUSTOM_DTFORMAT, ""));
                cfg.ft.setDTFormat(fileTransferPreferencesThisHost.get(PREF_DTFORMAT, ""));

                //VncViewer.pen.log( Level.CONFIG, "restore1 [" + p.ft.getPreferredClientDir() + "]");
                //VncViewer.pen.log( Level.CONFIG, "restore2 [" + p.ft.getPreferredServerDir() + "]");
                //VncViewer.pen.log( Level.CONFIG, "restore3 [" + dbg1 + "]");
                //VncViewer.pen.log( Level.CONFIG, "restore4 [" + dbg2 + "]");
            }
        } catch (SecurityException e) {
            VNCViewer.pen.log(Level.SEVERE, "Error while restoring preferences", e);
            return false;
        }

        return true;
    }

    /**
     * Tries to read a parameter from either the command line or the Applet's list
     * of parameters depending on the constructor used. If the desired parameter
     * can not be found, the method returns null or throws an exception.<p>
     *
     * @param name      The name of the parameter to be looking for.
     * @param required  If true, the method will throw a {@link MissingParameterException}
     *                  in case the desired parameter could not be found.
     * @param hideValue Where true, the method will not show the parameter value to be used.
     *                  this should be used with password parameters.
     * @throws MissingParameterException
     */
    private static String readParameter(String name, String defaultValue, boolean required, boolean hideValue)
            throws MissingParameterException {
        if (name == null || name.isEmpty())
            throw new MissingParameterException(YOKEL + "YOU MAY WANT TO GIVE ME A PARAMETER 2 READ!");

        if (required && defaultValue != null)
            throw new MissingParameterException(YOKEL + "MAKE UP YOUR MIND! Is the [" + name + "] option required or NOT???!");

        String p = null;

        if (applet != null) {
            p = applet.getParameter(name);
        } else {
            int i = 0;
            int argc = mainArgs.length;

            while (i < argc && p == null) {
                if (mainArgs[i].equalsIgnoreCase(name) && i + 1 < argc)
                    p = mainArgs[i + 1];
                i += 2;
            }
        }

        if (required && (p == null))
            throw new MissingParameterException(name + " parameter not specified");

        String rval = p == null ? defaultValue : p;
        String def = p == null ? " [FALLBACK]" : "";
        String msg = "Read Param [" + name + "] = [" + (hideValue ? "You wish :)" : rval) + "]" + def;
        System.err.println(msg);
        VNCViewer.pen.log(Level.CONFIG, msg);
        return rval;
    }

    private static String readLegacyDESParameter(String encPWD)
            throws MissingParameterException {
        String fail = "Failed to recover the ENCPASSWORD: ";
        String recm = "Install a suitable JCE provider (like Bouncy Castle http://www.bouncycastle.org/).";
        String ciph = "DES";
        String mode = "CBC";
        String padd = "NoPadding";

        try {
            byte[] pw = getBytes(encPWD, SEP_DES);

            DESKeySpec desKeySpec = new DESKeySpec(DEF_DES_KEY);
            SecretKeyFactory keyFactory = SecretKeyFactory.getInstance("DES");
            SecretKey secretKey = keyFactory.generateSecret(desKeySpec);

            Cipher cipher = Cipher.getInstance(ciph + "/" + mode + "/" + padd);
            cipher.init(Cipher.DECRYPT_MODE, secretKey, new IvParameterSpec(DEF_DES_KEY));

            return new String(cipher.doFinal(pw));
        } catch (InvalidKeyException e) {
            throw new MissingParameterException(fail + "The key is invalid! (Internal error)", e);
        } catch (NoSuchAlgorithmException e) {
            throw new MissingParameterException(fail + "Couldn't find a provider implementing the [" + ciph + "] cipher. " + recm, e);
        } catch (InvalidKeySpecException e) {
            throw new MissingParameterException(fail + "The key specification is faulty! (Internal error)", e);
        } catch (NoSuchPaddingException e) {
            throw new MissingParameterException(fail + "Couldn't find a privider implementing the [" + padd + "] padding. " + recm, e);
        } catch (InvalidAlgorithmParameterException e) {
            throw new MissingParameterException(fail + "The selected cipher doesn't support an IV parameter! (Internal error)", e);
        } catch (IllegalBlockSizeException e) {
            throw new MissingParameterException(fail + "Internal error while decrypting encoded password. Contact vendor.");
        } catch (BadPaddingException e) {
            throw new MissingParameterException(fail + "The password hasn't been padded as expected. Contact vendor.");
        }

    }


    private static String readAESParameter(String encPWD)
            throws MissingParameterException {
        int ivLen = 16;
        byte[] pw = getBytes(encPWD, SEP_AES);
        byte[] iv = new byte[ivLen];
        byte[] payload = new byte[pw.length - ivLen];
        String iname = "AES/CFB/PKCS5Padding";

        // 1st we need to take the IV off the crypt text
        System.arraycopy(pw, 0, iv, 0, ivLen);

        // then we copy the bytes containing the encrypted password into
        // a buffer from where we can decrypt it using the password and the IV.
        System.arraycopy(pw, ivLen, payload, 0, payload.length);

        SecretKeySpec keySpec = new SecretKeySpec(DEF_AES_KEY, "AES");
        IvParameterSpec ivSpec = new IvParameterSpec(iv);

        Cipher cipher;
        try {
            cipher = Cipher.getInstance(iname);
            cipher.init(Cipher.DECRYPT_MODE, keySpec, ivSpec);
            return new String(cipher.doFinal(payload));
        } catch (Exception e) {
            throw new MissingParameterException("Failed to read the (supposedly) AES encrypted Password.", e);
        }
    }

    /**
     * Read password parameters. If an "ENCPASSWORD" parameter is set, this method will read and
     * try to decrypt it. If not, it will look out for the "PASSWORD" parameter. Eventually the
     * cleartext password will be written to the appropriate sink, providing a password was passed
     * in via one of the two interfaces. If not, the sink will remain untouched.
     *
     * @throws MissingParameterException If neither a "PASSWORD" nor a "ENCPASSWORD" parameter was found.
     */
    private static String readPasswordParameters()
            throws MissingParameterException {
        String encPWD = readParameter("ENCPASSWORD", null, false, true);

        if (encPWD == null || encPWD.isEmpty()) {
            return readParameter("PASSWORD", null, false, true);
        }

        if (isASCIIArmouredString(encPWD, SEP_DES))
            return readLegacyDESParameter(encPWD);
        else return readAESParameter(encPWD);
    }

    private static boolean isASCIIArmouredString(String s, String sep) {
        if (s == null) return false;

        String reseg = "[\\d]{1,3}";
        String relst = reseg + "(\\" + sep + reseg + ")*";

        Pattern p = Pattern.compile(relst);
        Matcher m = p.matcher(s);
        return m.matches();
    }

    private static byte[] getBytes(String str, String sep) {
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        StringTokenizer st = new StringTokenizer(str, sep, false);
        while (st.hasMoreTokens()) {
            int i = Integer.parseInt(st.nextToken());
            bos.write((byte) i);
        }
        return bos.toByteArray();
    }

    private static int readIntParameter(String name, int defaultValue)
            throws MissingParameterException {
        return parseInt(readParameter(name, null, false, false), defaultValue);
    }

    private static int parseInt(String s, int defaultValue) {
        int result = defaultValue;
        if (s != null) {
            try {
                result = Integer.parseInt(s);
            } catch (NumberFormatException e) {
            }
        }
        return result;
    }

    private static boolean readBooleanParameter(String name, boolean defaultValue)
            throws MissingParameterException {
        return parseBool(readParameter(name, null, false, false), defaultValue);
    }

    private static boolean parseBool(String s, boolean defaultValue) {
        if (s == null) return defaultValue;

        return (s.equalsIgnoreCase("1") || s.equalsIgnoreCase("Yes") ||
                s.equalsIgnoreCase("Y") || s.equalsIgnoreCase("True") ||
                s.equalsIgnoreCase("On"));
    }

    private static String doConcat(String[] arr) {
        String s = "";
        if (arr == null || arr.length == 0) return s;

        for (int i = 0; i < arr.length; i++) s += arr[i] + SEP_DBG;

        return s.substring(0, s.length() - 1);
    }

    private static String[] splitSafely(String s) {
        if (s == null) s = "";
        return s.split(SEP_DBG);
    }
}