package com.wincornixdorf.vnc.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;

import com.wincornixdorf.vnc.global.VNCViewerOptions;

public class JAboutDialog extends JDialog implements ActionListener {
    private static final long serialVersionUID = 1888314916916979063L;

    private static final    int     clOptTabLen = 20;

    private                 JButton jbtnOk      = null;
    private                 String  clOpts      = null;

    public JAboutDialog( VNCViewerOptions clOptions)
    {
        setTitle( "About");
        
        setCLOptionsDebugStr( clOptions);

        setLayout( new BorderLayout( 3, 3));

        JTabbedPane jtp = new JTabbedPane();
        jtp.setPreferredSize( new Dimension( 550, 350));

        JTextArea jtaCopyrightNotice = createDisplay( getContentsOfJarFile( "DISCLAIMER.TXT"));
        JTextArea jtaGPL = createDisplay( getContentsOfJarFile( "LICENCE.TXT"));
        JTextArea jtaChangeLog = createDisplay( getContentsOfJarFile( "CHANGELOG.TXT"));
        JTextArea jtaCmdLineOptions = createDisplay( clOpts);

        jtp.add( "General",         new MSLogo());
        jtp.add( "Disclaimer",      new JScrollPane( jtaCopyrightNotice));
        jtp.add( "GPL",             new JScrollPane( jtaGPL));
        jtp.add( "Changelog",       new JScrollPane( jtaChangeLog));
        jtp.add( "CMDLN Options",   new JScrollPane( jtaCmdLineOptions));

        jtp.setMnemonicAt( 0, KeyEvent.VK_G);
        jtp.setMnemonicAt( 1, KeyEvent.VK_D);
        jtp.setMnemonicAt( 2, KeyEvent.VK_L);
        jtp.setMnemonicAt( 3, KeyEvent.VK_C);
        jtp.setMnemonicAt( 4, KeyEvent.VK_O);

        FlowLayout lojpButton = new FlowLayout( FlowLayout.RIGHT, 3, 3);
        JPanel jpButton = new JPanel( lojpButton);
        jbtnOk = new JButton( "Ok");
        jbtnOk.setMnemonic( KeyEvent.VK_O);
        jpButton.add( jbtnOk);
        jbtnOk.addActionListener( this);

        getContentPane().add( jtp, BorderLayout.CENTER);
        getContentPane().add( jpButton, BorderLayout.SOUTH);
        pack();
    }

    public void actionPerformed( ActionEvent evt) {
        if ( evt.getSource() == jbtnOk) setVisible( false);
    }

    private JTextArea createDisplay( String s) {
        JTextArea jta = new JTextArea();
        jta.setEditable( false);
        jta.setText( s);
        jta.setCaretPosition( 0);
        jta.setFont( new Font( "Monospaced", jta.getFont().getStyle(), 14));
        return jta;
    }

    private void setCLOptionsDebugStr( VNCViewerOptions clOptions) {
        if ( clOptions == null) return;

        StringBuffer        sb  = new StringBuffer();
        Dimension           cd  = clOptions.getCanvasDim();

        sb.append( stuff( "Pseudo URL"));
        sb.append( clOptions.getVNCServerConnectionParameters().getPseudoURL());
        sb.append( '\n');

        sb.append( stuff( "App. mode"));
        sb.append( clOptions.getAppMode().toString());
        sb.append( '\n');

        sb.append( stuff( "Default Encoding"));
        sb.append( clOptions.getDefaultEncoding().toString());
        sb.append( '\n');

        sb.append( stuff( "Canvas Dimension"));
        sb.append( cd.width + "x" + cd.height);
        sb.append( '\n');

        sb.append( stuff( "Blank Console"));
        sb.append( clOptions.wantsConsoleBlanked());
        sb.append( '\n');

        sb.append( stuff( "Show Controls"));
        sb.append( clOptions.wants2ShowControls());
        sb.append( '\n');

        sb.append( stuff( "REPMSG"));
        sb.append( clOptions.getVNCServerConnectionParameters().getRepeaterMsg());
        sb.append( '\n');
        
        clOpts = sb.toString();
    }

    /**Returns a String padded with whitespaces to a length of clOptTabLen.*/
    private String stuff( String s) {
        StringBuffer sb = new StringBuffer( s);
        while ( sb.length() < clOptTabLen) sb.append( ' ');
        sb.append( ':');
        return sb.toString();
    }

    private String getContentsOfJarFile( String fnInJar) {
        String s = "";

        try {
            InputStream isz = this.getClass().getClassLoader().getResourceAsStream( fnInJar);
            if ( isz != null) {
                BufferedReader br = new BufferedReader( new InputStreamReader( isz));
                while ( br.ready()) s += br.readLine() + "\n";
            } else s = "Could not find resource [" + fnInJar + "]. Please notify vendor.";
        } catch ( IOException e) {
            s = "IOException while reading [" + fnInJar + "]. Please notify vendor.";
        }

        return s;
    }
}

