package com.wincornixdorf.vnc.gui;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.nio.charset.Charset;
import java.util.Observable;
import java.util.Observer;
import java.util.logging.Level;

import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerListModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.VNCSession;
import com.wincornixdorf.vnc.rfb.conf.SettingsCN;
import com.wincornixdorf.vnc.rfb.conf.SettingsFB;
import com.wincornixdorf.vnc.rfb.conf.SettingsFB.CursorDisplayMode;
import com.wincornixdorf.vnc.rfb.conf.SettingsFT;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.rfb.fb.PixelFormat;
import com.wincornixdorf.vnc.rfb.fb.VNCEncoding;

public class JOptionsWidget extends JPanel
    implements ItemListener, ActionListener, ChangeListener, Observer, MouseListener {
    
    private static final long   serialVersionUID = -2501188935286175163L;
    private static final String jlTxtEncodingCap = "Text-Encoding";

    private static final double WEIGHTX_DEF      = 0.75;
    private static final double WEIGHTX_CONTROLS = 2.5;
    private static final double WEIGHTX_SLIDER   = 4.0;
    
    private SettingsCN          			cnPrefs             = null;
    private SettingsFB          			fbPrefs             = null;
    private SettingsFT          			ftPrefs             = null;

    // Connection related controls
    private JComboBox<VNCEncoding>  		jcbPixelEncoding    = null;

    private JComboBox<PixelFormat>			jcbPixFormat        = null;
    private JComboBox<CursorDisplayMode>	jcbCursorDM         = null;
    
    // The Text-Encoding is special in that it is displayed both on the FT&FB tab!
    // Hence the Label gets created only once and added accordingly!
    private JLabel              			jlTxtEncoding       = null;
    private JComboBox<Charset>  			jcbTxtEncoding      = null;
    
    private JSpinner            			jspCompQuality      = null;
    private JSpinner            			jspJPEGQuality      = null;

    // Display related controls
    private JCheckBox           			jckAutoScale        = null;
    private JCheckBox           			jckMatchZoomFactors = null;
    private JSlider             			jslScaleFactorDT    = null;
    private JSlider             			jslScaleFactorRC    = null;
    private JTextField          			tfSocketTO          = null;

    private JComboBox<String>   			jcbDTFormat         = null;
    private JTextField          			jtfDTFormat         = null;

    private GridBagLayout       			lo					= new GridBagLayout();
    private Insets              			dins                = new Insets( 2, 4, 2, 4);

    private boolean             			isFB                = true;
    
    public JOptionsWidget(boolean isFBIn) {
    	super();
    	setLayout( lo);

        isFB = isFBIn;

        jlTxtEncoding    = new JLabel( jlTxtEncodingCap);
        jcbTxtEncoding   = new JComboBox<Charset>();
        jlTxtEncoding.setLabelFor( jcbTxtEncoding);
        jlTxtEncoding.setDisplayedMnemonic( KeyEvent.VK_X);
        
        if ( isFB) buildConnectionOptionsPanel();
        else buildFileTransferOptionsPanel();
    }

    private GridBagConstraints gbc( int gridx, int gridy, double weightx) {
    	GridBagConstraints  c = new GridBagConstraints( gridx, gridy, 1, 1, weightx, 1.0,
        GridBagConstraints.CENTER,
        GridBagConstraints.HORIZONTAL, dins, 0, 0);
    	return c;
    }

    private void buildConnectionOptionsPanel() {
//        GridBagLayout       gb  = new GridBagLayout();
//        JPanel              jp  = new JPanel( gb);
        GridBagConstraints  c   = null;

        JLabel jlPixelEncoding  = new JLabel( "Pixel-Encoding");
        jcbPixelEncoding        = new JComboBox<VNCEncoding>();
        jlPixelEncoding.setLabelFor( jcbPixelEncoding);
        jlPixelEncoding.setDisplayedMnemonic( KeyEvent.VK_E);
                
        JLabel jlPixFormat       = new JLabel( "PixelFormat");
        jcbPixFormat             = new JComboBox<PixelFormat>();
        jlPixFormat.setLabelFor( jcbPixFormat);
        jlPixFormat.setDisplayedMnemonic( KeyEvent.VK_P);

        jcbPixFormat.addItem(PixelFormat.PixelFormatQuality.LOW.getFormat());
        jcbPixFormat.addItem(PixelFormat.PixelFormatQuality.MEDIUM.getFormat());
        jcbPixFormat.addItem(PixelFormat.PixelFormatQuality.HIGH.getFormat());

        JLabel jlCompQ          = new JLabel( "Compression Quality");
        jspCompQuality          = new JSpinner( new SpinnerListModel( SettingsCN.getAvailableCompQualityLevels()));
        jlCompQ.setLabelFor( jspCompQuality);
        jlCompQ.setDisplayedMnemonic( KeyEvent.VK_Q);

        JLabel jlJPEGQ          = new JLabel( "JPEG Quality");
        jspJPEGQuality          = new JSpinner( new SpinnerListModel( SettingsCN.getAvailableJPEGQualityLevels()));
        jlJPEGQ.setLabelFor( jspJPEGQuality);
        jlJPEGQ.setDisplayedMnemonic( KeyEvent.VK_J);

        JLabel jlFBSF           = new JLabel( "Zoom factor Desktop");
        jslScaleFactorDT        = new JSlider( SettingsFB.MINZOOMFACTOR, SettingsFB.MAXZOOMFACTOR, SettingsFB.DEFZOOMFACTOR);
        jlFBSF.setLabelFor( jslScaleFactorDT);
        jlFBSF.setDisplayedMnemonic( KeyEvent.VK_F);

        JLabel jlRCSF           = new JLabel( "Zoom factor Mouse pointer");
        jslScaleFactorRC        = new JSlider( SettingsFB.MINZOOMFACTOR, SettingsFB.MAXZOOMFACTOR, SettingsFB.DEFZOOMFACTOR);
        jlRCSF.setLabelFor( jslScaleFactorRC);
        jlRCSF.setDisplayedMnemonic( KeyEvent.VK_C);

        JLabel jlPMCursor       = new JLabel( "Mouse Paint Mode");
        jcbCursorDM             = new JComboBox<CursorDisplayMode>();
        jcbCursorDM.addItem( CursorDisplayMode.SYSTEM);
        jcbCursorDM.addItem( CursorDisplayMode.PAINT_N_SPARE_SYSTEM);
        jcbCursorDM.addItem( CursorDisplayMode.PAINT_N_HIDE_SYSTEM);
        jcbCursorDM.addItem( CursorDisplayMode.PAINT_N_UPDATE_SYSTEM);
        jlPMCursor.setLabelFor( jcbCursorDM);
        jlPMCursor.setDisplayedMnemonic( KeyEvent.VK_U);
        
        //jckReverseMouseBtns     = new JCheckBox( "Reverse Mouse Buttons");
        jckAutoScale            = new JCheckBox( "AutoScale");
        jckMatchZoomFactors     = new JCheckBox( "Match zoom factors");
        //jckReverseMouseBtns.setMnemonic( KeyEvent.VK_R);
        jckAutoScale.setMnemonic( KeyEvent.VK_A);

        fineTune( jslScaleFactorDT, 50, 1);
        fineTune( jslScaleFactorRC, 50, 1);
        
//        JLabel jlSocketTO       = null;

//        if ( VNCViewer.isDivine()) {
//            jlSocketTO              = new JLabel( "Socket timeout");
//            tfSocketTO              = new JTextField();
//            tfSocketTO.setEnabled( false);
//        }

        // Pixel-Encoding selector
        c = gbc( 0, 0, WEIGHTX_DEF);
        add( jlPixelEncoding, c);

        c = gbc( 1, 0, WEIGHTX_CONTROLS);
        add( jcbPixelEncoding, c);
        
        // PixelFormat
        c = gbc( 2, 0, WEIGHTX_DEF);
        add( jlPixFormat, c);

        c = gbc( 3, 0, WEIGHTX_CONTROLS);
        add( jcbPixFormat, c);

        // Compression Quality
        c = gbc( 0, 1, WEIGHTX_DEF);
        add( jlCompQ, c);

        c = gbc( 1, 1, WEIGHTX_CONTROLS);
        add( jspCompQuality, c);

        // JPEG Quality
        c = gbc( 2, 1, WEIGHTX_DEF);
        add( jlJPEGQ, c);

        c = gbc( 3, 1, WEIGHTX_CONTROLS);
        add( jspJPEGQuality, c);

        // Paint Mode Mouse Cursor 
        c = gbc( 0, 2, WEIGHTX_DEF);
        add( jlPMCursor, c);
        
        c = gbc( 1, 2, WEIGHTX_CONTROLS);
        add( jcbCursorDM, c);
        
        // Text-Encoding selector
        c = gbc( 2, 2, WEIGHTX_DEF);
        add( jlTxtEncoding, c);

        c = gbc( 3, 2, WEIGHTX_CONTROLS);
        add( jcbTxtEncoding, c);
        
//        if ( tfSocketTO != null) {
//            c = new GridBagConstraints( 0, 5, 1, 1, 1.0, 1.0,
//            GridBagConstraints.NORTH,
//            GridBagConstraints.HORIZONTAL, dins, 0, 0);
//            lo.setConstraints( jlSocketTO, c);
//            add( jlSocketTO);
//
//            c = new GridBagConstraints( 1, 5, 1, 1, 1.0, 1.0,
//            GridBagConstraints.NORTH,
//            GridBagConstraints.HORIZONTAL, dins, 0, 0);
//            lo.setConstraints( tfSocketTO, c);
//            add( tfSocketTO);
//        }

        // FB Scale Factor Slider
        c = gbc( 4, 0, WEIGHTX_DEF);
        add( jlFBSF, c);

        c = gbc( 5, 0, WEIGHTX_SLIDER);
        c.weightx=3.0;
        add( jslScaleFactorDT, c);

        c = gbc( 6, 0, WEIGHTX_CONTROLS);
        add( jckMatchZoomFactors, c);


        // RC Scale Factor Slider
        c = gbc( 4, 2, WEIGHTX_DEF);
        add( jlRCSF, c);

        c = gbc( 5, 2, WEIGHTX_SLIDER);
        add( jslScaleFactorRC, c);

        c = gbc( 6, 2, WEIGHTX_CONTROLS);
        add( jckAutoScale, c);

        jckAutoScale.addItemListener( this);
        jckMatchZoomFactors.addItemListener( this);
        jslScaleFactorDT.addChangeListener( this);
        jslScaleFactorRC.addChangeListener( this);
        
        jcbPixelEncoding.addActionListener( this);
        jcbPixFormat.addActionListener( this);
        jcbCursorDM.addActionListener( this);
        jspCompQuality.addChangeListener( this);
        jspJPEGQuality.addChangeListener( this);
    }

    private void buildFileTransferOptionsPanel() {
//        GridBagLayout       gb  = new GridBagLayout();
//        JPanel              jp  = new JPanel( gb);
        GridBagConstraints  c   = null;

        JLabel jlDTFormat = new JLabel( "Date/Time mode");
        jcbDTFormat = new JComboBox<String>();
        jlDTFormat.setLabelFor( jcbDTFormat);
        jlDTFormat.setDisplayedMnemonic( KeyEvent.VK_D);

        jcbDTFormat.addItem( SettingsFT.sDTCustom);
        jcbDTFormat.addItem( SettingsFT.sDTSHORT);
        jcbDTFormat.addItem( SettingsFT.sDTMEDIUM);
        jcbDTFormat.addItem( SettingsFT.sDTLONG);
        jcbDTFormat.addItem( SettingsFT.sDTFULL);

        JLabel jlDTCustFormat = new JLabel( "DT custom format");
        jtfDTFormat = new JTextField( "");
        jtfDTFormat.setEnabled( false);

        // Date-/Time-Format
        c = gbc( 0, 0, WEIGHTX_DEF);
        add( jlDTFormat, c);

        c = gbc( 1, 0, WEIGHTX_CONTROLS);
        add( jcbDTFormat, c);

        // Date-/Time-Format (Custom string)
        c = gbc( 2, 0, WEIGHTX_DEF);
        add( jlDTCustFormat, c);

        c = gbc( 3, 0, WEIGHTX_CONTROLS);
        add( jtfDTFormat, c);
        
        // Text-Encoding selector
        c = gbc( 4, 0, WEIGHTX_DEF);
        add( jlTxtEncoding, c);
        
        c = gbc( 5, 0, WEIGHTX_CONTROLS);
        add( jcbTxtEncoding, c);
        
        jcbDTFormat.addActionListener( this);
        jtfDTFormat.addActionListener( this);
    }

    /**Helps me save a few lines of code.*/
    private void fineTune( JSlider jsl, int maj, int min) {
        jsl.setMajorTickSpacing( maj);
        jsl.setMinorTickSpacing( min);
        jsl.setPaintLabels( true);
        jsl.setPaintTicks( true);
        jsl.setPaintTrack( true);
        jsl.addMouseListener( this);
    }

    public void setOptions( VNCSessionCfg cfg) {
        if ( cfg == null) return;
        setOptions( cfg.cn);
        setOptions( cfg.fb);
        setOptions( cfg.ft);
    }

    private void setOptions( SettingsCN o) {
        // i will not have a null value for my cnPrefs object
        // nor may the widget be reused with another cnPrefs object.
        if ( o == null || cnPrefs != null) return;

        cnPrefs = o;
        update( cnPrefs, SettingsCN.VNCProperty.PIXENCODING); // 2nd Parameter ensures proper update!
        cnPrefs.addObserver( this);
    }

    private void setOptions( SettingsFB o) {
        // I will not have a null value for my fbPrefs object
        // and once set, the object will not be replaced
        if ( o == null || fbPrefs != null) return;

        fbPrefs = o;
        update( fbPrefs, SettingsFB.FBProperty.SHOWBUTTONS); // 2nd Parameter ensures proper update!
        fbPrefs.addObserver( this);
    }

    private void setOptions( SettingsFT p) {
        // I will not have a null value for my ftPrefs object
        // and once set, the object will not be replaced
        if ( p == null || ftPrefs != null) return;

        ftPrefs = p;
        update( ftPrefs, SettingsFT.FTProperty.DTFORMAT); // 2nd Parameter ensures proper update!
        ftPrefs.addObserver( this);
    }

//    public void setOptions( SettingsCN o, VNCSessionCfg sp) {
//        setOptions( o);
//        if ( sp != null) {
//            setOptions( sp.fb);
//            setOptions( sp.ft);
//        }
//    }

    public void itemStateChanged( ItemEvent e) {
        if ( fbPrefs != null) {
            if ( e.getSource() == jckAutoScale)
                fbPrefs.setAutoScale( jckAutoScale.isSelected());

            if ( e.getSource() == jckMatchZoomFactors)
                fbPrefs.setMatchZoomFactors( jckMatchZoomFactors.isSelected());
        }

        verifyControls();
    }

    public void actionPerformed( ActionEvent e) {
        if ( e.getID() != ActionEvent.ACTION_PERFORMED) return;

        if ( cnPrefs != null) {
            if ( e.getSource() == jcbPixelEncoding)
                cnPrefs.setPreferredEncoding( (VNCEncoding)jcbPixelEncoding.getSelectedItem());

            if ( e.getSource() == jcbPixFormat)
                cnPrefs.setPixFormat( (PixelFormat)jcbPixFormat.getSelectedItem());
        }

        if ( fbPrefs != null) {
            if ( e.getSource() == jcbCursorDM)
                fbPrefs.setCursorDisplayMode( (CursorDisplayMode)jcbCursorDM.getSelectedItem());
        }

        if ( ftPrefs != null) {
            if ( e.getSource() == jcbDTFormat)
                ftPrefs.setDTFormat( (String)jcbDTFormat.getSelectedItem());

            if ( e.getSource() == jtfDTFormat)
                ftPrefs.setCustomDTFormatString( (String)jtfDTFormat.getText());
        }

        verifyControls();
    }

    public void stateChanged( ChangeEvent e) {
        if ( cnPrefs != null) {
            if ( e.getSource() == jspCompQuality)
                cnPrefs.setCompressLevel( ((Integer)jspCompQuality.getValue()).intValue());

            if ( e.getSource() == jspJPEGQuality)
                cnPrefs.setJpegQuality( ((Integer)jspJPEGQuality.getValue()).intValue());
        }

        if ( fbPrefs != null) {
            if ( e.getSource() == jslScaleFactorDT)
                fbPrefs.setScalingFactor( jslScaleFactorDT.getValue());

            if ( e.getSource() == jslScaleFactorRC)
                fbPrefs.setCursorScalingFactor( jslScaleFactorRC.getValue());           
        }

        verifyControls();
    }

    public void mouseClicked( MouseEvent e) {
        // In the event of a double click on any of the scaleFactor
        // sliders, we home in on 100%
        if ( e.getSource() == jslScaleFactorDT && e.getClickCount() == 2)
            jslScaleFactorDT.setValue( 100);

        if ( e.getSource() == jslScaleFactorRC && e.getClickCount() == 2)
            jslScaleFactorRC.setValue( 100);
    }

    public void mouseEntered( MouseEvent e) {}
    public void mouseExited( MouseEvent e) {}
    public void mousePressed( MouseEvent e) {}
    public void mouseReleased( MouseEvent e) {}

    private void verifyControls() {
        if ( isFB && cnPrefs != null) {
            VNCEncoding encoding = cnPrefs.getPreferredEncoding();
            jspCompQuality.setEnabled( encoding == VNCEncoding.Zlib || encoding == VNCEncoding.Tight);
            jspJPEGQuality.setEnabled( !cnPrefs.wantsEightBitColors());
        }

        if ( isFB && fbPrefs != null) {
            jslScaleFactorDT.setEnabled( !fbPrefs.wants2AutoScale());
            jslScaleFactorRC.setEnabled( !fbPrefs.wants2MatchZoomFactors() &&
            		                     !fbPrefs.wants2AutoScale() &&
            		                      fbPrefs.getCursorDisplayMode() != CursorDisplayMode.SYSTEM);
        }

        if ( !isFB && ftPrefs != null) {
            jtfDTFormat.setEnabled( ftPrefs.usesCustomFormat());
        }
    }

    public void update( Observable o, Object arg) {

        // The servers presumed encoding is displayed on both the FB & FT tabs!
        if ( cnPrefs != null)
            jcbTxtEncoding.setModel( cnPrefs.presumedServerEncoding);

        if ( isFB && cnPrefs != null && arg instanceof SettingsCN.VNCProperty) {
            jcbPixelEncoding.setModel( cnPrefs.preferredEncodings);

            jcbPixelEncoding.setSelectedItem( cnPrefs.getPreferredEncoding());
            jcbPixFormat.setSelectedItem( cnPrefs.getPixelFormatQuality().getFormat());

            jspCompQuality.setValue( cnPrefs.getCompressLevel());
            jspJPEGQuality.setValue( cnPrefs.getJpegQuality());

            if ( tfSocketTO != null) tfSocketTO.setText( "" + cnPrefs.getSocketTimeout());
            //jckReverseMouseBtns.setSelected( cnPrefs.wantsReversedMouseButtons());
        }

        if ( isFB && fbPrefs != null && arg instanceof SettingsFB.FBProperty) {
            // Hmmm ... IllegalArgumentException: invalid sequence element???!
            //      jspCompQuality.setValue( new Integer( cnPrefs.getCompressLevel()));
            //      jspJPEGQuality.setValue( new Integer( cnPrefs.getJpegQuality()));
            jslScaleFactorDT.setValue( fbPrefs.getScalingFactor());
            jslScaleFactorRC.setValue( fbPrefs.getCursorScalingFactor());
            
            jckMatchZoomFactors.setSelected( fbPrefs.wants2MatchZoomFactors());
            jckAutoScale.setSelected( fbPrefs.wants2AutoScale());
            
            CursorDisplayMode cdm = fbPrefs.getCursorDisplayMode();
            jcbCursorDM.setSelectedItem( cdm);
            jcbCursorDM.setToolTipText( cdm.getToolTip());
        }

        if ( !isFB && ftPrefs != null && arg instanceof SettingsFT.FTProperty) {
            jcbDTFormat.setSelectedItem( ftPrefs.getDTFormatById());
            jtfDTFormat.setText( ftPrefs.getCustomDTFormatString());
        }

        verifyControls();
    }
}
