package com.wincornixdorf.vnc.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;

public class SharedGraphicsUtils {

    public static void printDoubleBold( Graphics g, String msg, Rectangle2D r,
            Font f, Color cBack, Color cFront)
    {
        Font fMsg = scale2Fit( g, msg, r, g.getFont());

        g.setColor( cBack);
        drawStringInto( g, msg, r, fMsg, false, Component.CENTER_ALIGNMENT, Component.CENTER_ALIGNMENT);
        Rectangle rh = r.getBounds();
        rh.translate( 2, 2);
        g.setColor( cFront);
        drawStringInto( g, msg, rh, fMsg, false, Component.CENTER_ALIGNMENT, Component.CENTER_ALIGNMENT);
	}

	
    // Very basic. Still under construction.
    public static void drawStringInto( Graphics g, String s, Rectangle2D r, 
    								Font f, boolean scaleFont, float alignX, float alignY)
    {
        if ( g == null || s == null || r == null) return;

        Font fOld = g.getFont();
        if ( f == null) f = fOld;

        g.setFont( scaleFont ? scale2Fit( g, s, r, f) : f);

        int asc = g.getFontMetrics().getAscent();
        Rectangle2D rStr     = g.getFontMetrics().getStringBounds( s, g);

        double strTopLeftX = 0.0; 
        double strTopLeftY = 0.0;

        if ( alignX == Component.CENTER_ALIGNMENT)
        	strTopLeftX = r.getCenterX() - rStr.getWidth() / 2.0;

        if ( alignX == Component.RIGHT_ALIGNMENT)
        	strTopLeftX = r.getMaxX() - rStr.getWidth();
        
        if ( alignX == Component.LEFT_ALIGNMENT)
        	strTopLeftX = r.getX();

        
        if ( alignY == Component.CENTER_ALIGNMENT)
        	strTopLeftY = r.getCenterY() + asc - rStr.getHeight() / 2.0;

        if ( alignY == Component.TOP_ALIGNMENT)
        	strTopLeftY = r.getY();

        if ( alignY == Component.BOTTOM_ALIGNMENT)
        	strTopLeftY = r.getMaxY() - rStr.getHeight();
        
        g.drawString( s, (int)strTopLeftX, (int)strTopLeftY);
        g.setFont( fOld);
    }

    public static Font scale2Fit( Graphics g, String s, Rectangle2D r, Font font2size) {
        if ( g == null || s == null || r == null) return g.getFont();

        float size = (float)40.0;
        Font f = font2size != null ? font2size : g.getFont();
        Rectangle2D rNow = null;

        do {
            f = f.deriveFont( size);
            size -= (float)1.0;
            rNow = g.getFontMetrics( f).getStringBounds( s, g);
        } while ( (rNow.getWidth() > r.getWidth() || rNow.getHeight() > r.getHeight()) && !( size < 5.0));

        return f;
    }
    
}
