package com.wincornixdorf.vnc.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListCellRenderer;

import com.wincornixdorf.vnc.rfb.VNCSession;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;

class VNCSessionListCellRenderer extends JPanel implements ListCellRenderer<VNCSession> {
    private static final long serialVersionUID = -767958334396045265L;
    
    private JLabel              lHTML           = null;
    private boolean             selected        = false;
    private boolean             focused         = false;

    private JLabel              dbg             = new JLabel( "Messages");
    private RemoteDesktopPanel  rdp             = new RemoteDesktopPanel( true);

    public VNCSessionListCellRenderer() {
        super();
        setLayout( new BoxLayout( this, BoxLayout.Y_AXIS));

        JScrollPane spFB = new JScrollPane();
        spFB.setPreferredSize( new Dimension( 160, 120));
        spFB.setViewportView( rdp);

        add( dbg);
        add( spFB);
        add( lHTML = new JLabel( "<none>"));
    }

    @Override
    public Component getListCellRendererComponent(
        JList<? extends VNCSession> list, VNCSession value, int index, boolean isSelected, boolean cellHasFocus)
    {
        selected = isSelected;
        focused  = cellHasFocus;

        VNCSession    s   = value instanceof VNCSession ? (VNCSession)value : null;
        VNCSessionCfg cfg = s != null ? s.getConfiguration() : null;
       
        lHTML.setText( genHTML4Label( cfg));
        lHTML.setForeground( selected ? Color.BLACK : Color.LIGHT_GRAY);

        if ( !selected && !focused) {
            setBorder( BorderFactory.createLineBorder( Color.LIGHT_GRAY));
        } else {
            //int borderType = selected ? BevelBorder.LOWERED : BevelBorder.RAISED;
            setBorder( BorderFactory.createLineBorder( Color.RED));
        }

        if ( cfg != null && !cfg.cn.isFTOnly()){
            dbg.setText( "Fine");
            rdp.setConnection( s.getFB());
        }

        return this;
    }

    private String genHTML4Label( VNCSessionCfg cfg) {
        String host = "error";
        String rept = "Direct connection";
        String user = "error";
        String prot = "error";

        if ( cfg != null && cfg.vncserver != null) {
            VNCSrvCfg srv = cfg.vncserver;

            if ( srv.requiresRepeaterToConnect())
                rept = srv.getHostParams().getOriginalHostname();

            host = srv.getRepeaterParams().getOriginalHostname();
            user = srv.getUsername();
            prot = srv.getConMode().toString();
        }

        StringBuilder sb = new StringBuilder( 1024);

        sb.append( "<html>");
        sb.append( "<table border=\"0\" cellspacing=\"1\" cellpadding=\"1\" rules=\"groups\">");

        sb.append( "<tr><th align=\"left\">Host:</th><td>"            + host + "</td></tr>");
        sb.append( "<tr><th align=\"left\">Repeater:</th><td>"        + rept + "</td></tr>");
        sb.append( "<tr><th align=\"left\">User:</th><td>"            + user + "</td></tr>");
        sb.append( "<tr><th align=\"left\">Connection Mode:</th><td>" + prot + "</td></tr>");

        sb.append( "</table>");
        sb.append( "</html>");

        return sb.toString();
    }
}
