package com.wincornixdorf.vnc.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.wincornixdorf.vnc.global.VNCAction;
import com.wincornixdorf.vnc.global.VNCAction.Goal;
import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.VNCSession;
import com.wincornixdorf.vnc.rfb.VNCSessionPool;
import com.wincornixdorf.vnc.gui.VNCSessionPanel;

public class VNCSessionPoolPanel extends JPanel
        implements ListDataListener, ListSelectionListener 
{
    private static final long serialVersionUID	= 1350165378863852556L;
    private final static String  APOLOGY_ID		= "NO_SESSIONS";

    private VNCSessionPool  	pool            		= null;
    private VNCViewer       	owner           		= null;

    private JPanel          	jpSessions      		= null;
    private JPanel          	jpApology       		= null;

    public  JList<VNCSession>	jlSessions      		= null;
    public  JPanel          	panelSessLst    		= null;

    private boolean             areClipboardsVisible	= false;
    private boolean             areOptionsVisible		= false;

    public VNCSessionPoolPanel( VNCViewer v, VNCSessionPool p) {
        super( new BorderLayout());
        setName( "SessionPoolPanel");
        owner = v;
        pool = p;
        pool.addListDataListener( this);

        jpApology = new JPanel( new BorderLayout());
        jpApology.add( new JLabel( "No sessions have been created yet, or none is selected.", JLabel.CENTER), BorderLayout.CENTER);
        jpApology.setName( APOLOGY_ID);

//         panelSessLst = new JPanel( new BorderLayout());
        Icon slIcon = (Icon)owner.actions.getActionByGoal( Goal.LIST_SESSIONS).getValue( Action.SMALL_ICON);
//         JLabel l = new JLabel( "Sessions", slIcon, SwingConstants.LEFT);
        jlSessions = new JList<VNCSession>( p);
        jlSessions.setCellRenderer( new VNCSessionListCellRenderer());
        jlSessions.setPreferredSize( new Dimension( 300, 100));
        jlSessions.addListSelectionListener( this);

        new JScrollPane( jlSessions);
        panelSessLst = new ConcealablePanel( "Sessions", slIcon, jlSessions);
        add( panelSessLst, BorderLayout.WEST);

        jpSessions = new JPanel( new CardLayout());
        add( jpSessions, BorderLayout.CENTER);
        jpSessions.add( jpApology, APOLOGY_ID);

        handleActions( true, owner.actions.getGloballyAccessibleActions());

        showOnTop( APOLOGY_ID);
    }

    private void removeMyAction( VNCAction act) {
		if ( act == null) return;

        String 		actName = (String) act.getValue( Action.NAME);
        KeyStroke ks = (KeyStroke) act.getValue( Action.ACCELERATOR_KEY);
    	
        getInputMap( WHEN_IN_FOCUSED_WINDOW).remove( ks);
        getActionMap().remove( actName);
        act.setEnabled( false);
    }

	private void addMyAction( VNCAction act) {
		if ( act == null) return;

        String 		actName = (String) act.getValue( Action.NAME);
        KeyStroke ks = (KeyStroke) act.getValue( Action.ACCELERATOR_KEY);
        
        getInputMap( WHEN_IN_FOCUSED_WINDOW).put( ks, actName);
        getActionMap().put( actName, act);
        act.setEnabled( true);
	}

	/**Adds or removes a set of VNCActions to or from the InputMap, thus making them
	 * globally accessible through their keyboard shortcuts /revoking their visibility.
	 * @param addActions Set true if the actions are to be enabled, false if they are to
	 *        be disabled.
	 * @param actions The actions to add/remove.*/
	private void handleActions( boolean addActions, Vector<VNCAction> actions) {	
		Iterator<VNCAction> itAct = actions.iterator();
		VNCAction a = null;

		while ( itAct.hasNext()) {
			a = itAct.next();

			if ( addActions) addMyAction( a);
			else removeMyAction( a);		
		}
	}

    public VNCSessionPanel active() {
         VNCSession s = (VNCSession)jlSessions.getSelectedValue();
         if ( s == null) return null;
         return getSessionPanelBySID( s.getSID());
    }

    public VNCSession getActiveSession() {
        Object o = jlSessions.getSelectedValue();
        return o != null ? (VNCSession)o : null;
    }

    public void setOptionsVisible( boolean b) {
        Component   x  = null;
        Component[] c = jpSessions.getComponents();

        for ( int i = 0; i < c.length; i++) {
            x = c[ i];
            if ( x != jpApology) ((VNCSessionPanel)x).setOptionsVisible( b);
        }
        areOptionsVisible = b;
    }

    public boolean areOptionsVisible() {
    	return areOptionsVisible;
    }

    /**Convenience method for establishing if the visibility of the SessionStatusBar
     * may be toggled or not.
     * @return True, if the Bar doesn't presently display potentially security relevant
     * information. False, if the Bar is currently warning of an insecure connection.*/
    public boolean isConnectionStatusBarToggable() {
    	VNCSessionPanel act = active();
    	return act == null ? false : !act.statusPanel.isNecessary();
    }

    public void setClipboardsVisible( boolean b) {
        Component   x  = null;
        Component[] c = jpSessions.getComponents();

        for ( int i = 0; i < c.length; i++) {
            x = c[ i];
            if ( x != jpApology) ((VNCSessionPanel)x).setClipboardsVisible( b);
        }
        areClipboardsVisible = b;
    }

    public boolean areClipboardsVisible() {
    	return areClipboardsVisible;
    }

    private VNCSessionPanel createSessionPanel( String name) {
        VNCSessionPanel sp = new VNCSessionPanel();
        sp.setName( name);
        jpSessions.add( sp, name);
        return sp;
    }

    private void setUpSession( VNCSession s) {
        if ( s == null) return;
        String name = s.getSID();
        VNCSessionPanel sp = createSessionPanel( name);
        sp.setSession( s);
        // Ensure the new Session gets selected in the list view. This will
        // trigger the selection listener, which will update the actions.
        jlSessions.setSelectedValue( s, true);
    }

    private void showOnTop( String s) {
        if ( s == null /*|| s == currentTab*/) return;
        CardLayout l = (CardLayout)(jpSessions.getLayout());
        l.show( jpSessions, s);
    }

    private Component getComponentByName( String n) {
        if ( n == null) return jpApology;

        Component   r = jpApology;
        Component   x = null;
        Component[] c = jpSessions.getComponents();
        int         i = c.length;

        while ( r == jpApology && ( (x = c[ --i]) != null))
            r = n.equalsIgnoreCase( x.getName()) ? x : jpApology;

        return r;
    }

    private VNCSessionPanel getSessionPanelBySID( String sid) {
        Component c = getComponentByName( sid);
        if ( c == null || c == jpApology) return null;
        return (VNCSessionPanel)c;
    }

    public void intervalAdded( ListDataEvent e) {
        final VNCSession s = (VNCSession)(pool.get( e.getIndex0()));
        if ( SwingUtilities.isEventDispatchThread()) {
            setUpSession( s);
        } else {
            try {
                SwingUtilities.invokeLater( new Runnable() {
                    public void run() {
                        setUpSession( s);
                    }
                });
            } catch ( Exception ex) {
                ex.printStackTrace();
            }
        }
    }

    public void intervalRemoved( ListDataEvent e) { }
    public void contentsChanged( ListDataEvent e) { }

    public void valueChanged( ListSelectionEvent e) {
        VNCSession s = (VNCSession)jlSessions.getSelectedValue();
        if ( s == null) showOnTop( APOLOGY_ID);
        else {
            showOnTop( s.getSID());
            VNCSessionPanel sp = getSessionPanelBySID( s.getSID());
            if ( sp == null) return;
            // TODO: Check if this is really necessary
            sp.setSelectedKeys( owner.actions.getActionByGoal( Goal.SHOW_FB),
                            owner.actions.getActionByGoal( Goal.SHOW_FT),
                            owner.actions.getActionByGoal( Goal.SHOW_LG));

            sp.fbWidget.thePoolPanel = this;

            owner.checkActionsAvailability();  
        }
    }
}
