package com.wincornixdorf.vnc.gui.ft;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.Vector;
import java.util.ArrayList;
import java.util.Iterator;

import com.wincornixdorf.vnc.rfb.ft.FTPath;

/**Warning! This file is UTF-8 encoded and MUST NOT be edited in a non UTF-8
environment. Failure to do so may break the application. If the testcharacters
are broken check your editors encoding setting. If they are still garbled some
insensitive clod broke it :(
Testcharacters: Ä Ö Ü ß ä ö ü Խ Ф Щ д г 걎 겡 괱 궤 꿹 خ ج ث ؤ */

/**The FTBookmarker is a widget used to assist in navigating Filesystem trees
by storing the locations which have been visited in the past and providing a
simple bookmarking functionality. The widget will store as many as MAX_HISTORY
locations before removing the least frequently visited non bookmarked location
from its list. The widget discriminates between "paths" which have been added
by visiting a location, paths which where bookmarked and so called "solid"
paths which are expected to be mount points in a file system. Solid paths as
far as the widget is concerned are "untouchable" and will only be removed if
they get replaced by a new set of solid paths.

Please take note the widget does NOT provide any functionality for storing
bookmarks permanently. This is up to the application to do.*/
public class FTBookmarker extends JComboBox<FTPath> implements MouseListener
{
    private static final long serialVersionUID = -2083270462443574851L;

    private static int MAX_HISTORY = 7;

    //private boolean remote = false;

    private class BMR extends JLabel implements ListCellRenderer<FTPath> {
        private static final long serialVersionUID = -6271157274411053996L;
        final private Icon iconM = new ImageIcon( this.getClass().getResource( "/16x16/yellowStar.png"));
        final private Icon iconU = new ImageIcon( this.getClass().getResource( "/16x16/greyStar.png"));
        //final private Icon iconD = new ImageIcon( this.getClass().getResource( "/16x16/gohome.png"));
        final private Icon iconE = new ImageIcon( this.getClass().getResource( "/16x16/exit.png"));

        final private Icon iFDD = new ImageIcon( this.getClass().getResource( "/16x16/3floppy_unmount.png"));
        final private Icon iHDD = new ImageIcon( this.getClass().getResource( "/16x16/hdd_unmount.png"));
        final private Icon iODD = new ImageIcon( this.getClass().getResource( "/16x16/cdrom_unmount.png"));
        final private Icon iNDD = new ImageIcon( this.getClass().getResource( "/16x16/nfs_unmount.png"));

        //private FTBookmarker bm = null;

        public BMR( FTBookmarker in) {
            //bm = in;
            setOpaque( true);
        }

        @Override
        public Component getListCellRendererComponent( JList<? extends FTPath> list,
        											   FTPath value,
                                                       int index,
                                                       boolean isSelected,
                                                       boolean cellHasFocus)
        {
            Icon    ico = iconE;
            String  cap = "Error";

            if ( value != null) {
                FTPath p = (FTPath)value;
                switch ( p.getType()) {
                    case PATH       : ico = p.isMarked() ? iconM : iconU; break;
                    case FLOPPY     : ico = iFDD; break;
                    case CDROM      : ico = iODD; break;
                    case LOCAL      : ico = iHDD; break;
                    case NETWORK    : ico = iNDD; break;
                    default         : break;
                }

                cap = p.getDisplayString();
            }

            setIcon( ico);
            setText( cap);
            return this;
        }
    }

    public FTBookmarker( boolean remoteIn) {
        //remote = remoteIn;
        setEditable( false);
        setRenderer( new BMR( this));
        enableInputMethods( false);
        addMouseListener( this);
    }

    /**Adds the Argument to the history and makes it the currently highlighted
    entry. The method will cause the widget to fire an Action event.
    @param s A path to be "visited".
    @return The FTPath which represents arguments the location in the CB.*/
    public FTPath visitPath( String s) {
        return addPath( s, true);
    }

    public FTPath getCurrentPath() {
        return (FTPath)getSelectedItem();
    }

    public String getCurrentLocation() {
        FTPath p = (FTPath)getSelectedItem();
        return p != null ? p.getFQN() : "";
    }

    public String getRootOfCurrentPath() {
        FTPath p = (FTPath)getSelectedItem();
        return p != null ? p.getRoot() : "";
    }

    /**Adds the paths to the history as bookmarked FTPaths. This method will
    ensure the uniqueness of the entries while making sure the currently
    selected entry does not change.
    @param paths The paths to be added*/
    public void addBookmarks( String[] paths) {
        if ( paths == null) return;
        for ( int i = 0; i < paths.length; i++) {
            FTPath p = addPath( paths[ i], false);
            if ( p != null) p.setMarked( true);
        }
    }

    public void addSolidPaths( ArrayList<FTPath> paths) {
        if ( paths == null) return;

        removeSolidPaths();
        for ( Iterator<FTPath> i = paths.iterator(); i.hasNext();) {
            FTPath p = (FTPath)i.next();
            addItem( p);
        }
    }

    /**Use this method when trying to permanently store the list of bookmarked
    paths as the counterpart addBookmarks method can handle the output.
    @return a list of String representations of all bookmarked FTPaths.*/
    public String[] getBookmarks() {
        Vector<String> bm = new Vector<String>();
        int            i  = -1;
        FTPath         p  = null;

        while ( ( p = (FTPath)getItemAt( ++i)) != null)
            if ( p.isMarked()) bm.add( p.toString());

        return bm.toArray( new String[ 0]);
    }

    /**Gets the number of Path objects marked as solid.
    @return The number of solid Path objects in the history.*/
    public int countSolidPaths() {
        int     r = 0;
        FTPath  p = null;

        for ( int i = 0; i < getItemCount(); i++) {
            p = (FTPath)getItemAt( i);
            if ( p.isSolid()) r++;
        }

        return r;
    }

    /**Removes all Path objects marked as solid. This method will at best
    leave behind only Path objects, bookmarked or not.*/
    private void removeSolidPaths() {
        int            i  = 0;
        FTPath         p  = null;

        // may look cumbersome, but removing an item will mess with
        // the items order .... obviously
        while ( ( p = (FTPath)getItemAt( i)) != null) {
            if ( p.isSolid()) removeItem( p);
            else i++;
        }
    }

    /**Adds a path to the history, while at the same time ensuring the
    uniqueness of the entry.
    @param s The path to add to the history.
    @param goThere If set the path is made the presently selected one.
    @return The FTPath representing the argument.*/
    private FTPath addPath( String s, boolean goThere) {
        if ( s == null || s.isEmpty()) return null;

        boolean pathWasKnown = true;
        FTPath pNewLocat = getPath( s);
        FTPath pSelected = (FTPath)getSelectedItem();

        if ( pNewLocat == null) {
            pathWasKnown = false;
//             pNewLocat = new FTPath( false, remote, s);
            pNewLocat = new FTPath( s/*, true*/);
            if ( pNewLocat.isUsable()) {
                addItem( pNewLocat);

                int historyLength = numberOfPaths( false);
                if ( historyLength > MAX_HISTORY) {
                    FTPath lvp = getLeastVisitedPath();
                    removeItem( lvp);
                }
            }
        }

        if ( pSelected != null && pSelected.compareTo( pNewLocat) != 0 && goThere) {
            setSelectedIndex( getIndexOfPath( s));
            if ( pathWasKnown) pNewLocat.visit();
        }

        return pNewLocat;
    }

    /**Searches the history for a FTPath representing the argument.
    @param path The path to be looking for.
    @return The index of the corresponding FTPath within the history
            or -1 if unknown.*/
    private int getIndexOfPath( String path) {
        if ( path == null || path.isEmpty()) return -1;

        int     i     = -1;
        FTPath  p     = null;
        boolean found = false;

        while ( ( p = (FTPath)getItemAt( ++i)) != null && !found)
            found = p.toString().compareTo( path) == 0;

        return found ? i - 1 : -1;
    }

    /**Searches the history for a FTPath representing the argument.
    @param path The path to be looking for.
    @return The corresponding FTPath or null if unknown.*/
    private FTPath getPath( String path) {
        int i = getIndexOfPath( path);
        return i >= 0 ? (FTPath)getItemAt( i) : null;
    }

    private FTPath getLeastVisitedPath() {
        int     i     = 0;
        int     lv    = Integer.MAX_VALUE;
        FTPath  p     = null;
        FTPath  lvp   = null;

        while ( ( p = (FTPath)getItemAt( ++i)) != null) {
            if ( !p.isSolid() && !p.isMarked() && p.getVisits() < lv) {
                lvp = p;
                lv = lvp.getVisits();
            }
        }

        return lvp;
    }

    private void toggleBookmarkStatus() {
        FTPath p = (FTPath)getSelectedItem();
        if ( p == null) return;

        p.toggleMarker();
        repaint();
    }

    @Override
    public void mouseClicked( MouseEvent e) { }
    public void mouseReleased( MouseEvent e) { }
    public void mouseEntered( MouseEvent e) { }
    public void mouseExited( MouseEvent e) { }
    public void mousePressed( MouseEvent e) {
        // Ideally this would be done in mouseClicked, but the widgets
        // behavior is getting very strange if I do that. It's working
        // only half the times it is supposed to :(
        Rectangle rIcon = new Rectangle( 0, 0, getHeight(), getHeight());
        if ( rIcon.contains( e.getPoint())) {
            e.consume();
            toggleBookmarkStatus();
            hidePopup();
        }
    }

    /**Will count the number of FTPaths having isMarked == \a beingBookmarked,
    automatically skipping solid paths.*/
    private int numberOfPaths( boolean beingBookmarked) {
        int     i     = -1;
        int     cnt   =  0;
        FTPath  p     = null;

        while ( ( p = (FTPath)getItemAt( ++i)) != null)
            if ( !p.isSolid() && p.isMarked() == beingBookmarked) cnt++;

        return cnt;
    }
}
