package com.wincornixdorf.vnc.gui.ft;

import java.util.Comparator;

import org.apache.commons.lang3.StringUtils;

import com.wincornixdorf.vnc.rfb.ft.FTFile;

/**
 * Created by janfogl.
 */
public class FTComparator {

    public static Comparator<FTFile> getFilenameComparator() {
        return new Comparator<FTFile>() {
            @Override
            public int compare(FTFile f1, FTFile f2) {
                if (f1 == null || f2 == null) {
                    throw new IllegalArgumentException("Could not compare null objects.");
                }

                if (f1.getFilename().equalsIgnoreCase("..")) {
                    return -1;
                } else if (f2.getFilename().equalsIgnoreCase("..")) {
                    return 1;
                } else if (f1.getFilename().equalsIgnoreCase("..") && f2.getFilename().equalsIgnoreCase("..")) {
                    return 0;
                }

                if (f1.isDirectory() && !f2.isDirectory()) {
                    return -1;
                } else if (!f1.isDirectory() && f2.isDirectory()) {
                    return 1;
                }

                return f1.getFilename().compareToIgnoreCase(f2.getFilename());
            }
        };
    }

    public static Comparator<FTDateColumn> getDateComparator() {
        return new Comparator<FTDateColumn>() {
            @Override
            public int compare(FTDateColumn d1, FTDateColumn d2) {

                if (d1 == null || d2 == null || StringUtils.isEmpty(d1.getFilename()) || StringUtils.isEmpty(d2.getFilename())) {
                    throw new IllegalArgumentException("Could not compare null objects.");
                }

                if (d1.toString().equalsIgnoreCase("-") || d1.getFilename().equalsIgnoreCase("..")) {
                    return -1;
                } else if (d2.toString().equalsIgnoreCase("-") || d2.getFilename().equalsIgnoreCase("..")) {
                    return 1;
                }

                return d1.getDate().compareTo(d2.getDate());

            }
        };
    }

    public static Comparator<FTLongColumn> getSizeComparator() {
        return new Comparator<FTLongColumn>() {
            @Override
            public int compare(FTLongColumn l1, FTLongColumn l2) {

                if (l1 == null || l2 == null || l1.getSize() == null || l2.getSize() == null
                        || StringUtils.isEmpty(l1.getFilename()) || StringUtils.isEmpty(l2.getFilename())) {
                    throw new IllegalArgumentException("Could not compare null objects.");
                }

                if (l1.toString().equalsIgnoreCase("0") && l1.getFilename().equalsIgnoreCase("..")) {
                    return -1;
                } else if (l2.toString().equalsIgnoreCase("0") && l2.getFilename().equalsIgnoreCase("..")) {
                    return 1;
                }

                return l1.getSize().compareTo(l2.getSize());

            }
        };
    }

}
