/**
 *
 */
package com.wincornixdorf.vnc.rfb;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.rfb.debug.ToolBox;

import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.font.NumericShaper;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;

/**
 * @author jan.fogl
 */


public class RFBEventBuffer {


    private RFBEventSender rfbEventSender;

    private boolean disabled = false;

    /**
     * @param wantsMouseButtonsReversed If set, this Buffer will send a right MB click instead of a left one and vice versa.
     * @param disabled                  When disabled, the Event Buffer will not write out any events.
     */
    public RFBEventBuffer(boolean wantsMouseButtonsReversed, boolean disabled, OutputStream outputStream) {
        this.disabled = disabled;
        this.rfbEventSender = new RFBEventSender(wantsMouseButtonsReversed, outputStream);

    }

    public void writeKeyEvent(KeyEvent event, VNCSessionCfg.KeyboardLayout keyboardLayout) {

        if (disabled || event == null) {
            return;
        }

        boolean shortcut = isShortcut(event);

        if (shortcut && event.getID() == KeyEvent.KEY_PRESSED) {
            sendKeySequence(event, keyboardLayout);
            return;
        } else {
            rfbEventSender.writeKeyEvent(event);
        }

        // we need to reset all modifiers when the button is released
        if (event.getID() == KeyEvent.KEY_RELEASED) {
            rfbEventSender.resetKeyEventModifiers();
        }
    }

    private void sendKeySequence(KeyEvent event, VNCSessionCfg.KeyboardLayout layout) {

        switch (layout) {
            case US:
            case FR:
            case DE:
            case ES:

                // AT Sequence
                if (event.getKeyCode() == KeyEvent.VK_AT) {

                    // windows recognize altgr key as alt + ctrl. Need to reset modifiers
                    rfbEventSender.resetKeyEventModifiers();

                    KeyEvent sixEvt = new KeyEvent(event.getComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD6, '6');

                    KeyEvent fourEvt = new KeyEvent(event.getComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD4, '4');

                    KeyEvent sixEvtRelease = new KeyEvent(event.getComponent(), KeyEvent.KEY_RELEASED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD6, '6');

                    KeyEvent fourEvtRelease = new KeyEvent(event.getComponent(), KeyEvent.KEY_RELEASED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD4, '4');


                    rfbEventSender.writeKeyEvent(sixEvt, sixEvtRelease, fourEvt, fourEvtRelease);
                    rfbEventSender.resetKeyEventModifiers();
                }

                if (event.getKeyCode() == KeyEvent.VK_EURO_SIGN) {

                    // windows recognize altgr key as alt + ctrl. Need to reset modifiers
                    rfbEventSender.resetKeyEventModifiers();

                    KeyEvent zeroEvt = new KeyEvent(event.getComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD0, '0');

                    KeyEvent oneEvt = new KeyEvent(event.getComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD1, '1');

                    KeyEvent twoEvt = new KeyEvent(event.getComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD2, '2');

                    KeyEvent eightEvt = new KeyEvent(event.getComponent(), KeyEvent.KEY_PRESSED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD8, '8');

                    KeyEvent zeroEvtRelease = new KeyEvent(event.getComponent(), KeyEvent.KEY_RELEASED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD0, '0');

                    KeyEvent oneEvtRelease = new KeyEvent(event.getComponent(), KeyEvent.KEY_RELEASED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD1, '1');

                    KeyEvent twoEvtRelease = new KeyEvent(event.getComponent(), KeyEvent.KEY_RELEASED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD2, '2');

                    KeyEvent eightEvtRelease = new KeyEvent(event.getComponent(), KeyEvent.KEY_RELEASED, new Date().getTime(), KeyEvent.ALT_MASK, KeyEvent.VK_NUMPAD8, '8');

                    rfbEventSender.writeKeyEvent(zeroEvt, zeroEvtRelease, oneEvt, oneEvtRelease, twoEvt, twoEvtRelease, eightEvt, eightEvtRelease);
                    rfbEventSender.resetKeyEventModifiers();

                }

                break;
        }

    }

    private boolean isShortcut(KeyEvent keyEvent) {

        if (keyEvent.getKeyCode() == KeyEvent.VK_EURO_SIGN || keyEvent.getKeyCode() == KeyEvent.VK_AT) {
            return true;
        }

        return false;

    }

    public void writePointerEvent(MouseEvent event) {

        if (disabled || event == null) {
            return;
        }

        rfbEventSender.writePointerEvent(event);
    }

    public boolean isAltKeyJammed() {
        return rfbEventSender.isJammedAlt();
    }


    public boolean toggleAltJam() {
        return rfbEventSender.toggleAltJam();
    }


}
