package com.wincornixdorf.vnc.rfb;

import java.io.BufferedReader;
import java.io.File;
import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.Component;
import java.awt.event.KeyEvent;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.image.MemoryImageSource;
import java.net.Socket;
import java.nio.charset.Charset;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Date;
import java.util.Vector;
import java.util.Iterator;
import java.util.logging.Level;
import java.lang.Runnable;
import java.util.Observer;
import java.util.Observable;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.Transferable;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.net.ssl.SSLPeerUnverifiedException;
import javax.net.ssl.SSLSession;
import javax.net.ssl.SSLSocket;
import javax.swing.KeyStroke;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.global.VNCViewerOptions;
import com.wincornixdorf.vnc.gui.ft.FTStatus;
import com.wincornixdorf.vnc.rfb.auth.RFBHandlerLA;
import com.wincornixdorf.vnc.rfb.conf.SettingsCN;
import com.wincornixdorf.vnc.rfb.conf.SettingsFB;
import com.wincornixdorf.vnc.rfb.conf.SettingsFT;
import com.wincornixdorf.vnc.rfb.conf.SettingsGB;
import com.wincornixdorf.vnc.rfb.conf.VNCSessionCfg;
import com.wincornixdorf.vnc.rfb.conf.VNCSrvCfg;
import com.wincornixdorf.vnc.rfb.debug.ToolBox;
import com.wincornixdorf.vnc.rfb.exceptions.LoginException;
import com.wincornixdorf.vnc.rfb.fb.PixelFormat;
import com.wincornixdorf.vnc.rfb.fb.RFBHandlerFB;
import com.wincornixdorf.vnc.rfb.fb.RemoteCursor;
import com.wincornixdorf.vnc.rfb.fb.VNCEncoding;
import com.wincornixdorf.vnc.rfb.ft.RemoteFileSystem;
import com.wincornixdorf.vnc.rfb.ft.FTPath;
import com.wincornixdorf.vnc.rfb.ft.RFBHandlerFT;
import com.wincornixdorf.vnc.rfb.stream.RecordableInputStream;
import com.wincornixdorf.utils.httpSocket.HTTPClientSocket;

public class RFBHandler extends Observable
        implements Runnable, ClipboardOwner, Observer {

    public enum Phase {
        UNKNWON,        // Used internally.
        PENDING,        // Connection awaits users "GO!"
        PREPARING,      // Setting up socket etc.
        CONNECTING,     // Opening connection, waiting 4 initial message (RFB 003.XYZ)
        ROUTING,        // Talking to a repeater, trying to get forwarded.
        AUTHENTICATING, // Talking to the actual server, trying 2 get authenticated.
        NEGOTIATING,    // Negotiating the RFB parameters with the server.
        PROCESSING,     // Processing the stream like reading FB packets
        DISCONNECTING,   // Closing down sockets, preparing 4 reuse.
        RECONNECTING   // Reconnecting to vncserver.
    }


    private Phase phase = Phase.PENDING;

    private boolean isThreadDoomed = false;

    private ConcurrentLinkedQueue<RFBConnJob> jobs = new ConcurrentLinkedQueue<RFBConnJob>();

    // These are SO PRIVATE! Hands off or my wrath shall be eternal!
    /**
     * The sacred stream the session uses to read from the socket. Come heaven, hell high waters or
     * a newsflash saying hell had just frozen over DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!
     * On top of that it's important to know this stream is shared with several other objects to
     * facilitate the segmentation of the code into less bulky units.
     */
    private DataInputStream is = null;

    /**
     * The sacred stream the session uses to write to the socket. Come heaven, hell high waters or
     * a newsflash saying hell had just frozen over DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!
     * On top of that it's important to know this stream is shared with several other objects to
     * facilitate the segmentation of the code into less bulky units.
     */
    private OutputStream os = null;

    private RecordableInputStream vncis = null;

    /**
     * The socket, the DataInputStream is and the OutputStream os refer to.
     */
    private Socket sock = null;

    /**
     * A handle to get back at the VNCSession this connection is a part of.
     */
    private VNCSession sess = null;

    /**
     * The file transfer unit is an object devoted to encapsulating all code
     * necessary for authenticating this client against the server. For reasons of
     * thread safety this object must NEVER be shared with another Thread.
     */
    private RFBHandlerLA lau = null;

    /**
     * The file transfer unit is an object devoted to encapsulating all code
     * necessary for transferring files and directories using the RFB protocol.
     * For reasons of thread safety this object must NEVER be shared with another
     * Thread.
     */
    private RFBHandlerFT ftu = null;

    /**
     * The frame buffer unit is an object devoted to encapsulating all code
     * necessary for handling frame buffer messages. For reasons of thread safety
     * this object must NEVER be shared with another Thread.
     */
    private RFBHandlerFB fbu = null;

    private RFBEventBuffer bufEvents = null;

    /**
     * True if this is the VNCConnection is to transfer files only
     * operations. FT operations will be performed only if this flag has been
     * set.
     */
    private VNCViewerOptions.VNCMode mode = null;
    /**
     * Used only for internal debugging purposes.
     */
    private String connType = "";

    /**
     * Buffer used for storing clipboard text received from the server.
     */
    private String clipboardText = null;

    /**
     * Holds the last fatal error message thrown by this object. This will
     * usually be the stack trace of some exception. It is highly unlikely
     * the user will be able to understand it.
     */
    //private String stErrorMsg                       = null;
    private Throwable causeOfCrash = null;

    /**
     * Holds the last human readable fatal error message thrown by this object.
     * Unlike stErrorMsg this message is designed to be understood by
     * common users. It should give a hint at hot to resolve the problem.
     */
    private String humanreadableErrorMsg = null;

    /**
     * Used for distinguish whenever remote controlling should be enabled or not.
     */
    private boolean disabledRemoteControlling = false;

    private int conectionTriesCounter = 0;

    // TODO: This one stinks to high heaven! Get rid of it!
    public String logStr(String s) {
        return (connType + s);
    }

    public RFBHandler(VNCSession sess, VNCViewerOptions.VNCMode mode) {
        this.sess = sess;
        setConnectionMode(mode);

        this.mode = mode;

        ftu = new RFBHandlerFT(this);
        fbu = new RFBHandlerFB(this);
        lau = new RFBHandlerLA(this);

        // install the connection as an observer for the FB/FT unit so we can
        // forward the VNCConnectionEvents they emit.
        ftu.addObserver(this);
        fbu.addObserver(this);

        SettingsCN opts = getSettingsCN();
        opts.addObserver(this);
    }

    private boolean isFBConnection() {
        switch (mode) {
            case FB_ONLY:
            case FBANDFT:
                return true;
            default:
                return false;
        }
    }

    private boolean isFTConnection() {
        switch (mode) {
            case FT_ONLY:
            case FBANDFT:
                return true;
            default:
                return false;
        }
    }

    private void setConnectionMode(VNCViewerOptions.VNCMode mode) {
        switch (mode) {
            case FT_ONLY:
                connType = "[FT]";
                break;
            case FB_ONLY:
                connType = "[FB]";
                break;
            default:
                connType = "[FB]&[FT]";
        }
    }

    /**
     * @return Returns the VNCSession this connection belongs to.
     */
    public VNCSession getSession() {
        return sess;
    }

    public RFBHandler getPartnerConnection() {
        if (getSession() == null) return null;

        if (isFBConnection()) return (getSession().getFT());
        else return (getSession().getFB());
    }

    public RemoteCursor getCursor() {
        return fbu.getRemoteCursor();
    }

    public boolean queue(RFBConnJob j) {
        if (j != null) {
            return jobs.add(j);
        }
        return false;
    }

    public boolean hasAnErrorMessage() {
        return humanreadableErrorMsg != null || causeOfCrash != null;
    }

    public Throwable getCauseOfCrash() {
        return causeOfCrash;
    }

    public String getHRErrorMessage() {
        return humanreadableErrorMsg;
    }

    /**
     * @return Will return true, the connection uses a socket and it is a SSL socket.
     */
    public boolean isEncrypted() {
        return sock != null && sock instanceof SSLSocket;
    }

    public boolean isTrustingAll() {
        VNCSessionCfg cfg = sess.getConfiguration();
        if (cfg == null) return true;
        return cfg.isTrustingAllCertificates();
    }

    public boolean hasVerifiedPeer() {
        if (!isEncrypted() || isTrustingAll()) return false;

        try {
            SSLSession s = getSSLSession();
            Principal p = s != null ? s.getPeerPrincipal() : null;
            return p != null;
        } catch (SSLPeerUnverifiedException e) {
            return false;
        }
    }

    public SSLSession getSSLSession() {
        if (!isEncrypted()) return null;
        SSLSocket secs = (SSLSocket) sock;
        return secs.getSession();
    }

    public boolean canHandleFileTransfers() {
        return true;
    }

    public boolean canHandleFramebufferUpdates() {
        return true;
    }

    /**
     * Users can set the encoding using the fancy GUI all they want. If the
     * server does not offer the requested encoding it will fall back to whatever
     * encoding it does know. If all else fails, the server will probably start
     * sending RAW encoded images.
     *
     * @return Will return the encoding actually used by the server to encode the FB data.
     */
    public VNCEncoding getEncodingUsedByServer() {
        return fbu.getEncodingUsedByServer();
    }

    /**
     * @return Returns true, if this handler has a HTTPClientSocket for a source.
     */
    public boolean usesHTTPConnection() {
        if (sock == null) return false;
        return sock instanceof HTTPClientSocket;
    }

    private boolean isWebLoggerDown() {
//        if (getServerParams().isRemoteSupport()) {
//            //FIXME not needed ??
//            return false;
//        }

        if (VNCViewer.canLog2Web()) return false;
        // TODO: Get rid of this "security feature". The server / repeater should log the users actions. NOT THE CLIENT!
        humanreadableErrorMsg = "The web logger is down! For security reasons connections can't be created.";
        //stErrorMsg = hrErrorMsg;
        setPhase(Phase.PENDING);
        return true;
    }

    public boolean isAltKeyJammed() {
        return bufEvents.isAltKeyJammed();
    }

    /**
     * Implementation of the run method from the Runnable Interface.
     * It will begin a VNC session by reading the protocol version from
     * the server and reply with a corresponding Version String. Once
     * the "handshake" is complete, the authentication scheme will be
     * agreed upon and an attempt to authenticate against the server will
     * be made. On successful authentication the method will begin to
     * process the data stream coming from the server.
     */
    public void run() {
        SettingsCN opts = getSettingsCN();
        boolean webLoggerDown = isWebLoggerDown();
        isThreadDoomed = isThreadDoomed || opts == null || webLoggerDown;

        boolean reconnectingMode = false;
        while (!isThreadDoomed) {

            if (getPhase().equals(Phase.RECONNECTING)) {
                reconnectingMode = true;
            } else {
                setPhase(Phase.PENDING);
            }

            VNCViewer.pen.log(Level.FINEST, logStr("Entered PENDING state."));

            while (phase == Phase.PENDING && !isThreadDoomed) {
                processRequests();
                try {
                    Thread.sleep(750);
                } catch (Exception ex) {
                }
            }

            // These two lines are part of a dirty little trick to see if the web logger
            // is up and running, so we can leave this method very quickly!
            isThreadDoomed = isThreadDoomed || isWebLoggerDown(); // Don't know if this is such a grand idea ...

            // I don't like break, but it is fair to use it here:
            // If the connection is doomed as we leave the pending state, there
            // is no point in trying to bring up the connection.
            if (isThreadDoomed) break;

            forgetPreviousErrors();
            VNCViewer.pen.log(Level.FINEST, logStr("Quit hovering"));

            setPhase(Phase.PREPARING);

            try {
                setUpSocket();

                if (getServerParams().isRemoteSupport()) {
                    initRemoteSupportBroker();
                } else {
                    digg4Server();
                }

                setPhase(Phase.AUTHENTICATING);
                if (!lau.tryToConnect()) {
                    crash("Unable to log in. Please check username and password.", null);
                    VNCViewer.pen.log(Level.SEVERE, logStr("Logon failed :("));
                } else {
                    VNCViewer.pen.log(Level.CONFIG, logStr("Logon successfull :)"));

                    setPhase(Phase.NEGOTIATING);
                    doProtocolInitialisation();
                    fireEvent(RFBHandlerEvent.connect(this), null);

                    VNCViewer.pen.log(Level.CONFIG, logStr("PROTOCOL INITIALISATION Successfull :)"));

                    if (isFBConnection() && opts.wantsConsoleBlanked()) {
                        VNCViewer.pen.log(Level.CONFIG, logStr("Trying to power down remote console as configured."));
                        sendServerBlankConsoleRequest(true);
                    } else VNCViewer.pen.log(Level.CONFIG, logStr("Didn't try to power down remote console."));

                    if (reconnectingMode) {
                        processReconnecting();
                        reconnectingMode = false;
                    }
                    processRFBProtocol();
                }
            } catch (IOException ioe) {
                VNCViewer.pen.log(Level.SEVERE, logStr("The connection with the VNC server was lost!"), ioe);
                String errMsg = phase != Phase.NEGOTIATING ? "General network error. Please check network connection and ensure server is running." :
                        "Server did respond, but failed to offer a desktop. Please ensure the server is run as a service.";
                processCommunicationError(errMsg, ioe);
            } catch (LoginException lie) {
                VNCViewer.pen.log(Level.SEVERE, logStr("Could not log on to the server!"), lie);
                processCommunicationError("Fatal error trying to log on to server.", lie);
            } catch (Exception ge) {
                ge.printStackTrace(System.err);
                VNCViewer.pen.log(Level.SEVERE, logStr("An exception occurred in run method."), ge);
                processCommunicationError("Unexpected error.", ge);
            }

            VNCViewer.pen.log(Level.FINEST, logStr("Cleaning up the sockets."));
            // This disconnect is invoked if the CONNECTION died.
            if (getServerParams().isRemoteSupport() || conectionTriesCounter > 3 || conectionTriesCounter <= 0) {
                disconnect();
            }
        }

        // This disconnect is invoked, when the THREAD dies.
        disconnect();

        VNCViewer.pen.log(Level.FINEST, logStr("The RFBHandler thread has been closed."));
    }

    private void processReconnecting() {
        conectionTriesCounter = 0;
        fbu.processReconnecting();
    }

    private void processCommunicationError(String errMsg, Throwable ioe) {
        if (!getServerParams().isRemoteSupport()) {
            setPhase(Phase.RECONNECTING);
            conectionTriesCounter++;
            try {
                Thread.sleep(2000);
            } catch (InterruptedException e) {
                // no need
            }
        } else {
            crash(errMsg, ioe);
        }
    }

    private void forgetPreviousErrors() {
        causeOfCrash = null;
        humanreadableErrorMsg = null;
    }

    private void setUpSocket() throws Exception {
        SocketForge sf = new SocketForge(getServerParams(), isFBConnection());
        sock = sf.call();

        if (sock != null) VNCViewer.pen.log(Level.FINEST, logStr("The socket was set up."));

        is = new DataInputStream(new BufferedInputStream(sock.getInputStream()));
        os = sock.getOutputStream();
        bufEvents = new RFBEventBuffer(getSettingsCN().wantsReversedMouseButtons(), !isFBConnection(), os);


        fbu.setStreams(is, os);
        ftu.setStreams(is, os);
        lau.setStreams(is, os);
    }

    /**
     * This is de facto the main event loop of this thread. It handles the
     * server messages read from the stream and invokes the appropriate methods as
     * is necessary.
     */
    private void processRFBProtocol() throws Exception {
        if (isFTConnection()) {
            if (!readServerDriveList())
                VNCViewer.pen.log(Level.SEVERE, logStr("Failed to read the list of available drives."));

            SettingsFT sft = getSettingsFT();
            String defDir = sft != null ? sft.getPreferredServerDir() : null;
            if (defDir != null) {
                VNCViewer.pen.log(Level.CONFIG, logStr("Switching to last used dir [" + defDir + "]"));
                ftu.setCurrentDir(defDir);
                //ftu.changeServerDirectory( defDir);
            }
        }

        VNCViewer.pen.log(Level.FINEST, logStr("Starting to process the RFB protocol"));
        fbu.requestInitialFB();


        int msgType = -1;
        //int repDel = getSettings() != null ? getSettings().getReplayDelay() : SettingsCN.DEFREPLAYDELAY;

        while (!isThreadDoomed && phase == Phase.PROCESSING) {
            if (vncis != null) vncis.flush();
            ///if ( hasFlag( REPLAYING)) Thread.sleep( repDel);

            // Here we set the socket timeout to be used when waiting for the
            // msgType byte from the server. The timeout is deliberately set as
            // low as possible because each time we fail to read the msgType
            // byte, the thread will process the user requests in the queue.
            sock.setSoTimeout(100);

            try {
                msgType = -1;
                msgType = is.readUnsignedByte();
                //VncViewer.pen.log( Level.FINEST, logStr( "msgType read [" + msgType + "]"));
            } catch (java.net.SocketTimeoutException ste) {
            }

            // Once we have read the msgType byte, we need to increase the
            // timeout again, because unlike the above read, all consecutive
            // reads will cause an exception to be thrown if they do not
            // receive a byte by the time the socket times out.
            sock.setSoTimeout(sess.getConfiguration().cn.getSocketTimeout());

            // Process the message depending on its type.
            switch (msgType) {
                case -1:
                    if (fbu.isDecoderChanged()) {
                        VNCViewer.pen.log(Level.FINEST, logStr("Decoder changed. Processing full refresh of inactive screen."));
                        fbu.requestFullFBUpdate();
                        fbu.setDecoderChanged(false);
                    }
                    break;

                case RFB.FramebufferUpdate:
                    //if ( !isFBConnection)
                    //    throw new Exception( "File transfer connection received a framebuffer datagram.");
                    fbu.handleFBUpdate();
                    break;

                case RFB.rfbFileTransfer:
                    if (mode == VNCViewerOptions.VNCMode.FB_ONLY)
                        throw new Exception("Framebuffer connection received a file transfer datagram.");
                    ftu.handleFTUpdate();
                    break;

                case RFB.Bell:
                    Toolkit.getDefaultToolkit().beep();
                    break;

                case RFB.ServerCutText:
                    byte[] pad = new byte[3];
                    is.readFully(pad);
                    int len = is.readInt();
                    byte[] buffer = new byte[len];
                    is.readFully(buffer);
                    VNCViewer.pen.log(Level.FINEST, logStr("Received [" + len + "] bytes of clipboard text from Server!"));
                    clipboardText = new String(buffer, getPresumedEncoding());
                    setClipboardContents(clipboardText);
                    fireEvent(RFBHandlerEvent.clipchanged(this), null);
                    break;

                case RFB.SetColourMapEntries:
                    fbu.handleColorMapEntries();
                    break;

                default:
                    VNCViewer.pen.log(Level.SEVERE, logStr("Unknown RFB message type [" + msgType + "]"));
                    byte[] paddy = new byte[10];
                    is.readFully(paddy);

                    for (int i = 0; i < 10; i++)
                        VNCViewer.pen.log(Level.SEVERE, "" + i + " " + paddy[i]);

                    throw new Exception("Unknown RFB message type [" + msgType + "]");
            }

            while (processRequests()) {
            }
            try {
                Thread.yield();
            } catch (Exception e) {
            }
        }

        VNCViewer.pen.log(Level.FINE, logStr("Done processing the RFB protocol."));
    }

    /**
     * Retrieves the Charset which is being used when converting in between strings
     * and bytes in read and write operations.
     *
     * @return The current Charset or the Default Charset as defined by the RFB protocol.
     */
    public Charset getPresumedEncoding() {
        SettingsCN cn = getSettingsCN();
        if (cn == null || cn.presumedServerEncoding == null) return RFB.DEFAULT_CHARSET;
        return cn.presumedServerEncoding.getSelectedCharset();
    }

    private boolean processRequests() {

        isThreadDoomed = isThreadDoomed || isWebLoggerDown(); // Don't know if this is such a grand idea ...

        RFBConnJob j = jobs.poll();

        if (j == null || (phase != Phase.PROCESSING && j.requiresWorkingConnection())) {
            return false;
        }

        if (j.operation.isControlsType() && isDisabledRemoteControlling()) {
            return true;
        }

        switch (j.operation) {
            case NONE:
                break;
            case CONNECT:
                setPhase(Phase.PREPARING);
                break;
            case DISCONNECT:
                disconnect();
                break;
            case TERMINATE_THREAD:
                VNCViewer.pen.log(Level.SEVERE, logStr("Connection has been doomed."));
                isThreadDoomed = true;
                break;
            case TRANSFER:
                FTStatus result = performTransferOperation(j);
                if (result.equals(FTStatus.ANOTHER_FT_IN_PROGRESS)) {
                    queue(j);
                    return false;
                }
                break;
            case DELETEFILE:
                FTStatus delResult = deleteRemoteFile(j.srcFile);
                if (delResult.equals(FTStatus.ANOTHER_FT_IN_PROGRESS)) {
                    queue(j);
                    return false;
                }
                break;
            case CREATEDIR:
                createRemoteDirectory(j.srcFile);
                break;
            case CHANGEDIR:
                if (j.isDirectory) changeServerDirectory_wrapper(j.dstFile);
                else navigateTo(j.dstFile);
                break;
            case TOGGLEJAMALT:
                boolean isJammedNow = bufEvents.toggleAltJam();
                setChanged();
                notifyObservers(RFBHandlerEvent.keyJamChanged(this, isJammedNow, InputEvent.ALT_MASK));
                break;
            case LISTDRIVES:
                readServerDriveList();
                break;
            case REFRESH:
                sendScreenRefresh();
                break;
            case SHUTDOWNCONSOLE:
                sendServerBlankConsoleRequest(true);
                break;
            case POWERUPCONSOLE:
                sendServerBlankConsoleRequest(false);
                break;
            case MOUSE:
                sendMouseEvent(j.mouseEvent);
                break;
            case KEYPRESS:
                sendKeyPress(j.keyEvent);
                break;
            case KEYSTROKE:
                sendKeyStroke(j.component, j.keyStroke);
                break;
            case META:
                sendMetaKeypress(j.component);
                break;
            case MONKEY:
                sendMonkeyGrip(j.component);
                break;
            case CLIPBOARD:
                sendToRemoteClipboard(j.srcFile);
                break;
            case REENCODE:
                setEncodings();
                break;
            case PIXFORMAT:
                break;
//            case PIXFORMAT          : fbu.hasNewPixmapRequest=true;					break;
//            case PIXFORMAT          : fbu.setPixelFormat();   break;
        }
        return true;
    }

    public MemoryImageSource getRemoteDesktopSource() {
        return fbu.getRemoteDesktopSource();
    }

    public MemoryImageSource getRemotePointerSource() {
        RemoteCursor rCursor = fbu.getRemoteCursor();
        return rCursor != null ? rCursor.getImageSource() : null;
    }

    public String getRemoteDesktopName() {
        return fbu.getRemoteDesktopName();
    }

    public Dimension getRemoteDesktopOriginalSize() {
        return fbu.getFBDim();
    }

    public int getRemoteDesktopsBytesPerPixel() {
        return fbu.getBytesPerPixel();
    }

    public VNCSrvCfg getServerParams() {
        return sess.getHost();
    }

    public SettingsGB getSettingsGB() {
        VNCSessionCfg cfg = sess.getConfiguration();
        return cfg != null ? cfg.gb : null;
    }

    public SettingsCN getSettingsCN() {
        VNCSessionCfg cfg = sess.getConfiguration();
        return cfg != null ? cfg.cn : null;
    }

    public SettingsFB getSettingsFB() {
        VNCSessionCfg cfg = sess.getConfiguration();
        return cfg != null ? cfg.fb : null;
    }

    public SettingsFT getSettingsFT() {
        VNCSessionCfg cfg = sess.getConfiguration();
        return cfg != null ? cfg.ft : null;
    }

    private void disconnect() {

        VNCViewer.pen.log(Level.FINEST, logStr("Trying to disconnect from the server."));
        if (sock == null) {
            VNCViewer.pen.log(Level.WARNING, logStr("FAILED to disconnect: No socket to close!"));
            return;
        }

        setPhase(Phase.DISCONNECTING);

        if (sock.isClosed()) {
            VNCViewer.pen.log(Level.FINEST, logStr("Didn't try to disconnect as socked was already closed."));
            return;
        }

        jobs.clear();
        fbu.resetDecoders();

        try {
            if (is != null) {
                is.close();
                is = null;
            }
            if (os != null) {
                os.close();
                os = null;
            }

            sock.close();
            sock = null;
            VNCViewer.pen.log(Level.FINE, logStr("The socket has been closed."));

            //lau = null; fbu = null; ftu = null;

            fireEvent(RFBHandlerEvent.disconnect(this, true), null);
        } catch (IOException e) {
            VNCViewer.pen.log(Level.WARNING, logStr("Failed to close the socket."));
        }

        VNCSrvCfg s = getServerParams();
        String hn = s != null ? s.getHostParams().getOriginalHostname() : "unknown";
        VNCViewer.pen.log(Level.INFO, logStr("Disconnected from [" + hn + "]"));

        // This is NOT a goof. Finer messages are relayed to the Web-Logger!
        VNCViewer.pen.log(Level.FINER, logStr("Disconnected from [" + hn + "]"));
    }

    private void digg4Server() throws IOException {
        if (!is.markSupported())
            VNCViewer.pen.log(Level.SEVERE, logStr("Failed to connect to Server: Input stream does not support Marks!"));

        VNCViewer.pen.log(Level.FINEST, logStr("Reading ahead of stream to detect if we are at the Server already."));

        is.mark(RFB.REPEATER_IDENTIFIER.length() + 15);

        byte[] hostMsgAsBytes = new byte[RFB.REPEATER_IDENTIFIER.length()];
        is.read(hostMsgAsBytes);

        String supposedVersionMessage = new String(hostMsgAsBytes, RFB.DEFAULT_CHARSET); // US-ASCII as per the standard
        VNCViewer.pen.log(Level.FINEST, logStr("Server identified itslf as : [" + ToolBox.escapeCRLF(supposedVersionMessage) + "]"));

        // If this looks like a Server, we can leave here. Otherwise, we have to ask what
        // may be a Repeater to forward us to the desired Host.
        if (supposedVersionMessage.compareTo(RFB.REPEATER_IDENTIFIER) != 0) {
            if (getServerParams().requiresRepeaterToConnect())
                VNCViewer.pen.log(Level.WARNING, logStr("Expected to find a Repeater, but encountered a Server! Trying to continue."));
            else
                VNCViewer.pen.log(Level.FINEST, logStr("Found what may be a Server."));

            is.reset();
            VNCViewer.pen.log(Level.FINEST, logStr("Resetting stream as reader has read bytes ahead of schedule."));
            return;
        }

        // VNCSrvCfg s = vc != null ? vc.getServerParams() : null;
        // if ( s != null && s.requiresTunnelToConnect()) {
        //     VncViewer.pen.log( Level.FINEST, vc.logStr( "Reading connection ID for stateful HTTP(s) connection."));
        //     String id = is.readLine();
        //     VncViewer.pen.log( Level.INFO, vc.logStr( "Connection was assigned ID [" + id + "] by the repeater."));
        //     ((HTTPClientSocket)sock).
        // }

        // This code is an extension to the original VNC protocol. If connecting
        // via a repeater using a HTTP/HTTPS tunnel, the client is expecting the
        // server to send an ID which the client will as of then have to send
        // with every request so the HTTP/HTTPS server will know what client it is talking to.
        // if ( usesHTTPConnection()) ((HTTPClientSocket)sock).readID();
        //         if ( usesHTTPConnection()) {
        //             String conid = is.readLine();
        //             VncViewer.pen.log( Level.CONFIG, logStr( "This socket will use the ID [" + conid + "] as of now."));
        //             ((HTTPClientSocket)sock).setID( conid);
        //         }

        setPhase(Phase.ROUTING);

        String repMsg = composeRepeaterMsg();
        VNCViewer.pen.log(Level.FINEST, logStr("Sending answer: [" + ToolBox.escapeCRLF(repMsg) + "]"));
        os.write(repMsg.getBytes(RFB.DEFAULT_CHARSET));
        VNCViewer.pen.log(Level.FINEST, logStr("Wrote the answer. Will flush now."));
        os.flush();
        VNCViewer.pen.log(Level.FINEST, logStr("sent answer: [" + ToolBox.escapeCRLF(repMsg) + "]"));
    }


    private void initRemoteSupportBroker() throws IOException {

        VNCSrvCfg s = getServerParams();
        String repMsg = s.getRepeaterMsg();//s.getRepeaterMsgFullLength(250);
        VNCViewer.pen.log(Level.FINEST, logStr("Full msg: " + repMsg));

        // tell broker that we are vnc viewer
        String repMsgType = "testB";
        VNCViewer.pen.log(Level.FINEST, logStr("Sending vnc viewer msg: [" + ToolBox.escapeCRLF(repMsgType) + "]"));
        os.write(repMsgType.getBytes(RFB.DEFAULT_CHARSET));
        os.flush();

        // read zeros from broker
        byte[] hostMsgAsBytes = new byte[RFB.REPEATER_IDENTIFIER.length()];
        is.read(hostMsgAsBytes);
        String supposedVersionMessage = new String(hostMsgAsBytes, RFB.DEFAULT_CHARSET); // US-ASCII as per the standard
        VNCViewer.pen.log(Level.FINEST, logStr("Server identified itself as : [" + ToolBox.escapeCRLF(supposedVersionMessage) + "]"));
        if (supposedVersionMessage.compareTo(RFB.REPEATER_IDENTIFIER) != 0) {
            is.reset();
            VNCViewer.pen.log(Level.FINEST, logStr("Resetting stream as reader has read bytes ahead of schedule."));
            return;
        }

        //send 250 characters starting with VNC endpoint ID 
        VNCViewer.pen.log(Level.FINEST, logStr("Sending answer: [" + ToolBox.escapeCRLF(repMsg) + "]"));
        os.write(doZeroPaddingBytes(repMsg, 250));
        VNCViewer.pen.log(Level.FINEST, logStr("Wrote the answer. Will flush now. 1"));
        os.flush();

        setPhase(Phase.ROUTING);
    }

    private byte[] doZeroPaddingBytes(String repMsg, int length) {
        byte[] zeros = new byte[length];
        byte[] msg = repMsg.getBytes(RFB.DEFAULT_CHARSET);
        for (int i = 0; i < msg.length; i++) {
            zeros[i] = msg[i];
        }
        return zeros;
    }

    /**
     * A convenience method which will "interpret the "Repeater message" where necessary.
     *
     * @return The request string to send to the Repeater instead of the usual host:port one.
     * Unlike {@link getRepeaterMsg()}, this method will add the
     */
    private String composeRepeaterMsg() throws IOException {
        VNCSrvCfg s = getServerParams();
        String repeaterMsg = s.getRepeaterMsg();

        if (repeaterMsg.isEmpty()) {
            VNCViewer.pen.log(Level.FINEST, logStr("No RepMsg specified. Requesting Hostname:Port"));
            return s.getHostnameAndPort(false, true) + "\n";
        }

        String r = repeaterMsg;

        // RepMsgFileReader: This is where either the RepMsg parameter
        // is appended to the original Repeater message, or if the RepMsg
        // starts with a "$" sign the contents of the file specified there.
        if (repeaterMsg.startsWith("$")) {
            String fn = repeaterMsg.substring(1);
            VNCViewer.pen.log(Level.WARNING, logStr("Trying to read RepMsg from local file [" + fn + "]!"));
            FileReader fr = new FileReader(fn);
            BufferedReader br = new BufferedReader(fr);
            r = br.readLine();
            fr.close();
        } else VNCViewer.pen.log(Level.FINEST, logStr("RepMsg is NOT to be read from a file."));

        // scrub any potentially trailing crlf's as the string is not expected to have any   
        if (r.endsWith("\n")) {
            r = r.substring(0, r.length() - 1);
            VNCViewer.pen.log(Level.WARNING, logStr("Had to remove a trailing CRLF! This shouldn't have been here in the 1st place."));
        }

        // We only append the connection type indicator if this is an attempt at redeeming an invitation.
        // If we were to append the indicator to every request, non invitation requests would get turned
        // down by the Bouncer as presently only invited connections are typed. (Might be a good idea to
        // allow the indicator as an optional hint too ...)
        if (r.endsWith("/F")) {
            VNCViewer.pen.log(Level.FINEST, logStr("Appending connection type indicator to RepMsg."));
            r = r + (isFBConnection() ? "B" : "T");
        } else
            VNCViewer.pen.log(Level.WARNING, logStr("The RepMsg didn't look like one. Connection type indicator has been omitted."));

        return r + "\n";
    }

    /**
     * @return The list of Drives presently known to exist on the server. This
     * method will return an empty list if readServerDriveList() has never been
     * invoked before and null if this is not the FT Thread. The list is
     * guaranteed to contain {@link FTPath} objects only.
     */
    public ArrayList<FTPath> getKnownServerDrives() {
        return ftu.getKnownServerDrives();
    }

    /**
     * @return The full path to the directory the Filetransfer operations will be performed in.
     */
    public String getCurrentDir() {
        return ftu.getCurrentDir();
    }

    /**
     * @return Get the name of the Sessions host.
     */
    public String getHost() {
        //return getServerParams().getHostParams().getHostname();
        return getServerParams().getHostnameAndPort(false, false, getServerParams().isRemoteSupport());
    }

    /**
     * This method is intended for windumb environments and will likely fail in *NIX environments.
     *
     * @return The Name of the drive of the current directory.
     */
    public String getCurrentDrive() {
        return ftu.getCurrentDrive();
    }

    /**
     * @return The number of files in the present remote directory.
     */
    public int getFilesCount() {
        return ftu.getFilesCount();
    }

    /**
     * @return The number of directories in the present remote directory.
     */
    public int getDirectoriesCount() {
        return ftu.getDirectoriesCount();
    }

    public RemoteFileSystem getRFS() {
        return ftu.getRFS();
    }


    private FTStatus performTransferOperation(RFBConnJob job) {

        boolean isDownload = job.isDownload;
        boolean isDirectory = job.isDirectory;

        if (isDownload && isDirectory) {
            return downloadDirectory(job.srcFile, job.dstFile);
        } else if (isDownload && !isDirectory) {
            return downloadFile(job.srcFile, job.dstFile, false);
        } else if (isDirectory) {
            return uploadDirectory(new File(job.srcFile), job.dstFile);
        } else {
            return uploadFile(job.srcFile, job.dstFile, null);
        }
    }

    /**
     * Will send a download request to the server, asking it to send the file
     * \a remoteFN. The file will then be written to \a localPath. The method will
     * overwrite existing files. If another FT operation is in progress the method
     * will return false without sending the request. The \a isDirectory parameter has to
     * be set true, if the file to be transferred is in fact a directory.
     *
     * @return True if download *request* was successfully sent.
     */
    private FTStatus downloadFile(String remoteFn, String localPath, boolean isDirectory) {
        return ftu.downloadFile(remoteFn, localPath, isDirectory);
    }

    /**
     * Will send an upload file offer to the server, asking it to write the file
     * to the path specified in \a destinationPath. The \a fakeFilename will be
     * sent to the server instead of the real filename, which is necessary to use
     * the method for directory uploads. Set the variable to null to send the
     * original filename. The method will overwrite existing files.
     *
     * @param localFn         Name of the file in the local FS.
     * @param destinationPath Desired destination directory for the file to be uploaded.
     * @param fakeFilename    will be sent to the server instead of the real filename
     * @return Returns false, without sending the offer, if another FT operation is in progress.
     */
    public FTStatus uploadFile(String localFn, String destinationPath, String fakeFilename) {
        return ftu.uploadFile(localFn, destinationPath, fakeFilename);
    }

    private FTStatus downloadDirectory(String dir2Download, String targetPath) {
        return ftu.downloadDirectory(dir2Download, targetPath);
    }

    private FTStatus uploadDirectory(File dir2Upload, String targetPath) {
        return ftu.uploadDirectory(dir2Upload, targetPath, this);
    }

    /**
     * Will send a delete file request to the server, asking it to delete the file
     * \a filename. The path must be fully qualified.
     *
     * @return Returns false, without sending the request, if another FT operation is in progress.
     */
    private FTStatus deleteRemoteFile(String filename) {
        return ftu.deleteRemoteFile(filename);
    }

    /**
     * Will send a create directory request to the server, asking it to create
     * the path specified in \a dirName. The path must be fully qualified. This is
     * NOT a create tree function.
     *
     * @return Returns false, without sending the request, if another FT operation is in progress.
     */
    private boolean createRemoteDirectory(String dirName) {
        return ftu.createRemoteDirectory(dirName);
    }

    /**
     * Convenience method, which will invoke changeServerDirectory in an
     * attempt to change to directory \a relDir, which it assumes to be
     * a directory relative to the current directory.
     */
    private void navigateTo(String relDir) {
        ftu.navigateTo(relDir);
    }

    /**
     * Will change the current directory to \a dir. A fully qualified Path is expected in \a dir.
     *
     * @return Returns false, without sending the request, if another FT operation is in progress.
     */
    private boolean changeServerDirectory_wrapper(String dir) {
        return ftu.changeServerDirectory(dir);
    }

    /**
     * Will send a "list drives" request to the server.
     *
     * @return Returns false, without sending the request, if another FT operation is in progress.
     */
    private boolean readServerDriveList() {
        return ftu.readServerDriveList();
    }

    private void sendMouseEvent(MouseEvent me) {
        bufEvents.writePointerEvent(me);
    }

    private void sendKeyPress(KeyEvent ke) {
        bufEvents.writeKeyEvent(ke, getSession().getConfiguration().getUsedKeyboardLayout());
    }

    /**
     * This method is supposed to send a Key stroke (that's essentially a key press
     * complete with modifiers keys) to the server. Alas the method does not work
     * properly. Developers are advised to use this method with <b>extreme care</b>
     * as results may be heretic. At the moment, it is only used to send Alt+Tab
     * keystrokes to servers.
     */
    private void sendKeyStroke(Component source, KeyStroke ks) {
        if (ks == null) {
            VNCViewer.pen.log(Level.SEVERE, logStr("Can't send [NULL] KeyStroke!"));
            return;
        }

        Date d = new Date();
        int modifiers = ks.getModifiers();
        int keycode = ks.getKeyCode();
        char keychar = ks.getKeyChar();
        KeyEvent.getKeyText(keycode);

        KeyEvent kevt = new KeyEvent(source, KeyEvent.KEY_PRESSED, d.getTime(), modifiers, keycode, keychar);
        bufEvents.writeKeyEvent(kevt, getSession().getConfiguration().getUsedKeyboardLayout());

        d = new Date();
        kevt = new KeyEvent(source, KeyEvent.KEY_RELEASED, d.getTime(), modifiers, keycode, keychar);
        bufEvents.writeKeyEvent(kevt, getSession().getConfiguration().getUsedKeyboardLayout());
    }

    private void sendMonkeyGrip(Component source) {
        Date d = new Date();
        final int modifiers = InputEvent.CTRL_MASK | InputEvent.ALT_MASK;
        KeyEvent ctrlAltDelEvent = new KeyEvent(source, KeyEvent.KEY_PRESSED, d.getTime(), modifiers, 127, KeyEvent.CHAR_UNDEFINED);
        bufEvents.writeKeyEvent(ctrlAltDelEvent, getSession().getConfiguration().getUsedKeyboardLayout());

        d = new Date();
        ctrlAltDelEvent = new KeyEvent(source, KeyEvent.KEY_RELEASED, d.getTime(), modifiers, 127, KeyEvent.CHAR_UNDEFINED);
        bufEvents.writeKeyEvent(ctrlAltDelEvent, getSession().getConfiguration().getUsedKeyboardLayout());
    }

    private void sendMetaKeypress(Component source) {
        Date d = new Date();
        KeyEvent winMenuEvent = new KeyEvent(source, KeyEvent.KEY_PRESSED, d.getTime(), KeyEvent.VK_UNDEFINED, KeyEvent.VK_WINDOWS, KeyEvent.CHAR_UNDEFINED);
        bufEvents.writeKeyEvent(winMenuEvent, getSession().getConfiguration().getUsedKeyboardLayout());

        d = new Date();
        winMenuEvent = new KeyEvent(source, KeyEvent.KEY_RELEASED, d.getTime(), KeyEvent.VK_UNDEFINED, KeyEvent.VK_WINDOWS, KeyEvent.CHAR_UNDEFINED);
        bufEvents.writeKeyEvent(winMenuEvent, getSession().getConfiguration().getUsedKeyboardLayout());
    }

    private void sendScreenRefresh() {
        try {
            fbu.requestFullFBUpdate();
        } catch (IOException e) {
            VNCViewer.pen.log(Level.SEVERE, logStr("Failed to send a screen refresh request to the server."), e);
        }
    }

    public void tryToAbortFileTransfer() {
        ftu.tryToAbortFileTransfer();
    }

    /**
     * Warning! Experimental method. Do not call this one unless u know what u do!
     *
     * @return false if something went wrong. Chances are, it will!
     */
    private boolean sendServerBlankConsoleRequest(boolean powerDown) {
        // A RFB blank remote Console request may only be sent from the FB
        // connection. In FT mode there's nothing 4 the user 2 b seen anyway.
        if (!isFBConnection()) return false;

        byte[] b = new byte[4];

        b[0] = (byte) RFB.ServerInputUpdateRequest;
        b[1] = (byte) (powerDown ? 1 : 0);
        b[2] = (byte) 0;
        b[3] = (byte) 0;

        try {
            os.write(b);
            String msg = powerDown ? "The remote computer was asked to power down its monitor." :
                    "The remote computer was asked to switch its monitor on again.";
            VNCViewer.pen.log(Level.CONFIG, logStr(msg));
            fireEvent(RFBHandlerEvent.consoleshutdown(this), null);
            return true;
        } catch (IOException e) {
            VNCViewer.pen.log(Level.WARNING, logStr("Failed to send blank console request to server."), e);
            return false;
        }
    }

    private void doProtocolInitialisation() throws IOException {
        writeClientInit();
        readServerInit();
        setEncodings();
    }

    /**
     * Will send a message to the server containing a list of features which
     * are supported by the client. This list is being sent during the handshake
     * phase and later whenever the user wants a different pixel encoding type
     * to be used.
     */
    private void setEncodings() {
        if (phase != Phase.PROCESSING) return;

        try {
            SettingsCN o = getSettingsCN();
            if (o == null) {
                VNCViewer.pen.log(Level.WARNING, logStr("No SettingsCN found when trying to write the encodings"));
                return;
            }

            Vector<VNCEncoding> encs = o.preferredEncodings.getEncodings(true, isFBConnection());
            int nEncodings = encs.size();
            byte[] b = new byte[4 + 4 * nEncodings];

            b[0] = (byte) RFB.SetEncodings;
            b[1] = (byte) 0; // Padding
            b[2] = (byte) ((nEncodings >> 8) & 0xff);
            b[3] = (byte) (nEncodings & 0xff);

            int j = 0;
            Iterator<VNCEncoding> i = encs.iterator();
            while (i.hasNext()) {
                VNCEncoding e = i.next();
                int encodingEnum = e.toInt();
                //VncViewer.pen.log( Level.FINEST, logStr( "Writing encoding [ " + e.toString() + " / " + encodingEnum + "]"));
                for (int k = 0; k <= 3; k++)
                    b[4 + k + 4 * j] = (byte) ((encodingEnum >> 24 - k * 8) & 0xff);
                j++;
            }

            os.write(b);
            //VncViewer.pen.log( Level.FINEST, logStr( "The encodings list has been sent to the server."));
            // if (vc != null) vc.softCursorFree();
        } catch (Exception e) {
            VNCViewer.pen.log(Level.WARNING, logStr("An exception occurred while setting the encodings."), e);
        }
    }

    private void writeClientInit() throws IOException {
        VNCViewer.pen.log(Level.FINEST, logStr("Writing client init."));
        os.write(1);
        //os.write( op.wants2ShareDesktop() ? 1 : 0);
    }

    private void readServerInit() throws IOException {
        VNCViewer.pen.log(Level.FINEST, logStr("Reading server init."));

        int fbWidth = is.readUnsignedShort();
        int fbHeight = is.readUnsignedShort();

        // This PixelFormat constructor will read from the sacred IS.
        // It may do so because this method is private and will only
        // get invoked in due course of THIS thread.
        PixelFormat pfProposedByServer = new PixelFormat(is);

        byte[] pad = new byte[3];
        is.readFully(pad);

        int nameLength = is.readInt();
        byte[] name = new byte[nameLength];
        is.readFully(name);
        String desktopName = new String(name, RFB.DEFAULT_CHARSET); // US-ASCII as per the protocol

        VNCViewer.pen.log(Level.FINEST, logStr("Read server init. => Marking this connection as running."));

        // Only as of this moment, we can be sure we have a working connection
        // with the VNC server. Before that we could have misinterpreted a
        // servers "silence" as approval of our connection attempts. It is here
        // the server has to answer us or hold his silence forever.
        setPhase(Phase.PROCESSING);

        // This is not a goof. Finer messages are relayed to the Web-Logger!
        VNCViewer.pen.log(Level.FINER, logStr("Connected to [" + getHost() + "] [" + desktopName + "]"));

        // OUCH! Didn't see that one coming :(
        fbu.setRDProperties(pfProposedByServer, new Dimension(fbWidth, fbHeight), desktopName, isFBConnection());
    }

    @SuppressWarnings("unused")
    private String getDBGModifiers(KeyEvent e) {
        String evtM = "";

        if (e.isControlDown()) evtM += "Ctrl|";
        if (e.isShiftDown()) evtM += "Shift|";
        if (e.isMetaDown()) evtM += "Meta|";
        if (e.isAltDown()) evtM += "Alt|";
        if (e.isAltGraphDown()) evtM += "AltGr";

        return evtM;
    }

    @SuppressWarnings("unused")
    private String getDBGModifiers(int m) {
        String evtM = "";
        if ((m & InputEvent.CTRL_MASK) != 0) evtM += "Ctrl|";
        if ((m & InputEvent.SHIFT_MASK) != 0) evtM += "Shift|";
        if ((m & InputEvent.META_MASK) != 0) evtM += "Meta|";
        if ((m & InputEvent.ALT_MASK) != 0) evtM += "Alt|";
        if ((m & InputEvent.ALT_GRAPH_MASK) != 0) evtM += "AltGr";
        return evtM;
    }

    private void sendToRemoteClipboard(String text) {
        if (phase != Phase.PROCESSING || text.equals(clipboardText)) return;

        byte[] textBytes = text.getBytes(getPresumedEncoding());
        byte[] b = new byte[8 + textBytes.length];

        b[0] = (byte) RFB.ClientCutText;
        b[4] = (byte) ((text.length() >> 24) & 0xff);
        b[5] = (byte) ((text.length() >> 16) & 0xff);
        b[6] = (byte) ((text.length() >> 8) & 0xff);
        b[7] = (byte) (text.length() & 0xff);

        System.arraycopy(textBytes, 0, b, 8, text.length());

        try {
            os.write(b);
            clipboardText = text;
        } catch (IOException e) {
            VNCViewer.pen.log(Level.SEVERE, logStr("Failed to read clipboard contents from the server."), e);
        }
    }

    /**
     * @return Returns true, if the Connection with the server is beyond repair and any
     * attempt at sending something to the server or waiting for an answer is pointless.
     */
    public boolean isDeadInTheWater() {
        return isThreadDoomed || phase != Phase.PROCESSING;
    }

    public Phase getPhase() {
        return phase;
    }

    private void setPhase(Phase p) {
        if (phase == p) return;
        Phase oldPhase = phase;
        phase = p;
        VNCViewer.pen.log(Level.FINEST, logStr("Entered phase [" + phase.toString() + "]"));
        fireEvent(RFBHandlerEvent.phasechanged(this, oldPhase, phase), null);
    }

    private void crash(String hrErrMsgIn, Throwable cause) {
        //stErrorMsg = stErrMsgIn;
        causeOfCrash = cause;
        humanreadableErrorMsg = hrErrMsgIn;

        String develreadableErrorMsg = cause != null ? cause.getLocalizedMessage() :
                "It would seem no 'cause' (object) came with this crash ??!";
        // This is not a goof. Finer messages are relayed to the Web-Logger!
        VNCViewer.pen.log(Level.FINER, logStr(hrErrMsgIn));

        fireEvent(RFBHandlerEvent.disconnect(this, false), develreadableErrorMsg);
        return;
    }

    // implementation of the ClipboardOwner Interface
    public void lostOwnership(Clipboard aClipboard, Transferable aContents) {
    }

    public String getClipboardText() {
        return clipboardText;
    }

    private void setClipboardContents(String clipIn) {
        clipboardText = clipIn;
    }

    public void fireEvent(RFBHandlerEvent e, String errmsg) {
        e.success = errmsg == null;
        //e.errMsg = errmsg;
        setChanged();
        notifyObservers(e);
    }

    public void update(Observable o, Object arg) {
        SettingsCN connectionSettings = getSettingsCN();

        if (o == connectionSettings) {
            SettingsCN.VNCProperty p = (SettingsCN.VNCProperty) arg;

            if (p == SettingsCN.VNCProperty.PIXENCODING) queue(RFBConnJob.reencode(this));
            if (p == SettingsCN.VNCProperty.EIGHTBIT) queue(RFBConnJob.changePixelFormat(this));
        }

        // Here we forward VNCConnectionEvents coming from the FT/FB unit.
        if (o == ftu || o == fbu) {
            setChanged();
            notifyObservers(arg);
        }
    }

    public boolean isDisabledRemoteControlling() {
        return disabledRemoteControlling;
    }

    public void setDisabledRemoteControlling(boolean disabledRemoteControlling) {
        this.disabledRemoteControlling = disabledRemoteControlling;
    }
}
