package com.wincornixdorf.vnc.rfb;

import java.awt.event.KeyEvent;
import java.util.HashMap;

import com.wincornixdorf.vnc.global.VNCViewer;

/**
 * Created by jan.fogl.
 */
public class X11Mapper {

    private static boolean traceKeybd = VNCViewer.isDivine() && System.getenv("TRACEKEYBD") != null;

    private HashMap<Integer, Integer> keycode2x11 = new HashMap<Integer, Integer>();

    private HashMap<Integer, Integer> keycharExceptions = new HashMap<Integer, Integer>();


    public X11Mapper() {
        init();
    }

    private void init() {

        keycode2x11.put(KeyEvent.VK_HOME, 0xff50);
        keycode2x11.put(KeyEvent.VK_LEFT, 0xff51);
        keycode2x11.put(KeyEvent.VK_UP, 0xff52);
        keycode2x11.put(KeyEvent.VK_RIGHT, 0xff53);
        keycode2x11.put(KeyEvent.VK_DOWN, 0xff54);
        keycode2x11.put(KeyEvent.VK_PAGE_UP, 0xff55);
        keycode2x11.put(KeyEvent.VK_PAGE_DOWN, 0xff56);
        keycode2x11.put(KeyEvent.VK_END, 0xff57);
        keycode2x11.put(KeyEvent.VK_BEGIN, 0xff58);

        keycode2x11.put(KeyEvent.VK_INSERT, 0xff63);
        keycode2x11.put(KeyEvent.VK_NUM_LOCK, 0xff7f);

        keycode2x11.put(KeyEvent.VK_F1, 0xffbe);
        keycode2x11.put(KeyEvent.VK_F2, 0xffbf);
        keycode2x11.put(KeyEvent.VK_F3, 0xffc0);
        keycode2x11.put(KeyEvent.VK_F4, 0xffc1);
        keycode2x11.put(KeyEvent.VK_F5, 0xffc2);
        keycode2x11.put(KeyEvent.VK_F6, 0xffc3);
        keycode2x11.put(KeyEvent.VK_F7, 0xffc4);
        keycode2x11.put(KeyEvent.VK_F8, 0xffc5);
        keycode2x11.put(KeyEvent.VK_F9, 0xffc6);
        keycode2x11.put(KeyEvent.VK_F10, 0xffc7);
        keycode2x11.put(KeyEvent.VK_F11, 0xffc8);
        keycode2x11.put(KeyEvent.VK_F12, 0xffc9);

        keycode2x11.put(KeyEvent.VK_WINDOWS, 0xffeb);
        keycode2x11.put(KeyEvent.VK_BACK_SPACE, 0xff08);
        keycode2x11.put(KeyEvent.VK_TAB, 0xff09);
        keycode2x11.put(KeyEvent.VK_ENTER, 0xff0d);
        keycode2x11.put(KeyEvent.VK_WINDOWS, 0xffeb);
        keycode2x11.put(KeyEvent.VK_ESCAPE, 0xff1b);
        keycode2x11.put(KeyEvent.VK_DELETE, 0xffff);
        keycode2x11.put(KeyEvent.VK_DEAD_GRAVE, 0xfe50);
        keycode2x11.put(KeyEvent.VK_DEAD_ACUTE, 0xfe51);
        keycode2x11.put(KeyEvent.VK_DEAD_CIRCUMFLEX, 0xfe52);
        keycode2x11.put(KeyEvent.VK_EURO_SIGN, 0x20ac);
        keycode2x11.put(KeyEvent.VK_AT, 0x0040);

        keycode2x11.put(KeyEvent.VK_NUMPAD0, 0xffb0);
        keycode2x11.put(KeyEvent.VK_NUMPAD1, 0xffb1);
        keycode2x11.put(KeyEvent.VK_NUMPAD2, 0xffb2);
        keycode2x11.put(KeyEvent.VK_NUMPAD3, 0xffb3);
        keycode2x11.put(KeyEvent.VK_NUMPAD4, 0xffb4);
        keycode2x11.put(KeyEvent.VK_NUMPAD5, 0xffb5);
        keycode2x11.put(KeyEvent.VK_NUMPAD6, 0xffb6);
        keycode2x11.put(KeyEvent.VK_NUMPAD7, 0xffb7);
        keycode2x11.put(KeyEvent.VK_NUMPAD8, 0xffb8);
        keycode2x11.put(KeyEvent.VK_NUMPAD9, 0xffb9);

        keycharExceptions.put(8364, 0x0065);    //Euro sign 101 = VK_NUMPAD5
        keycharExceptions.put(176, 0x00b0);    //Degree ° sign
        keycharExceptions.put(176, 0x0060);    //Degree ° sign
        keycharExceptions.put(0x00A2, 0x00a2);    //Degree ° sign

    }

    public int keycode2x11(int keyCode, int keyChar) {
        Integer exception = keycharExceptions.get(keyChar);

        if (exception != null) {
            if (traceKeybd)
                System.err.println("KBDTrace KBDEvs: Exceptional character input detected. Casting KeyChar (" + keyChar + " | " + "0x" + Integer.toHexString(keyChar) +
                        ") → (" + keyCode + " | " + "0x" + Integer.toHexString(keyCode) + ")");
            return exception.intValue();
        }

        Integer x11KeySym = keycode2x11.get(keyCode);
        int rVal = x11KeySym == null ? keyChar : x11KeySym.intValue();

        if (traceKeybd)
            System.err.println("KBDTrace KBDEvs: KeyCode (" + keyCode + " | " + "0x" + Integer.toHexString(keyCode) +
                    ") KeyChar (" + keyChar + " | " + "0x" + Integer.toHexString(keyChar) +
                    ") → (" + rVal + " | " + "0x" + Integer.toHexString(rVal) + ")");
        return rVal;
    }

}
