package com.wincornixdorf.vnc.rfb.conf;

import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.util.Vector;

/**
 * Stores all information available about a VNC server.
 */
//TODO: Give this a COMPLETE overhaul ... this code STINKS!
//TODO: Also rename this class to VNCConnectionDetails or VNCServerDetails
public class VNCSrvCfg {

    // For now SSL and TLS are considered equal in that they do the same
    // no distinction will be made between the two, which in my opinion
    // is potentially misleading.
    public enum ConMode {
        PLAIN, SSL, TLS, HTTP, HTTPS, FILE
    }

    ;

    private static final String SINGLE_CLICK_PREFIX = "ID:";

    private static final Object repeaterMsgFullLengthPlaceholder = ".kone";

    private static String PURL_FORMAT = "<PROTOCOL>://USER@HOST:PORT[::DESKTOPNO][?REPEATERHOST:REPEATERPORT][;LABEL]\n" +
            "Supported protocols: [" + ConMode.PLAIN.toString() + ","
            + ConMode.TLS.toString() + ","
            + ConMode.SSL.toString() + ","
            + ConMode.HTTPS.toString() + ","
            + ConMode.HTTPS.toString() + ","
            + ConMode.FILE.toString() + "]";

    public String filename = "";
    private LoginParameters lpSrv = new LoginParameters(isRemoteSupport());
    private LoginParameters lpRep = new LoginParameters(isRemoteSupport());

    public boolean supportsFT = true;

    private ConMode conMode = ConMode.TLS;

    // TODO: Storing protocol relevant data here is a bad idea. This is a supposed to be a configuration container.
    private int serverMajor = -1;
    private int serverMinor = -1;
    private int clientMajor = -1;
    private int clientMinor = -1;

    private String repeaterMsg = "";

    /**
     * The desktop number is presently unused. It will probably be discontinued
     * soon. Turns out the protocol has no notion of multiple screens anyway
     * and the UltraVNC Server, which this Applet is optimized for, supports
     * multiple screens by creating one large representation thereof.
     */
    private String desktopNo = null;

    private String label = "unknown";
    private String originalURL = null;
    private String destIP = "";


    public VNCSrvCfg(String pseudoURL, String repMsg, String password) throws MalformedURLException {
        setRepeaterMsg(repMsg);
        configure(pseudoURL, password);
    }

    public VNCSrvCfg(String srvHostname, int srvPort,
                     String repHostname, int repPort,
                     String username, String repMsg, String password) {
        setRepeaterMsg(repMsg);
        lpSrv = new LoginParameters(true, srvHostname, srvPort, username, password, isRemoteSupport());
        lpRep = new LoginParameters(false, repHostname, repPort, "", repMsg, false);
    }

    /**
     * This method will parse the String argument in an effort to draw the
     * parameters required to set up a connection with a particular host from
     * it. These parameters will be written into member variables and can then
     * be retrieved through the getters and setters below. As one can imagine,
     * this method is intended to be used as a pseudo constructor.
     * <p>
     * <p>
     * The pseudo URL is expected to abide by the following syntax: session
     * '<PROTOCOL>://USER@HOSTNAME:PORT[::DESKTOPNO][?REPEATERHOST:REPEATERPORT][;LABEL]
     * <p>
     *
     * @param pseudoUrl A URL like String to be parsed.
     *                  <p>
     *                  Supported protocols are:
     *                  <p>
     *                  <ul>
     *                  <li>PLAIN plain socket connection. No encryption
     *                  <li>TLS TLS/SSL secured socket.
     *                  <li>HTTP Tunnel via HTTP. At the moment only possible via the
     *                  WNXVNCRepeater.
     *                  <li>HTTPS Tunnel via HTTPS. At the moment only possible via
     *                  the WNXVNCRepeater.
     *                  <li>FILE Reads the stream from a local file.
     *                  </ul>
     *                  <p>
     *                  The USER should be the name of the account the viewer is to
     *                  attempt logging into.
     *                  <p>
     *                  The HOSTNAME:PORT and REPEATERHOST:REPEATERPORT parameters
     *                  should be self explaining.
     *                  <p>
     *                  The desktop number will be ignored for now. Future versions
     *                  may support it.
     *                  <p>
     *                  The LABEL will be used for display purposes only and gets set
     *                  to hostname if omitted.
     * @throws MalformedURLException In the event of grievous format errors, which
     *                               would otherwise lead to or have been the cause of ArrayOutOfBound
     *                               exceptions. This will likely change in the future (I hope).
     */
    private void configure(String pseudoUrl, String pwd)
            throws MalformedURLException {
        try {
            if (pseudoUrl == null || pseudoUrl.isEmpty()) return;

            originalURL = pseudoUrl;

            int idxProtocolSep = pseudoUrl.indexOf("://");
            int idxDsktpNumSep = pseudoUrl.indexOf("::");
            int idxRepeaterSep = pseudoUrl.indexOf("?");
            int idxLabelExtSep = pseudoUrl.indexOf(";");

            if (idxProtocolSep == -1)
                throw new MalformedURLException("Protocol separator '://' missing!\n" + PURL_FORMAT);

            boolean dsktpNumWasSet = idxDsktpNumSep != -1;
            boolean repeaterWasSet = idxRepeaterSep != -1;
            boolean labelExtWasSet = idxLabelExtSep != -1;

            //* '<PROTOCOL>://USER@HOSTNAME:PORT[::DESKTOPNO][?REPEATERHOST:REPEATERPORT][;LABEL]<p>
            int idxUsrHostPEnd = pseudoUrl.length();
            int idxRepHostPEnd = pseudoUrl.length();
            int idxDsktpNumEnd = pseudoUrl.length();

            if (labelExtWasSet) {
                idxUsrHostPEnd = idxLabelExtSep;
                idxRepHostPEnd = idxLabelExtSep;
                idxDsktpNumEnd = idxLabelExtSep;
            }

            if (repeaterWasSet) {
                idxUsrHostPEnd = idxRepeaterSep;
                idxDsktpNumEnd = idxRepeaterSep;
            }

            // If the desktop number was set, the user/host/port substring ends there of course.
            if (dsktpNumWasSet) idxUsrHostPEnd = idxDsktpNumSep;

            // Here the protocol to be used is extracted.
            String protocol = pseudoUrl.substring(0, idxProtocolSep);

            try {
                conMode = ConMode.valueOf(protocol);
            } catch (IllegalArgumentException e) {
                throw new MalformedURLException("[" + protocol + "] is NOT a supported Protocol!\n" + PURL_FORMAT);
            } catch (NullPointerException npe) {
                throw new MalformedURLException("No protocol was specified.\n" + PURL_FORMAT);
            }

            if (conMode == ConMode.FILE) {
                filename = pseudoUrl.substring(idxProtocolSep + 3);
                return;
            }

            String usrHostPort = pseudoUrl.substring(idxProtocolSep + 3, idxUsrHostPEnd);
            desktopNo = dsktpNumWasSet ? pseudoUrl.substring(idxUsrHostPEnd + 2, idxDsktpNumEnd) : "";
            String repAndport = repeaterWasSet ? pseudoUrl.substring(idxRepeaterSep + 1, idxRepHostPEnd) : "";

            int idxUserHostSep = usrHostPort.indexOf("@");
            idxDsktpNumSep = usrHostPort.indexOf("::");
            int idxPortnum = usrHostPort.indexOf(":");

            if (idxUserHostSep == -1)
                throw new MalformedURLException("Username Hostname separator '@' missing!\n" + PURL_FORMAT);

            if (idxPortnum == -1)
                throw new MalformedURLException("Hostname Port# separator ':' missing!\n" + PURL_FORMAT);

            String hostname = usrHostPort.substring(idxUserHostSep + 1, idxPortnum);
            String usrname = usrHostPort.substring(0, idxUserHostSep);

            int portno = toInt(usrHostPort.substring(idxPortnum + 1), -1);

            label = labelExtWasSet ? pseudoUrl.substring(idxLabelExtSep + 1) : hostname;

            lpSrv = new LoginParameters(true, hostname, portno, usrname, pwd, isRemoteSupport());

            if (repeaterWasSet) {
                idxRepeaterSep = repAndport.indexOf(":");
                String rephostname = repAndport.substring(0, idxRepeaterSep);
                int repportno = toInt(repAndport.substring(idxRepeaterSep + 1), -1);
                lpRep = new LoginParameters(false, rephostname, repportno, "", "", false);
            }
        } catch (Exception e) {
            throw new MalformedURLException("General Error trying to parse the URL!\n" + PURL_FORMAT + "\n" + e.toString());
        }
    }

    /**
     * Builds or returns a string that looks quite a lot like a URL but will likely
     * violate important rules for URLs. It is designed to be easily readable and
     * present key parameters of VNC connections to advanced users. Please mind
     * this is a "Pseudo-URL". It is NOT fit to be used anywhere else but in this viewer.
     * Please see @link configure( String) for details<p>
     *
     * @return The pseudoURL used to set up this object or a pseudoURL generated
     * from the values presently set in this object
     **/
    public String getPseudoURL() {
        if (originalURL != null) return originalURL;

        String portNo = lpSrv.getPort() != -1 ? ":" + lpSrv.getPort() : "";
        String deskNo = desktopNo != null ? "::" + desktopNo : "";
        String rs = conMode + "://" + lpSrv.getUsername() + "@" + lpSrv.getOriginalHostname() + portNo + deskNo;
        String repHaP = !requiresRepeaterToConnect() ? "" : "?" + getHostnameAndPort(true, true);

        return rs + repHaP + ";" + label;
    }

    public LoginParameters getHostParams(boolean repeater) {
        return repeater ? lpRep : lpSrv;
    }

    public LoginParameters getHostParams() {
        return lpSrv;
    }

    public LoginParameters getRepeaterParams() {
        return lpRep;
    }

    public LoginParameters get1stHopParams() {
        return getHostParams(requiresRepeaterToConnect());
    }

    public ConMode getConMode() {
        return conMode;
    }

    public void setConMode(ConMode m) {
        conMode = m;
    }

    public boolean isConfigured() {
        return conMode == ConMode.FILE || (lpSrv.isInitialized() && lpSrv.hasAUsername());
    }

    public String getUsername() {
        return lpSrv.getUsername() != null ? lpSrv.getUsername() : "";
    }

    public String getLabel() {
        return label == null ? "unknown" : label;
    }

    public void setLabel(String l) {
        if (l != null && !l.isEmpty()) label = l;
    }

    public Vector<InetSocketAddress> get1stHops() {
        LoginParameters p = get1stHopParams();
        return p != null ? p.getPotentialInetAdresses() : new Vector<InetSocketAddress>();
    }

    public String getHostnameAndPort(boolean repeater, boolean protocol) {
        return getHostnameAndPort(repeater, protocol, false);
    }

    public String getHostnameAndPort(boolean repeater, boolean protocol, boolean remoteSupport) {
        LoginParameters p = getHostParams(repeater);
        if (p == null) return "ERROR";

        String result;
        if (remoteSupport) {
            result = p.getSessionId();
        } else if (protocol) {
            result = p.getHostSpecial() + ":" + p.getPort();
        } else {
            result = p.getHostAddress().toString();
        }
        return result;

    }

    public void setVersion(boolean server, int major, int minor) {
        if (server) {
            serverMajor = major;
            serverMinor = minor;
        } else {
            clientMajor = major;
            clientMinor = minor;
        }
    }

    public int getVersionMajor(boolean server) {
        return server ? serverMajor : clientMajor;
    }

    public void setVersionMajor(boolean server, int major) {
        if (server) serverMajor = major;
        else clientMajor = major;
    }

    public int getVersionMinor(boolean server) {
        return server ? serverMinor : clientMinor;
    }

    public void setVersionMinor(boolean server, int minor) {
        if (server) serverMinor = minor;
        else clientMinor = minor;
    }

    /**
     * This method serves debugging purposes only. This is method does <b>NOT</b>
     * generate the version string to be sent to the server during the handshake phase.
     *
     * @param server If true, the protocol version of the server will be
     *               recreated. If false, the protocol version desired by the client.
     * @return A string reading "RFB {major}.{minor} where major and minor are
     * the those used by the respective party. In case the handshake did not
     * take place, "UNKNOWN" will be returned instead.
     */
    public String getProtocolVersion(boolean server) {
        int major = getVersionMajor(server);
        int minor = getVersionMinor(server);

        if (major < 0 || minor < 0) return "UNKNOWN";
        return String.format("RFB %1$d.%2$d", major, minor);
    }

    /**
     * @return True, if the server can only be accessed via a repeater.
     */
    public boolean requiresRepeaterToConnect() {
        return lpRep.isInitialized();
    }

    /**
     * @return True, if the server can only be accessed via a HTTP or HTTPS tunnel.
     */
    public boolean requiresTunnelToConnect() {
        return conMode == ConMode.HTTP || conMode == ConMode.HTTPS;
    }

    /**
     * @return The "Repeater message", which is to be sent on encountering a Repeater
     * instead of the usual host:port request.
     */
    public String getRepeaterMsg() {
        return repeaterMsg;
    }

    private void setRepeaterMsg(String s) {
        // This is to make sure the repeaterMsg value will never be null
        repeaterMsg = s != null ? s : "";
    }

    public String getRepeaterMsgFullLength(int length) {
        StringBuilder fullMsg = new StringBuilder(repeaterMsg);
        while (fullMsg.length() < length) {
            fullMsg.append(repeaterMsgFullLengthPlaceholder);
        }
        return fullMsg.toString().substring(0, length);
    }

    public String getDestIP() {
        return destIP;
    }

    public void setDestIP(String destIP) {
        this.destIP = destIP;
    }

    /**
     * @return true if applet is launched for Remote Support (Single Click - SCIII), false if applet is launched for Remote Control
     */
    public boolean isRemoteSupport() {
        return repeaterMsg != null && repeaterMsg.startsWith(SINGLE_CLICK_PREFIX);
    }

    private int toInt(String str, int defaultValue) {
        int result = defaultValue;
        if (str != null) {
            try {
                result = Integer.parseInt(str);
            } catch (NumberFormatException e) {
            }
        }
        return result;
    }

    public String toString() {
        return getPseudoURL();
    }

    public String toUserFriendlyString() {
        if (lpSrv.isInitialized()) return "This VNC connection is unusable as it doesn't specify a valid host.";

        StringBuffer sb = new StringBuffer(250);

        sb.append("Connection details:\n");
        sb.append("===================\n");
        sb.append("Protocol  : [" + conMode + "]\n");
        sb.append("Host      : [" + lpSrv.getOriginalHostname() + "]\n");
        sb.append("Port      : [" + lpSrv.getPort() + "]\n");
        sb.append("User      : [" + lpSrv.getUsername() + "]\n");
        sb.append("Password  : [ You wish ;) ]\n");

        String dc = "N/A";
        String repHost = !lpRep.isInitialized() ? dc : lpRep.getOriginalHostname();
        String repPort = !lpRep.isInitialized() ? dc : String.valueOf(lpRep.getPort());

        sb.append("Rep. Host : [" + repHost + "]\n");
        sb.append("Rep. Port : [" + repPort + "]");

        return sb.toString();
    }
}
