package com.wincornixdorf.vnc.rfb.decoder;

import java.awt.Color;
import java.io.DataInputStream;
import java.io.IOException;

import com.wincornixdorf.vnc.rfb.RFB;
import com.wincornixdorf.vnc.rfb.fb.RemoteDesktop;
import com.wincornixdorf.vnc.rfb.fb.VNCEncoding;

public class FBDecoderHextile extends FBDecoder {
    /**These colors should be kept between handleHextileSubrect() calls.*/
    private Color hextile_bg    = new Color( 0);
    private Color hextile_fg    = new Color( 0);

    public FBDecoderHextile( RemoteDesktop data) {
        super( data);
    }

    public VNCEncoding getEncoding() {
        return VNCEncoding.Hextile;
    }

    public void process( DataInputStream is, int x, int y, int w, int h) throws Exception {
        hextile_bg = new Color( 0);
        hextile_fg = new Color( 0);

        for ( int ty = y; ty < y + h; ty += 16) {
            int th = 16;
            if ( y + h - ty < 16)
                th = y + h - ty;

            for ( int tx = x; tx < x + w; tx += 16) {
                int tw = 16;
                if ( x + w - tx < 16)
                    tw = x + w - tx;

                handleHextileSubrect( is, tx, ty, tw, th);
            }
        }
        rd.srcRDesktop.newPixels( x, y, w, h);
    }

    //
    // Handle one tile in the Hextile-encoded data.
    //

    void handleHextileSubrect( DataInputStream is, int tx, int ty, int tw, int th) throws IOException {
        int subencoding = is.readUnsignedByte();
        //if ( rfb.rec != null)
        //    rfb.rec.writeByte( subencoding);
        int bpp = rd.getBytesPerPixel();

        // Is it a raw-encoded sub-rectangle?
        if (( subencoding & RFB.HextileRaw) != 0) {
            processRAW( is, tx, ty, tw, th, bpp);
            return;
        }

        // Read and draw the background if specified.
        byte[] cbuf = new byte[ bpp];
        if (( subencoding & RFB.HextileBackgroundSpecified) != 0) {
            is.readFully( cbuf);

            switch ( bpp) {
            	case 1  : hextile_bg = new Color( rd.cm08.getRGB( cbuf[ 0] & 0xFF)); break;
                case 2  : hextile_bg = decode16BitColor( cbuf, 0); break;
                default : hextile_bg = new Color( crunch( cbuf)); break;
            }

            //if ( rfb.rec != null)
            //    rfb.rec.write( cbuf);
        }

        rd.fillRect( tx, ty, tw, th, hextile_bg);

        // Read the foreground color if specified.
        if (( subencoding & RFB.HextileForegroundSpecified) != 0) {
            is.readFully( cbuf);

            switch ( bpp) {
            	case 1  : hextile_fg = new Color( rd.cm08.getRGB( cbuf[ 0] & 0xFF)); break;
                case 2  : hextile_fg = decode16BitColor( cbuf, 0); break;
                default : hextile_fg = new Color( cbuf[ 2] & 0xFF, cbuf[ 1] & 0xFF, cbuf[ 0] & 0xFF); break;
            }

            //if ( rfb.rec != null)
            //    rfb.rec.write( cbuf);
        }

        // Done with this tile if there is no sub-rectangles.
        if (( subencoding & RFB.HextileAnySubrects) == 0) return;

        int nSubrects = is.readUnsignedByte();
        int bufsize = nSubrects * 2;

        if (( subencoding & RFB.HextileSubrectsColoured) != 0)
            bufsize += nSubrects * bpp;

        byte[] buf = new byte[ bufsize];
        is.readFully( buf);

        //if (rfb.rec != null) {
        //    rfb.rec.writeByte(nSubrects);
        //    rfb.rec.write(buf);
        //}

        int b1, b2, sx, sy, sw, sh;
        int i = 0;

        if (( subencoding & RFB.HextileSubrectsColoured) == 0) {
            // Sub-rectangles are all of the same color.
            for ( int j = 0; j < nSubrects; j++) {
                b1 = buf[ i++] & 0xFF;
                b2 = buf[ i++] & 0xFF;
                sx = tx + ( b1 >> 4);
                sy = ty + ( b1 & 0xf);
                sw = ( b2 >> 4) + 1;
                sh = ( b2 & 0xf) + 1;
                rd.fillRect( sx, sy, sw, sh, hextile_fg);
            }
            return;
        }

        for ( int j = 0; j < nSubrects; j++) {

            switch ( bpp) {
                case 1  : hextile_fg = new Color( rd.cm08.getRGB( buf[ i++] & 0xFF)); break;               
                case 2  : hextile_fg = decode16BitColor( buf, i); i += 2; break;
                default : hextile_fg = new Color( buf[ i + 2] & 0xFF, buf[ i + 1] & 0xFF, buf[ i] & 0xFF); i += 4; break;
            }

            b1 = buf[ i++] & 0xFF;
            b2 = buf[ i++] & 0xFF;
            sx = tx + ( b1 >> 4);
            sy = ty + ( b1 & 0xf);
            sw = ( b2 >> 4) + 1;
            sh = ( b2 & 0xf) + 1;
            rd.fillRect( sx, sy, sw, sh, hextile_fg);
        }
    }
}
