package com.wincornixdorf.vnc.rfb.fb;

import java.lang.Cloneable;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.nio.ByteOrder;

import com.wincornixdorf.vnc.rfb.RFB;

public class PixelFormat extends Object implements Cloneable {

    public enum PixelFormatQuality {
        EMPTY(-1, PixelFormat.defaultLoColor),
        LOW(8, PixelFormat.defaultLoColor),
        MEDIUM(16, PixelFormat.defaultMidColor),
        HIGH(32, PixelFormat.defaultHiColor);

        private int code;
        private PixelFormat format;

        PixelFormatQuality(int code, PixelFormat format) {
            this.code = code;
            this.format = format;
        }

        public int getCode() {
            return code;
        }

        public PixelFormat getFormat() {
            return format;
        }

        public static PixelFormatQuality getTypeByCode(int code) {
            for (PixelFormatQuality type : values()) {
                if (type.getCode() == code) {
                    return type;
                }
            }
            return PixelFormatQuality.EMPTY;
        }
    }

	// 32 bpp @ 24 bit depth True colour
    private static PixelFormat  defaultHiColor  = new PixelFormat( false, true, 32, 24, 255, 255, 255, 16, 8, 0);
	// 16 bpp @ 16 bit depth High colour
    private static PixelFormat  defaultMidColor = new PixelFormat( false, true, 16, 16,  31,  63,  31, 11, 5, 0);
	//  8 bpp @  8 bit depth 265 colours   	
    private static PixelFormat  defaultLoColor  = new PixelFormat( false, true,  8,  8,   7,   7,   3,  0, 3, 6);
	// 8 bpp @   6 bit depth 64 colours Amiga Chipset style (Thanks Wikipedia)
    public static PixelFormat  pf4 = new PixelFormat( false, true,  8,  6,   3,   3,   3,  4, 2, 0);
	// 8 bpp @   3 bit depth  8 colours ZX Spectrum style (Thanks Wikipedia)
    public static PixelFormat  pf5 = new PixelFormat( false, true,  8,  3,   1,   1,   1,  2, 1, 0);
	// 8 bpp @   2 bit depth  4 colours CGA         GRAY SCALE!
    public static PixelFormat  pf6 = new PixelFormat( false, true,  8,  6,   3,   3,   3,  4, 2, 0);
	// 8 bpp @   1 bit depth  2 colours monocrome	GRAY SCALE!
    public static PixelFormat  pf7 = new PixelFormat( false, true,  8,  3,   1,   1,   1,  2, 1, 0);

    
    public static String str08Bit =  "8 Bit";
    public static String str16Bit = "16 Bit";
    public static String str32Bit = "32 Bit";

    private boolean bigEndian   = true;
    private boolean trueColour  = true;

    private int bitsPerPixel    = 32;
    private int depth           = 24;
    private int maxRed          = 255;
    private int maxGreen        = 255;
    private int maxBlue         = 255;
    private int shiftRed        = 16;
    private int shiftGreen      = 8;
    private int shiftBlue       = 0;

    public enum Colour { RED, GREEN, BLUE};

    /**Constructs a hi color PixelFormat object.*/
    public PixelFormat() {
    }

//    private PixelFormat( int pfType) {
//    	if ( pfType > 0) {
//            bitsPerPixel    = 32;
//            depth           = 24;
//            maxRed          = 255;
//            maxGreen        = 255;
//            maxBlue         = 255;
//            shiftRed        = 16;
//            shiftGreen      = 8;
//            shiftBlue       = 0;
//            bigEndian       = false;
//            trueColour      = true;
//            return;
//        }
//
//        if ( pfType < 0) {
//            bitsPerPixel    = 8;
//            depth           = 8;
//            maxRed          = 7;
//            maxGreen        = 7;
//            maxBlue         = 3;
//            shiftRed        = 0;
//            shiftGreen      = 3;
//            shiftBlue       = 6;
//            bigEndian       = false;
//            trueColour      = true;
//            return;
//        }
//
//        bitsPerPixel    = 16;
//        depth           = 16;
//        maxRed          = 31;
//        maxGreen        = 63;
//        maxBlue         = 31;
//        shiftRed        = 11;
//        shiftGreen      = 5;
//        shiftBlue       = 0;
//        bigEndian       = false;
//        trueColour      = true;
//    }

    /**This constructor specifically addresses the needs of the VNCViewer.
    Please do not change the order in which it will read the Items from
    the stream nor share the stream with another thread.
    @param is The {@link DataInputStream} to read from.
    @throws IOException in case the input stream is doesn't cooperate.*/
    public PixelFormat( DataInputStream is) throws IOException {
        bitsPerPixel    = is.readUnsignedByte();
        depth           = is.readUnsignedByte();
        bigEndian       = (is.readUnsignedByte() != 0);
        trueColour      = (is.readUnsignedByte() != 0);
        maxRed          = is.readUnsignedShort();
        maxGreen        = is.readUnsignedShort();
        maxBlue         = is.readUnsignedShort();
        shiftRed        = is.readUnsignedByte();
        shiftGreen      = is.readUnsignedByte();
        shiftBlue       = is.readUnsignedByte();
    }

    private PixelFormat( boolean bigEnd, boolean trueCol,
                        int bitspp, int dpth,
                        int maxR, int maxG, int maxB,
                        int shiftR, int shiftG, int shiftB)
    {
        bigEndian       = bigEnd;
        trueColour      = trueCol;
        bitsPerPixel    = bitspp;
        depth           = dpth;

        maxRed          = maxR;
        maxGreen        = maxG;
        maxBlue         = maxB;

        shiftRed        = shiftR;
        shiftGreen      = shiftG;
        shiftBlue       = shiftB;
    }

    public Object clone() throws CloneNotSupportedException {
        return super.clone();
    }

//    public boolean matches( int t) {
//        return type == t;
//    }

    public boolean equals( PixelFormat pf) {
    	if ( pf == null) return false;

        return  bigEndian       == pf.bigEndian &&
                trueColour      == pf.trueColour &&
                bitsPerPixel    == pf.bitsPerPixel &&
                depth           == pf.depth &&

                maxRed          == pf.maxRed &&
                maxGreen        == pf.maxGreen &&
                maxBlue         == pf.maxBlue &&

                shiftRed        == pf.shiftRed &&
                shiftGreen      == pf.shiftGreen &&
                shiftBlue       == pf.shiftBlue;
    }

    public int getBitsPerPixel() {
        return bitsPerPixel;
    }

    public int getBytesPerPixel() {
    	return (int)( bitsPerPixel / 8);
    }

    public int getMaxColors() {
        return (int)Math.pow( 2.0, depth);
    }

    public int getDepth() {
        return depth;
    }

    public int getRedShift()   { return shiftRed; }
    public int getGreenShift() { return shiftGreen; }
    public int getBlueShift()  { return shiftBlue; }

    public int getRedMax()   { return maxRed; }
    public int getGreenMax() { return maxGreen; }
    public int getBlueMax()  { return maxBlue; }

    public int getColor( boolean shift, Colour c) {
        switch ( c) {
            case RED   : return ( shift ? shiftRed   : maxRed  );
            case GREEN : return ( shift ? shiftGreen : maxGreen);
            case BLUE  : return ( shift ? shiftBlue  : maxBlue );
            default : return -1;
        }
    }

    public boolean isBigEndian() {
        return bigEndian;
    }

    public boolean isTrueColour() {
        return trueColour;
    }

    public boolean need2swap() {
        boolean hostByteOrderIsBigEndian = ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN;
        return bigEndian == hostByteOrderIsBigEndian;
    }

    public byte[] getAsRFBProtocolCompliantByteArray() {
        byte[] b = new byte[ 20];

        b[  0] = (byte) RFB.SetPixelFormat;
        // 3 bytes padding
        b[  4] = (byte) bitsPerPixel;
        b[  5] = (byte) depth;
        b[  6] = (byte) (bigEndian ? 1 : 0);
        b[  7] = (byte) (trueColour ? 1 : 0);
        b[  8] = (byte) ((maxRed >> 8) & 0xff);
        b[  9] = (byte) (maxRed & 0xff);
        b[ 10] = (byte) ((maxGreen >> 8) & 0xff);
        b[ 11] = (byte) (maxGreen & 0xff);
        b[ 12] = (byte) ((maxBlue >> 8) & 0xff);
        b[ 13] = (byte) (maxBlue & 0xff);
        b[ 14] = (byte) shiftRed;
        b[ 15] = (byte) shiftGreen;
        b[ 16] = (byte) shiftBlue;
        // 3 bytes padding

        return b;
    }

    public String toString() {
    	return bitsPerPixel + " Bits @ " + depth;
    }

    public void dump() {
        printProps( this);
    }

    // Don't ask. Tried to be smart. Turned out that was one stupid idea. 
    public static String print( PixelFormat pf) {
    	StringWriter sw = new StringWriter();
    	PrintWriter pw = new PrintWriter( sw, true);
        boolean bigEndianLoc = ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN;
        String ms = pf.need2swap() ? "No swapping necessary." : "Must swap bytes.";
        
    	String f = "Bytes per Pixel          : [%d]\n" +
                   "Depth                    : [%d] = [%d] colors max\n" +
    			   "BigEndian (remote/local) : [%b/%b] => %s\n" +
                   "TrueColour               : [%b]\n" +
                   "max(R/G/B)               : (%d/%d/%d)\n" +
                   "shift(R/G/B)             : (%d/%d/%d)\n";
    	pw.format( f, pf.getBytesPerPixel(), pf.getDepth(), pf.getMaxColors(),
    			      pf.bigEndian, bigEndianLoc, ms, pf.trueColour,
    			      pf.maxRed, pf.maxGreen, pf.maxBlue,
    			      pf.shiftRed, pf.shiftGreen, pf.shiftBlue);
    	return sw.toString();
    }

    public static void printProps( PixelFormat pf)
    {
        boolean bigEndianLoc = ByteOrder.nativeOrder() == ByteOrder.BIG_ENDIAN;
        String ms = pf.need2swap() ? "No swapping necessary." : "Must swap bytes.";
        System.out.println( "Bytes per Pixel          : [" + pf.getBytesPerPixel() + "]");
        System.out.println( "Depth                    : [" + pf.getDepth() + "] = [" + pf.getMaxColors() + "] colors max");
        System.out.println( "BigEndian (remote/local) : [" + pf.bigEndian  + "/" + bigEndianLoc + "] => " + ms);
        System.out.println( "TrueColour               : [" + pf.trueColour + "]");
        System.out.println( "max(R/G/B)               : (" + pf.maxRed   + "/" + pf.maxGreen   + "/" + pf.maxBlue   + ")");
        System.out.println( "shift(R/G/B)             : (" + pf.shiftRed + "/" + pf.shiftGreen + "/" + pf.shiftBlue + ")");
    }
}
