package com.wincornixdorf.vnc.rfb.fb;

import com.wincornixdorf.vnc.rfb.RFB;

// TODO: Make VNCEncoding a proper enum where humanly possible!
public class VNCEncoding extends Object {

    public static final VNCEncoding Auto        = new VNCEncoding( RFB.EncodingidAuto, false, false);
    public static final VNCEncoding Raw         = new VNCEncoding( RFB.EncodingRaw);
    public static final VNCEncoding CopyRect    = new VNCEncoding( RFB.EncodingCopyRect, false, false);
    public static final VNCEncoding RRE         = new VNCEncoding( RFB.EncodingRRE);
    public static final VNCEncoding CoRRE       = new VNCEncoding( RFB.EncodingCoRRE);
    public static final VNCEncoding Hextile     = new VNCEncoding( RFB.EncodingHextile);
    public static final VNCEncoding Zlib        = new VNCEncoding( RFB.EncodingZlib, false, true);
    public static final VNCEncoding Tight       = new VNCEncoding( RFB.EncodingTight);
    public static final VNCEncoding ZRLE        = new VNCEncoding( RFB.EncodingZRLE);

    // pseudo encodings
    public static final VNCEncoding CompressLvl = new VNCEncoding( RFB.EncodingCompressLevel0, true, true);
    public static final VNCEncoding QualityLvl  = new VNCEncoding( RFB.EncodingQualityLevel0, true, true);
    public static final VNCEncoding XCursor     = new VNCEncoding( RFB.EncodingXCursor, true, true);
    public static final VNCEncoding RichCursor  = new VNCEncoding( RFB.EncodingRichCursor, true, true);
    public static final VNCEncoding PointerPos  = new VNCEncoding( RFB.EncodingPointerPos, true, true);
    public static final VNCEncoding LastRect    = new VNCEncoding( RFB.EncodingLastRect, true, true);
    public static final VNCEncoding NewFBSize   = new VNCEncoding( RFB.EncodingNewFBSize, true, true);

    public static final VNCEncoding DEFAULT_ENC = Tight;
    
    /**The constant of the encoding as defined in the RFB protocol.*/
    private int type                            = -2;

    /**A human readable String to identify the encoding.*/
    private String idStr                        = "";

    /**May look like an ingenious way to encode the compression and quality
    level values into the VNCEncoding objects intestines, but this is nothing
    but a dirty hack. The support for this value is sketchy at most :( This
    value is added to the value of type and returned via toInt.*/
    private int specialValue                    = 0;

    /**This flag is set to mark encodings as a pseudo encodings. The original
    protocol only referred to encodings as the identifiers used by client and
    server to negotiate in what way the FRAMEBUFFER DATA being exchanged in
    between each other was encoded ( compressed). Pseudo encodings are
    identifiers which have obviously been introduced later. They are used today
    to negotiate settings relevant to the client server communication that do
    NOT refer to the framebuffer packets being exchanged.*/
    private boolean isPseudo                    = false;

    /**If set the viewer will treat the (pseudo)-encoding as an encoding it
    can handle. It is primarily intended for testing and debugging purposes.
    If in doubt, leave it alone :)*/
    private boolean isSupported                 = false;
 
    /** Convenience method for looking up an Encoding by its name.
     * @param s The name of the {@link VNCEncoding} to look for. Expects a String generated by getEncodingAsString. 
     * @param skipPseudo If set, pseudo encodings will not be considered.
     * @return A reference to the VNCEncoding object represented by the String or VNCEncoding.Auto if unknown.  */
    public static VNCEncoding getEncoding( String s, boolean skipPseudo) {
        if ( s.equals( getEncodingAsString( RFB.EncodingidAuto)))   return Auto;
        if ( s.equals( getEncodingAsString( RFB.EncodingRaw)))      return Raw;
        if ( s.equals( getEncodingAsString( RFB.EncodingCopyRect))) return CopyRect;
        if ( s.equals( getEncodingAsString( RFB.EncodingRRE)))      return RRE;
        if ( s.equals( getEncodingAsString( RFB.EncodingCoRRE)))    return CoRRE;
        if ( s.equals( getEncodingAsString( RFB.EncodingHextile)))  return Hextile;
        if ( s.equals( getEncodingAsString( RFB.EncodingZlib)))     return Zlib;
        if ( s.equals( getEncodingAsString( RFB.EncodingTight)))    return Tight;
        if ( s.equals( getEncodingAsString( RFB.EncodingZRLE)))     return ZRLE;
        if ( skipPseudo) return Auto;

        if ( s.equals( getEncodingAsString( RFB.EncodingCompressLevel0)))   return CompressLvl;
        if ( s.equals( getEncodingAsString( RFB.EncodingQualityLevel0)))    return QualityLvl;
        if ( s.equals( getEncodingAsString( RFB.EncodingXCursor)))          return XCursor;
        if ( s.equals( getEncodingAsString( RFB.EncodingRichCursor)))       return RichCursor;
        if ( s.equals( getEncodingAsString( RFB.EncodingPointerPos)))       return PointerPos;
        if ( s.equals( getEncodingAsString( RFB.EncodingLastRect)))         return LastRect;
        if ( s.equals( getEncodingAsString( RFB.EncodingNewFBSize)))        return NewFBSize;
        return Auto;
    }

    /**@param rfbEncoding Must be one of the Encoding constants defined in {@link RFB}.
     * @return A String representing a particular encoding. These Strings can be used as
     * references in various places as the inverse function getEncoding can map the String
     * back to a VNCEncoding object at any time.*/
    public static String getEncodingAsString( int rfbEncoding) {
        switch ( rfbEncoding) {
            case RFB.EncodingidAuto             : return( "Auto");
            case RFB.EncodingRaw                : return( "Raw");
            case RFB.EncodingCopyRect           : return( "CopyRect");
            case RFB.EncodingRRE                : return( "RRE");
            case RFB.EncodingCoRRE              : return( "CoRRE");
            case RFB.EncodingHextile            : return( "Hextile");
            case RFB.EncodingZlib               : return( "Zlib");
            case RFB.EncodingTight              : return( "Tight");
            case RFB.EncodingZRLE               : return( "ZRLE");

            case RFB.EncodingCompressLevel0     : return( "Pseudo-encoding CompressLevel");
            case RFB.EncodingQualityLevel0      : return( "Pseudo-encoding QualityLevel");
            case RFB.EncodingXCursor            : return( "Pseudo-encoding XCursor");
            case RFB.EncodingRichCursor         : return( "Pseudo-encoding RichCursor");
            case RFB.EncodingPointerPos         : return( "Pseudo-encoding Pointer Pos");
            case RFB.EncodingLastRect           : return( "Pseudo-encoding Last Rect");
            case RFB.EncodingNewFBSize          : return( "Pseudo-encoding New FBSize");
            default: return( "Error");
        }
    }

    /**@param rfbConstant Must be one of the encoding constants defined in the RFB protocol.
     * @return The VNCEncoding object wrapping a particular RFB encoding.*/
    public static VNCEncoding getEncoding( int rfbConstant) {
        switch ( rfbConstant) {
            case RFB.EncodingidAuto          : return( VNCEncoding.Auto);
            case RFB.EncodingRaw             : return( VNCEncoding.Raw);
            case RFB.EncodingCopyRect        : return( VNCEncoding.CopyRect);
            case RFB.EncodingRRE             : return( VNCEncoding.RRE);
            case RFB.EncodingCoRRE           : return( VNCEncoding.CoRRE);
            case RFB.EncodingHextile         : return( VNCEncoding.Hextile);
            case RFB.EncodingZlib            : return( VNCEncoding.Zlib);
            case RFB.EncodingTight           : return( VNCEncoding.Tight);
            case RFB.EncodingZRLE            : return( VNCEncoding.ZRLE);

            case RFB.EncodingCompressLevel0  : return( VNCEncoding.CompressLvl);
            case RFB.EncodingQualityLevel0   : return( VNCEncoding.QualityLvl);
            case RFB.EncodingXCursor         : return( VNCEncoding.XCursor);
            case RFB.EncodingRichCursor      : return( VNCEncoding.RichCursor);
            case RFB.EncodingPointerPos      : return( VNCEncoding.PointerPos);
            case RFB.EncodingLastRect        : return( VNCEncoding.LastRect);
            case RFB.EncodingNewFBSize       : return( VNCEncoding.NewFBSize);

            default: return( VNCEncoding.Auto);
        }
    }

    private VNCEncoding( int i) {
        type = i;
        idStr = getEncodingAsString( i);
        isPseudo = false;
        isSupported = true;
    }

    private VNCEncoding( int i, boolean pseudo, boolean supported) {
        type = i;
        idStr = getEncodingAsString( i);
        isPseudo = pseudo;
        isSupported = supported;
    }

    public int getType() { return type; }

    public int toInt() { return type + specialValue; }

    public int getSpecialValue() { return specialValue; }
    public boolean setSpecialValue( int v) {
        if ( !isPseudo) return false;
        specialValue = v;
        return true;
    }

    public String toString() {
        return( idStr);
    }

    public boolean isPseudo() {
        return isPseudo;
    }

    public boolean isSupported() {
        return isSupported;
    }

    public static VNCEncoding craftCompressionLvl( int i) {
        VNCEncoding e = new VNCEncoding( RFB.EncodingCompressLevel0, true, true);
        e.specialValue = i;
        return e;
    }

    public static VNCEncoding craftQualityLvl( int i) {
        VNCEncoding e = new VNCEncoding( RFB.EncodingQualityLevel0, true, true);
        e.specialValue = i;
        return e;
    }
}
