package com.wincornixdorf.vnc.rfb.ft;

import java.io.*;
import java.util.zip.*;
import java.util.*;

import com.wincornixdorf.vnc.gui.ft.FTStatus;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent;

public class DirectoryTransferAssistant extends Observable implements Runnable {

    private File            dir2zip = null;
    private File            zipFile = null;
    private RFBHandler   conn    = null;
    private String          ffn     = null;

    public DirectoryTransferAssistant( String fndir2zip, RFBHandler parent) throws IOException {
        this.dir2zip = new File( fndir2zip);
        this.conn = parent;

        // Create a temporary file to download the zipped directory to.
        String tmpZipFnExt  = ".zip";
        String tmpZipFnBase = "!UVNCDIR-" + dir2zip.getName();
        ffn = tmpZipFnBase + tmpZipFnExt;
        this.zipFile = File.createTempFile( tmpZipFnBase, tmpZipFnExt);
    }

    public void run() {
        boolean zipSuccess = false;

        try {
            conn.fireEvent( RFBHandlerEvent.zipbegin( conn,
                dir2zip.getAbsolutePath(), zipFile.getAbsolutePath(), true, true), null);

            ZipOutputStream zos = new ZipOutputStream( new FileOutputStream( zipFile));

            int fcnt = countObservers() > 0 ? countFilesInFolder( dir2zip) : 0;
            zip( dir2zip, zos, dir2zip.getAbsolutePath().length() + 1, fcnt, 0);
            zos.close();
            zipSuccess = true;
            //rmFilePostUpload = true;
        } catch ( IOException e) {
            e.printStackTrace();
        } finally {
            String msg = zipSuccess ? null : "Failed to zip directory [" + dir2zip + "].\nPlease check file privilleges.";
            conn.fireEvent( RFBHandlerEvent.zipend( conn,
                dir2zip.getAbsolutePath(), zipFile.getAbsolutePath(), "", true, true, true), msg);
        }

        if ( zipSuccess) {
            while (conn.uploadFile(zipFile.getAbsolutePath(), null, ffn).equals(FTStatus.ANOTHER_FT_IN_PROGRESS)){}
        }
    }

    public File getZipFile() {
        return zipFile;
    }

    private int countFilesInFolder( File f) {
        if ( f == null || !f.exists()) return -1;
        if ( !f.isDirectory()) return 1;

        File[] fList = f.listFiles();
        if ( fList == null) return 0;

        int c = 0;

        for ( int i = 0; i < fList.length; i++)
            c += countFilesInFolder( fList[ i]);

        return c;
    }

    private int zip( File dir2zip, ZipOutputStream zos, int cp, int fcount, int zipped) throws IOException {
        if ( dir2zip.isDirectory()) {
            File[] dirList = dir2zip.listFiles();

            if ( dirList == null) return zipped;

            for ( int i = 0; i < dirList.length; i++)
                zipped = zip( dirList[ i], zos, cp, fcount, zipped);
        } else {
            String hs = dir2zip.getAbsolutePath().substring( cp);
            zos.putNextEntry( new ZipEntry( hs));

            FileInputStream is = new FileInputStream( dir2zip);

            byte[] b = new byte[ 1024];
            int len = 0;

            while (( len = is.read( b)) >= 0)
                zos.write( b, 0, len);

            zos.closeEntry();
            is.close();

            RFBHandlerEvent e1 = RFBHandlerEvent.zipprogress( conn,
                dir2zip.getAbsolutePath(), zipFile.getAbsolutePath(),
                true, true, true);

            if ( countObservers() > 0) {
                int pcnt = fcount != 0 ? (int)((zipped++ * 100) / fcount) : 0;
                e1.progress = pcnt;
                setChanged();
                notifyObservers( e1);
            }
        }
        return zipped;
    }
}

