package com.wincornixdorf.vnc.rfb.ft;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.Queue;
import java.util.logging.Level;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.Inflater;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import com.wincornixdorf.vnc.global.VNCViewer;
import com.wincornixdorf.vnc.gui.ft.FTStatus;
import com.wincornixdorf.vnc.rfb.RFB;
import com.wincornixdorf.vnc.rfb.RFBHandler;
import com.wincornixdorf.vnc.rfb.RFBHandlerEvent;

/**
 * The VNCConnectionFT class was created to separate those parts of the code
 * in the VNCConnection class which deal with file transfer operations. Since many
 * methods in this class operate directly on the sacred stream, these methods may
 * only be called from the connection thread. Failure to abide by that rule will
 * result in catastrophic crashes and must be avoided at all cost.
 */
public class RFBHandlerFT extends Observable implements Observer {
    private static final long MAXWORD = 0xffffffffL;

    // These are SO PRIVATE! Hands off or my wrath shall be eternal!
    /**
     * This is a COPY of the sacred stream the session uses to read from the socket.
     * Come heaven, hell high waters or a newsflash saying hell had just frozen over
     * DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!
     */
    private DataInputStream is = null;
    /**
     * This is a COPY of the sacred stream the session uses to write to the socket.
     * Come heaven, hell high waters or a newsflash saying hell had just frozen over
     * DO NOT SHARE THIS STREAM WITH ANY OTHER THREAD!
     */
    private OutputStream os = null;

    /**
     * Used to compute the time taken to transfer a file.
     */
    private Date ftStarted = null;

    /**
     * Name of the local file which is to be uploaded to the server.
     */
    private String fnUpload = null;

    /**
     * Name of the file being written to when downloading a file or a
     * compressed directory from the server.
     */
    private String fnDownload = null;
    /**
     * Name of the file being deleted on the servers side.
     */
    private String fnDeleting = null;

    /**
     * Name of the directory on the servers side which is supposed to be
     * downloaded. This is the one without the silly brackets!
     */
    private String dn2bDownloaded = null;
    /**
     * Name of the directory to be created on the server.
     */
    private String dnCreating = null;

    /**
     * Reference to a file containing a ZIP archive. The archive has either
     * just been downloaded from the server and awaits being inflated or
     * contains the zipped contents of a directory which is to be uploaded to the
     * server. Which is which, depends on whether \a unzip2Dir is set or not.
     */
    private File zippedDir = null;
    /**
     * Reference to a directory where the contents of \a zippedDir are to be
     * expanded to, once the compressed directory has arrived. This variable
     * should only be set when downloading a directory from the server.
     */
    private File unzip2Dir = null;

    /**
     * Files being downloaded are written to this Output Stream. Since the
     * download process is split across several methods, this Stream must be
     * globally accessible.
     */
    private FileOutputStream fosFileDownload = null;

    /**
     * A List of Directories known to exist on the server. The FTPath objects
     * contain information on the drive letters associated with them in windumb
     * environments and whether those are CD-ROM, Floppy, Network or physical
     * drives.
     */
    private ArrayList<FTPath> remoteDrivesList = new ArrayList<FTPath>();

    /**
     * A copy of the remoteDrivesList is being held here while the original
     * list gets updated. The new list will be compared to the old one to detect
     * differences if the list has changed. To save on memory, the list will be
     * deleted again, once it has been compared to the new list. You should
     * generally expect the list to be null and void.
     */
    private ArrayList<FTPath> oldRDL = new ArrayList<FTPath>();

    /**
     * This Vector contains the contents of the "Remote File System".
     */
    private RemoteFileSystem rfs = new RemoteFileSystem("server filesystem");

    /**
     * This variable gets checked repeatedly throughout up- and downloads. If set
     * an attempt will be made at aborting the file transfer. Providing the server
     * accepts the request, the file transfer will be interrupted.
     */
    private boolean fAbort = false;

    /**
     * This flag will be set, whenever the VNC server is asked to do something which
     * can vaguely be considered a file transfer operation. File Up- and Downloads are
     * considered FT operations. So are requests to delete Files and list directory
     * contents. Future versions may know many more operations not mentioned here. The
     * flag is used to make sure that only one such operation is performed at any one
     * time. The VNC protocol is unfit to handle more than one at a time. Please
     * use the 'setBusy' method to toggle it's state.
     */
    private boolean ftActionInProgress = false;
    private Object ftActionInProgressLock = new Object();

    /**
     * Name of the current directory on the server side.
     */
    private String currentDir = "";
    /**
     * The name of the server directory which is currently being listed.
     */
    private String desiredDir = "";

    /**
     * If set files being uploaded will be deleted once they have been transferred.
     */
    private boolean rmFilePostUpload = false;
    private boolean inDirectory2;

    private long downloadBytesTotal = 0;
    private long downloadBytesRead = 0;

    private int npBytesRead = 0;

    /**
     * Used 4 internal reasons only. Should ideally be removed soon.
     */
    private RFBHandler vc = null;

    public RFBHandlerFT(RFBHandler c/*, String iDir*/) {
        this.vc = c;
    }

    //TODO: Storing references to the streams here is one VERY bad idea.
    public void setStreams(DataInputStream isOfSession, OutputStream osOfSession) {
//      if ( is == null) this.is = isOfSession;
//      if ( os == null) this.os = osOfSession;
        this.is = isOfSession;
        this.os = osOfSession;
    }

    /**
     * @return The Name of the drive of the current directory. This method is intended
     * for windumb environments and will likely fail in *NIX environments.
     */
    public String getCurrentDrive() {
        if (currentDir == null) return "C:\\";
        int i = currentDir.indexOf(':');
        return (i < 0 ? "C:\\" : currentDir.substring(0, i + 2));
    }

    /**
     * @return The number of files in the present remote directory.
     */
    public int getFilesCount() {
        return rfs.getFilesCount();
    }

    /**
     * @return The number of directories in the present remote directory.
     */
    public int getDirectoriesCount() {
        return (rfs.getDirectoriesCount() - 1); // One is the parent directory '..'
    }

    public RemoteFileSystem getRFS() {
        return rfs;
    }

    public ArrayList<FTPath> getKnownServerDrives() {
        return remoteDrivesList;
    }

    private FTPath getFirstDrive(char t) {
        if (remoteDrivesList == null) return null;

        FTPath rval = null;
        Iterator<FTPath> i = remoteDrivesList.iterator();
        while (i.hasNext() && rval == null) {
            FTPath drv = i.next();
            if (drv.isOfType(t)) rval = drv;
        }

        return rval;
    }

    /**
     * Will search the list of known drives for a FTPath with the properties of the argument.
     *
     * @param d A {@link FTPath} describing the one to search.
     * @return The {@link FTPath} from the list of known Drives having the same properties as the
     * argument, or null if none is found.
     */
    private FTPath getDrive(FTPath d) {
        if (remoteDrivesList == null) return null;

        FTPath rval = null;
        Iterator<FTPath> i = remoteDrivesList.iterator();
        while (i.hasNext() && rval == null) {
            FTPath drv = i.next();
            if (drv.equals(d)) rval = drv;
        }

        return rval;
    }

    public FTPath getFallbackDrive() {
        FTPath rval = null;

        // A local drive would be ideal
        rval = getFirstDrive('l');
        if (rval != null) return rval;

        // An optical media is nice too
        rval = getFirstDrive('c');
        if (rval != null) return rval;

        // Network drives only if need be
        rval = getFirstDrive('n');
        if (rval != null) return rval;

        // Floppy drives are so dead
        rval = getFirstDrive('f');
        return rval;
    }

    /**
     * Will send a download request to the server, asking it to send the file
     * \a remoteFN. The file will then be written to \a localPath. The method will
     * overwrite existing files. If another FT operation is in progress the method
     * will return false without sending the request. The \a isDirectory parameter has to
     * be set true, if the file to be downloaded is in fact a directory.
     *
     * @param remoteFn    The Filename of the file to request.
     * @param localPath   Where the file is to be downloaded to.
     * @param isDirectory If the file is in fact a Directory.
     * @return False if another FT action is currently in progress.
     */
    public FTStatus downloadFile(String remoteFn, String localPath, boolean isDirectory) {

        if ((ftActionInProgress || fnDownload != null) && !isDirectory) {
            return FTStatus.ANOTHER_FT_IN_PROGRESS;
        }

        fnDownload = localPath;
        ftStarted = new Date();

        vc.fireEvent(RFBHandlerEvent.ftbegin(vc, remoteFn, localPath, true, isDirectory), null);

        // Who ever wrote that server should really be taught a lesson
        // on how to separate data and it's logical representation.
        // Don't mess with the braces! The blanks are intentional too :(
        if (isDirectory) remoteFn = getCurrentDir() + "[ " + remoteFn + " ]";

        VNCViewer.pen.log(Level.CONFIG, vc.logStr("Downloading [" + remoteFn + "] to [" + localPath + "]"));

        try {
            setBusy(true);
            // 3rd parameter : 0 = compression not supported; 1 = compression supported
            sendFTOperationRequest(RFB.rfbFileTransferRequest, 0, 1, remoteFn.length(), remoteFn);
        } catch (IOException e) {
            setBusy(false);
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to download [" + remoteFn + "] to [" + localPath + "]"), e);
            return FTStatus.FT_ERROR;
        }

        return FTStatus.FT_REQUEST_PROCESSED;
    }

    /**
     * Will send an upload file offer to the server, asking it to write the file
     * to the path specified in \a destinationPath. The \a fakeFilename will be
     * sent to the server instead of the real filename, which is necessary to use
     * the method for directory uploads. Set the variable to null to send the
     * original filename. The method will overwrite existing files.
     *
     * @param localFn         Name of the file in the local FS.
     * @param destinationPath Desired destination directory for the file to be uploaded.
     * @param fakeFilename    will be sent to the server instead of the real filename
     * @return Returns false, without sending the offer, if another FT operation is in progress.
     */
    public FTStatus uploadFile(String localFn,
                               String destinationPath,
                               String fakeFilename) {

        if (ftActionInProgress || fnUpload != null) {
            return FTStatus.ANOTHER_FT_IN_PROGRESS;
        }

        try {
            ftStarted = new Date();
            setBusy(true);
            fnUpload = localFn;

            File f = new File(fnUpload);
            long lSize = f.length();
            int iLowSize = (int) (lSize & 0x00000000FFFFFFFF);
            int iHighSize = (int) (lSize >> 32);

            if (destinationPath == null) destinationPath = currentDir;

            vc.fireEvent(RFBHandlerEvent.ftbegin(vc, localFn, destinationPath, false, fakeFilename != null), null);

            String temp = destinationPath + (fakeFilename != null ? fakeFilename : f.getName());

            VNCViewer.pen.log(Level.CONFIG, vc.logStr("Uploading file [" + fnUpload + "] to [" + destinationPath + "]"));
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Filename sent in FT offer: [" + temp + "]"));

            sendFTOperationRequest(RFB.rfbFileTransferOffer, 0, iLowSize, temp.length(), temp);

            byte b[] = new byte[4];
            b[0] = (byte) ((iHighSize & 0xFF000000) >>> 24);
            b[1] = (byte) ((iHighSize & 0xFF0000) >>> 16);
            b[2] = (byte) ((iHighSize & 0xFF00) >>> 8);
            b[3] = (byte) ((iHighSize & 0xFF));
            os.write(b);
        } catch (IOException e) {
            setBusy(false);
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to upload file [" + fnUpload + "] to [" + destinationPath + "]"), e);
            return FTStatus.FT_ERROR;
        }


        return FTStatus.FT_REQUEST_PROCESSED;
    }

    public FTStatus downloadDirectory(String dir2Download, String targetPath) {


        if(ftActionInProgress){
            return FTStatus.ANOTHER_FT_IN_PROGRESS;
        }

        try {

            setBusy(true);

            // Bounce if targetPath does NOT point to a directory ...
            unzip2Dir = new File(targetPath);
            if (!unzip2Dir.isDirectory()) {
                unzip2Dir = null;
                return FTStatus.FT_ERROR;
            }

            dn2bDownloaded = dir2Download;

            // We shall not create this directory until the requested zip file
            // has actually arrived, so we won't have to delete it again in the
            // event of an error.
            unzip2Dir = new File(unzip2Dir, dir2Download);

            // Create a temporary file to download the zipped directory to.
            zippedDir = File.createTempFile("dirdownload", ".zip");
            VNCViewer.pen.log(Level.FINEST, vc.logStr("created tmp file " + zippedDir.getAbsolutePath() + " for requested dir " + dir2Download));

            // Download the directory as a zipped file to the temp file
            return downloadFile(dir2Download, zippedDir.getAbsolutePath(), true);
        } catch (IOException e) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to download Directory [" + dir2Download + "] to [" + targetPath + "]"), e);
            setBusy(false);
            return FTStatus.FT_ERROR;
        }

    }

    public FTStatus uploadDirectory(File dir2Upload, String targetPath, RFBHandler s) {

        if (ftActionInProgress || fnUpload != null) {
            return FTStatus.ANOTHER_FT_IN_PROGRESS;
        }

        try {
            DirectoryTransferAssistant z = new DirectoryTransferAssistant(dir2Upload.getAbsolutePath(), s);
            z.addObserver(this);
            Thread t = new Thread(z);
            t.setName("FTUpload_Thread");
            t.start();
        } catch (IOException e) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to upload Directory [" + dir2Upload.getAbsoluteFile() + "] to [" + targetPath + "]"), e);
            setBusy(false);
            return FTStatus.FT_ERROR;
        }

        return FTStatus.FT_REQUEST_PROCESSED;
    }

    /**
     * Will send a delete file request to the server, asking it to delete the file
     * \a filename. The path must be fully qualified. If another FT operation is in
     * progress the method will return false without sending the request.
     *
     * @param filename The fully qualified filename of the file to be deleted on the server.
     * @return False if another FT action is currently in progress.
     */
    public FTStatus deleteRemoteFile(String filename) {
        if (ftActionInProgress || fnDeleting != null) {
            return FTStatus.ANOTHER_FT_IN_PROGRESS;
        }
        VNCViewer.pen.log(Level.CONFIG, vc.logStr("Deleting remote file [" + filename + "]"));

        try {
            setBusy(true);
            fnDeleting = filename;
            sendFTOperationRequest(RFB.rfbCommand, RFB.rfbCFileDelete, 0, filename.length(), filename);
        } catch (IOException e) {
            setBusy(false);
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to upload delete remote file [" + filename + "]"), e);
            return FTStatus.FT_ERROR;
        }

        return FTStatus.FT_REQUEST_PROCESSED;
    }

    /**
     * Will send a create directory request to the server, asking it to create
     * the path specified in \a dirName. The path must be fully qualified. This is
     * <b>NOT</b> a create tree function.
     *
     * @param dirName The fully qualified name of the directory to create.
     * @return Returns False if another FT operation is in progress, in which case the request won't be sent.
     */
    public boolean createRemoteDirectory(String dirName) {
        if (ftActionInProgress) return false;

        try {
            setBusy(true);
            dnCreating = dirName;
            VNCViewer.pen.log(Level.CONFIG, vc.logStr("Creating directory [" + dirName + "] on the server."));
            sendFTOperationRequest(RFB.rfbCommand, RFB.rfbCDirCreate, 0, dirName.length(), dirName);
        } catch (IOException e) {
            setBusy(false);
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to create remote directory [" + dirName + "]"), e);
            return false;
        } finally {
            setBusy(false);
        }

        return true;
    }

    /**
     * This method is purely for testing purposes. Do NOT use it! In fact I
     * looked into the code to find the server doesn't even implement directory
     * deletion.
     *
     * @param dirName Name of folder to delete.
     */
    @Deprecated
    public void deleteRemoteDirectory(String dirName) {
        if (ftActionInProgress) return;

        try {
            setBusy(true);
            String hs = "[ " + ".aaatest" + " ]";
            VNCViewer.pen.log(Level.CONFIG, vc.logStr("Deleting remote directory [" + hs + "] on the server."));
            sendFTOperationRequest(RFB.rfbCommand, RFB.rfbCDirDelete, 0, hs.length(), hs);
        } catch (IOException e) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("IO exception occurred while trying to delete directory [" + dirName + "] on the server."));
            e.printStackTrace();
        } finally {
            setBusy(false);
        }
    }

    public String getCurrentDir() {
        return currentDir;
    }

    /**
     * Tries to set the current directory. The method will do nothing if the current
     * directory is already set as this method is designed to be invoked only when
     * creating the connection. Callers are expected to attempt setting the current
     * to the one last used.
     *
     * @param dir The fully qualified path to the directory
     */
    public void setCurrentDir(String dir) {
        if (dir == null || currentDir != null ||
                dir.isEmpty() || !currentDir.isEmpty()) return;
        currentDir = dir;
    }

    /**
     * Convenience method, which will invoke changeServerDirectory in an attempt
     * to change to directory relDir, which it assumes to be a directory
     * relative to the current directory.
     *
     * @param relativePath The directory to change to, specified as a path relative to the current directory.
     */
    public void navigateTo(String relativePath) {
        if (relativePath.equalsIgnoreCase("..")) {
            String hs = currentDir.substring(0, currentDir.length() - 1);
            int lp = hs.lastIndexOf('\\');
            String parentFolder = currentDir.substring(0, lp + 1);
            changeServerDirectory(parentFolder);
        } else changeServerDirectory(currentDir + relativePath + "\\");
    }

    /**
     * Sends a request asking the current server directory be changed to the argument.
     *
     * @param absolutePath The path to change to. Must be fully qualified.
     * @return If another FT operation is in progress the method will return false without
     * sending the request, otherwise true.
     */
    public boolean changeServerDirectory(String absolutePath) {
        // One directory update at a time
        if (ftActionInProgress) {
            VNCViewer.pen.log(Level.FINE, vc.logStr("changeServerDirectory bounced at [" + absolutePath + "] because another FT is in progress."));
            return false;
        }

        if (absolutePath == null || absolutePath.length() == 0) absolutePath = getCurrentDrive();
        if (absolutePath == null || absolutePath.length() == 0) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Couldn't change server directory."));
            return false;
        }

        try {
            //absolutePath="/Apollo";
//        	absolutePath="C:\\home\\dic\\aaa\\bbb"; // funktioniert
//        	absolutePath="/home/dic/aaa/bbb";
//        	absolutePath="C:\\var\\log\\"; //funktioniert auch

            VNCViewer.pen.log(Level.CONFIG, vc.logStr("Requesting [" + absolutePath + "] ..."));
//            if ( absolutePath.endsWith( "\\") && !absolutePath.endsWith( ":\\"))
//            	absolutePath = absolutePath.substring( 0, absolutePath.length() - 1);

            desiredDir = absolutePath;


            VNCViewer.pen.log(Level.CONFIG, vc.logStr("Changing to directory__ [" + desiredDir + "] on the server."));
            setBusy(true);
            rfs.clear();
            vc.fireEvent(RFBHandlerEvent.cdbegin(vc, currentDir, desiredDir), null);
            sendFTOperationRequest(RFB.rfbDirContentRequest, RFB.rfbRDirContent, 0, desiredDir.length(), desiredDir);
        } catch (IOException e) {
            setBusy(false);
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to change server directory to [" + desiredDir + "]"), e);
            return false;
        } finally {
            setBusy(false);
        }

        return true;
    }

    /**
     * Will send a "list drives" request to the server.
     *
     * @return If another FT operation is in progress the method will return false without sending the request.
     */
    public boolean readServerDriveList() {
        if (ftActionInProgress) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Could not read list of remote drives. Another FT operation is in Progress."));
            return false;
        }

        try {
            setBusy(true);
            oldRDL = remoteDrivesList;
            remoteDrivesList = new ArrayList<FTPath>();

            sendFTOperationRequest(RFB.rfbDirContentRequest, RFB.rfbRDrivesList, 0, 0, null);
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Requested a list of drives available to the server."));
        } catch (IOException e) {
            setBusy(false);
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to read the list of remote drives."), e);
            return false;
        } finally {
            setBusy(false);
        }
        return true;
    }

    public void tryToAbortFileTransfer() {
        fAbort = true;
    }

    /**
     * Writes an RFB message to the server, indicating a file transfer request.
     * size = 0 : compression not supported; 1 : compression supported.
     *
     * @param contentType  A protocol specific constant indicating the desired operation (I.e. RFB.rfbDirContentRequest)
     * @param contentParam Another protocol specific constant serving as an argument to the previous one (RFB.rfbRDrivesList)
     * @param size         Some FT operations require knowing the length of a file before up-/downloading.
     * @param length       Same goes for file-/directory names.
     * @param text         A textual representation of the target of the operation. Often this is a directory or filename.
     *                     Set null if unused.
     */
    private void sendFTOperationRequest(
            int contentType, int contentParam,
            long size, long length, String text)
            throws IOException {
        byte b[] = new byte[12];

        b[0] = (byte) RFB.rfbFileTransfer;
        b[1] = (byte) contentType;
        b[2] = (byte) contentParam;
        b[3] = 0; // decided to leave nothing 2 chance here ... (Wild guess)

        // WORKINPROGRESS!!!
        Charset cs = vc.getPresumedEncoding();
//        Charset cs = RFB.DEFAULT_CHARSET;
        byte[] textAsBytes = null;
        int textLen = -1;

        if (text != null) {
            textAsBytes = text.getBytes(cs);
            textLen = textAsBytes.length;
        }

//        long lengthIn = length; 
        length = textLen == -1 ? length : textLen;
        length++; // This one looks bogus to me

//      boolean isCDReq = contentType == RFB.rfbDirContentRequest && 
//      contentParam == RFB.rfbRDirContent;
//        if ( isCDReq) {
//            String hs = "CD size       [" + size     + "]\n" +
//                        "CD lengthIn   [" + lengthIn + "]\n" +
//                        "CD length     [" + length   + "]\n" +
//                        "CD text       [" + text     + "]\n" +
//                        "CD encoding   [" + cs       + "]";
//        
//            VNCViewer.pen.log( Level.FINEST, vc.logStr( hs));
//        }

        b[4] = (byte) ((size & 0xFF000000) >>> 24);
        b[5] = (byte) ((size & 0xFF0000) >>> 16);
        b[6] = (byte) ((size & 0xFF00) >>> 8);
        b[7] = (byte) ((size & 0xFF));

        b[8] = (byte) ((length & 0xFF000000) >>> 24);
        b[9] = (byte) ((length & 0xFF0000) >>> 16);
        b[10] = (byte) ((length & 0xFF00) >>> 8);
        b[11] = (byte) ((length & 0xFF));

        os.write(b);

        if (text == null || length > Integer.MAX_VALUE || length <= 0) return;

        int l = (int) length;

        byte text0Terminated[] = new byte[l];
        System.arraycopy(textAsBytes, 0, text0Terminated, 0, l - 1);
        text0Terminated[l - 1] = 0;

//        byte text0Terminated[] = new byte[ textLen + 1];
//        System.arraycopy( textAsBytes, 0, text0Terminated, 0, textLen);
//        text0Terminated[ text0Terminated.length - 1] = 0;

//        for ( int i = 0; i < textLen; i++)
//            byteArray2[ i] = textAsBytes[ i];
//        byteArray2[ byteArray2.length - 1] = 0;

        os.write(text0Terminated);
    }

    /**
     * This is a callback method getting invoked whenever the server has signaled it is
     * ready to accept a file awaiting its upload. It should <b>never</b> be called directly!
     *
     * @param contentType  RFB.rfbFilePacket
     * @param contentParam Unused → 0
     * @param size         Unused → 0
     * @param length       Unused → 0
     * @param text         Name of File/Directory to be uploaded.
     * @return Returns -1 in the event of an error and 1 if successful.
     */
    private int sendFileToServer(int contentType, int contentParam,
                                 long size, long length,
                                 String filename) throws IOException {
        if (filename == null) return -1;

        File f = new File(filename);
        FileInputStream fis = new FileInputStream(f);
        byte byteBuffer[] = new byte[RFB.sz_rfbBlockSize];
        VNCViewer.pen.log(Level.FINEST, vc.logStr("Reading source file"));
        int bytesRead = fis.read(byteBuffer);
        long counter = 0;
        boolean fError = false;

        // sf@ - Manage compression
        boolean fCompress = true;
        VNCViewer.pen.log(Level.FINEST, vc.logStr("Setting up deflater to compress source file."));
        Deflater myDeflater = new Deflater();
        byte[] CompressionBuffer = new byte[RFB.sz_rfbBlockSize + 1024];
        int compressedSize = 0;

        RFBHandlerEvent e1 = RFBHandlerEvent.ftprogress(vc, filename, false, 0, 0);
        e1.success = true;

        while (bytesRead != -1) {
            counter += bytesRead;
            myDeflater.setInput(byteBuffer, 0, bytesRead);
            myDeflater.finish();
            compressedSize = myDeflater.deflate(CompressionBuffer);
            myDeflater.reset();

            // If the compressed data is larger than the original one, we are
            // probably dealing with data which has already been compressed.
            // This one looks REALLY bogus to me since the compression is NOT
            // guaranteed to ALWAYS reduce the chunks size.
            if (compressedSize > bytesRead)
                fCompress = false;

            sendFTOperationRequest(contentType, contentParam, (fCompress ? 1 : 0),
                    (fCompress ? compressedSize - 1 : bytesRead - 1), null);

            os.write(fCompress ? CompressionBuffer : byteBuffer, 0, fCompress ? compressedSize : bytesRead);

            bytesRead = fis.read(byteBuffer);

            if (countObservers() > 0) {
                Date d = new Date();
                long secs = (d.getTime() - ftStarted.getTime()) / 1000;
                int pcnt = f.length() != 0 ? (int) ((counter * 100) / f.length()) : 0;

                e1.progress = pcnt;
                e1.bps = secs != 0 ? counter / secs : 0;
                setChanged();
                notifyObservers(e1);
            }

            if (fAbort) {
                fAbort = false;
                fError = true;
                VNCViewer.pen.log(Level.WARNING, vc.logStr("The filetransfer was aborted."));
                break;
            }
        }

        // Delete the file if the respective flag is set and it is safe
        // to assume is has been received without errors.
        if (rmFilePostUpload && (!fAbort || fError)) {
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Deleting source file."));
            zippedDir.deleteOnExit();
        }

        // I'd rather risk running out of disk space than to
        // accidentally delete files I don't intend to.
        rmFilePostUpload = false;

        sendFTOperationRequest(fError ? RFB.rfbAbortFileTransfer : RFB.rfbEndOfFile, 0, 0, 0, null);
        fis.close();
        return (fError ? -1 : 1);
    }

    /**
     * This method is invoked internally when the server signals it is ready to
     * receive a file to be uploaded. To upload a file the client is sending a
     * message to the server indicating it wants to send the server a file. The
     * server then has to signal it will accept the file.
     */
    private void onReceivingAnUploadFileRequest() {
        try {
            int size = is.readInt();
            int length = is.readInt();

//            String dbgFilename = "";
//            for ( int i = 0; i < length; i++)
//                dbgFilename += ((char) is.readUnsignedByte());

            byte[] bErrMsg = new byte[length];
            is.readFully(bErrMsg);
            String dbgFilename = new String(bErrMsg, vc.getPresumedEncoding());

            if (size <= -1) {
                String errMsg = "The server has refused to accept file [" + dbgFilename + "].\n" +
                        "Please check available space on device and access privileges.";

                VNCViewer.pen.log(Level.WARNING, vc.logStr(errMsg));

                fnUpload = null;
                if (unzip2Dir == null) zippedDir = null;
                vc.fireEvent(RFBHandlerEvent.ftError(vc, dbgFilename, currentDir, errMsg), errMsg);
                return;
            }

            VNCViewer.pen.log(Level.FINEST, vc.logStr("The server has accepted file [" + dbgFilename + "]."));
            int ret = sendFileToServer(RFB.rfbFilePacket, 0, 0, 0, fnUpload);

            String rs = ret == 1 ? "succeeded." : "failed.";
            String tpe = zippedDir != null ? "directory" : "file";
            String hs = "Upload of " + tpe + " [" + fnUpload + "] to [" + currentDir + "] " + rs;
            VNCViewer.pen.log(ret != 1 ? Level.INFO : Level.CONFIG, vc.logStr(hs));

            // This is NOT a glitch. "Finer" messages are relayed to a Web-Logger!
            VNCViewer.pen.log(Level.FINER, vc.logStr(hs));
            String errMsg = ret == 1 ? null : hs + "\nThe file on the server is either corrupted or incomplete.";
            vc.fireEvent(RFBHandlerEvent.ftend(vc, fnUpload, currentDir, false, zippedDir != null, false), errMsg);
        } catch (IOException e) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to satisfy file/directory upload request."), e);
        } finally {
            fnUpload = null;
            setBusy(false);
        }
    }

    public void handleFTUpdate() throws IOException {
        //VncViewer.pen.log( Level.FINEST, vc.logStr( "reading file transfer Msg"));
        int contentType = is.readUnsignedByte();
        int contentParamT = is.readUnsignedByte();
        int contentParam = contentParamT;
        contentParamT = is.readUnsignedByte();
        contentParamT = contentParamT << 8;
        contentParam = contentParam | contentParamT;

        contentParam |= contentParamT;

        if (contentType == RFB.rfbRDrivesList || contentType == RFB.rfbDirPacket) {
            //VncViewer.pen.log( Level.FINEST, vc.logStr( "Incoming Drive or Directory List detected."));
            onReceivingADriveOrDirectoryList(contentParam);
            return;
        }

        if (contentType == RFB.rfbFileHeader) {
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Incoming File detected."));
            onReceivingAFileHeader();
            return;
        }

        if (contentType == RFB.rfbFilePacket) {
            //VncViewer.pen.log( Level.FINEST, vc.logStr( "It's a File Packet"));
            onReceivingAFileChunk();
            return;
        }

        if (contentType == RFB.rfbEndOfFile || contentType == RFB.rfbAbortFileTransfer) {
            //VncViewer.pen.log( Level.FINEST, vc.logStr( "It's the end of a File as we know it :>"));
            onReceivingEndOfFile(contentType == RFB.rfbEndOfFile); // Ok
            return;
        }

        if (contentType == RFB.rfbCommandReturn) {
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Incoming command returnvalue detected."));
            if (contentParam == RFB.rfbADirCreate) onCreateRemoteDirectoryFeedback();
            else onDeleteRemoteFileFeedback();
            return;
        }

        if (contentType == RFB.rfbFileAcceptHeader) {
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Incoming file request detected."));
            onReceivingAnUploadFileRequest();
            return;
        }

        if (contentType == RFB.rfbFileChecksums) {
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Incoming file checksum detected."));
            onReceivingAFileChecksum();
            return;
        }

        VNCViewer.pen.log(Level.WARNING, vc.logStr("Unknown RFB file transfer message: [" + contentType + "]"));
    }

    /**
     * Handles the servers response to a "create remote directory" request from
     * the server.
     */
    private void onCreateRemoteDirectoryFeedback() throws IOException {
        int rval = is.readInt();
        int length = is.readInt();

        byte[] bErrMsg = new byte[length];
        is.readFully(bErrMsg);
        String f = new String(bErrMsg, vc.getPresumedEncoding());

        setBusy(false);

        String msg = rval == 0 ? "Successfully created directory [" + dnCreating + "]" :
                "Failed to create directory [" + dnCreating + "]";

        if (f != null && !f.isEmpty()) msg += " : " + f;

        // VncViewer.pen.log( Level.INFO, vc.logStr( "onCreateRemoteDirectoryFeedback rval: [" + rval + "]"));
        // VncViewer.pen.log( Level.INFO, vc.logStr( "onCreateRemoteDirectoryFeedback len : [" + length + "]"));
        // VncViewer.pen.log( Level.INFO, vc.logStr( "onCreateRemoteDirectoryFeedback f   : [" + f + "]"));

        VNCViewer.pen.log(Level.FINER, vc.logStr(msg));
        vc.fireEvent(RFBHandlerEvent.dircreate(vc, dnCreating, currentDir, msg), rval == 0 ? null : msg);
        dnCreating = null;
    }

    /**
     * Handles acknowledgment that the file has been deleted on the server.
     */
    private void onDeleteRemoteFileFeedback() throws IOException {
        int rval = is.readInt();
        int length = is.readInt();

        byte[] bErrMsg = new byte[length];
        is.readFully(bErrMsg);
        String f = new String(bErrMsg, vc.getPresumedEncoding());

        String msg = rval != -1 ? "File [" + fnDeleting + "] has been deleted" :
                "Failed to delete [" + fnDeleting + "]";

        if (f != null && !f.isEmpty()) msg += " : " + f;

        // This is not a mistake. Finer messages are relayed to the Web-Logger!
        VNCViewer.pen.log(Level.FINER, vc.logStr(msg));

        fnDeleting = null;
        setBusy(false);
        vc.fireEvent(RFBHandlerEvent.filedelete(vc, f.substring(0, length - 1), msg), rval != -1 ? null : msg);
    }

    private void onReceivingAFileChecksum() throws IOException {
        is.readInt(); // padding? (used to be labeled "size")
        int length = is.readInt();

        byte[] ReceptionBuffer = new byte[length + 32];

        // Read the incoming file data
        is.readFully(ReceptionBuffer, 0, length);
    }

    /**
     * Internally used when transferring file from server. Here, the server sends
     * a RFB packet signaling that it is ready to send the file requested
     */
    private void onReceivingAFileHeader() throws IOException {
        int size = is.readInt();
        int length = is.readInt();

        byte[] bErrMsg = new byte[length];
        is.readFully(bErrMsg);
        String tempName = new String(bErrMsg, vc.getPresumedEncoding());

        // sf@2004 - Read the high part of file size (not yet in rfbFileTransferMsg for
        // reasons of backwards compatibility ...)
        int sizeH = is.readInt();
        long lSize = ((long) (sizeH) << 32) + size;

        if (size <= -1) {
            String errMsg = "The server has refused access to file [" + tempName + "]. Please check access privileges.";
            VNCViewer.pen.log(Level.WARNING, vc.logStr(errMsg));
            vc.fireEvent(RFBHandlerEvent.ftend(vc, tempName, currentDir, true, zippedDir != null, false), errMsg);
            return;
        }

        downloadBytesTotal = lSize;
        downloadBytesRead = 0;
        fosFileDownload = new FileOutputStream(fnDownload);
        sendFTOperationRequest(RFB.rfbFileHeader, 0, 0, 0, null);
    }

    /**
     * Internal function for reading raw file data from the stream and writing
     * it to a previously configured file. There should be no need whatsoever to invoke
     * this method outside this context.
     */
    private void onReceivingAFileChunk() throws IOException {
        // sf@2004 - Size = 0 means file chunk not compressed
        int size = is.readInt();
        boolean fCompressed = (size != 0);
        int length = is.readInt();

        // sf@2004 - allocates buffers for file chunk reception and decompression
        byte[] ReceptionBuffer = new byte[length + 32];

        // Read the incoming file data
        is.readFully(ReceptionBuffer, 0, length);

        if (fCompressed) {
            int bufSize = RFB.sz_rfbBlockSize + 1024;
            int decompressedSize = 0;
            byte[] DecompressionBuffer = new byte[bufSize];
            Inflater myInflater = new Inflater();
            myInflater.setInput(ReceptionBuffer);
            try {
                decompressedSize = myInflater.inflate(DecompressionBuffer);
            } catch (DataFormatException e) {
                VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to inflate file chunk read from server. The downloaded file is probably damaged."), e);
            }
            fosFileDownload.write(DecompressionBuffer, 0, decompressedSize);
            downloadBytesRead += decompressedSize;
        } else {
            fosFileDownload.write(ReceptionBuffer, 0, length);
            downloadBytesRead += length;
        }

        if (fAbort == true) {
            fAbort = false;
            sendFTOperationRequest(RFB.rfbAbortFileTransfer, 0, 0, 0, null);
            VNCViewer.pen.log(Level.FINER, vc.logStr("User requested the download be aborted."));
        }

        if (countObservers() > 0) {
            Date d = new Date();
            long secs = (d.getTime() - ftStarted.getTime()) / 1000;
            int pcnt = downloadBytesTotal != 0 ? (int) ((downloadBytesRead * 100) / downloadBytesTotal) : 0;
            vc.fireEvent(RFBHandlerEvent.ftprogress(vc, fnDownload, true, pcnt, secs != 0 ? downloadBytesRead / secs : 0), fAbort != true ? null : "Download aborted");
        }
    }

    /**
     * Internally used when transferring file from server. Server signals end of file.
     */
    private void onReceivingEndOfFile(boolean fReceptionOk) throws IOException {
        String fnZip = zippedDir != null ? zippedDir.getAbsolutePath() : null;
        String fnDstDir = unzip2Dir != null ? unzip2Dir.getAbsolutePath() : null;
        boolean fileIsAZippedDir = zippedDir != null;

        boolean success = true;
        is.readInt();   // padding? (used to be labeled "size"
        is.readInt();   // padding? (used to be labeled "length"
        fosFileDownload.close();
        fosFileDownload = null;
        downloadBytesRead = 0;

        VNCViewer.pen.log(Level.FINEST, vc.logStr("End of file " + desiredDir + " received!"));

        // If the file we just received is a zipped directory ...
        if (fileIsAZippedDir) {
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Zipped directory transfer detected. Zipfile: [" + fnZip + "]"));

            // we shall extract it's contents into the present directory
            try {
                vc.fireEvent(RFBHandlerEvent.zipbegin(vc, fnZip, fnDstDir, true, false), null);

                VNCViewer.pen.log(Level.FINEST, vc.logStr("Unzipping [" + fnZip + "] to directory [" + fnDstDir + "]"));
                unzip2Dir.mkdirs();

                VNCViewer.pen.log(Level.FINEST, vc.logStr("Zipped dir  " + zippedDir.getAbsolutePath()));


                ZipFile zf = new ZipFile(zippedDir);
                Enumeration<? extends ZipEntry> zfEntries = zf.entries();

                int entriesInZip = zf.size();
                int entriesProcessed = 1;

                RFBHandlerEvent e2 = RFBHandlerEvent.zipprogress(vc, fnZip, fnDstDir, true, true, false);

                while (zfEntries.hasMoreElements()) {
                    ZipEntry e = zfEntries.nextElement();
                    String dn = unzip2Dir.getAbsolutePath() + File.separatorChar + e.getName();

                    if (countObservers() > 0) {
                        int pcnt = entriesInZip != 0 ? (int) ((entriesProcessed++ * 100) / entriesInZip) : 0;
                        e2.progress = pcnt;
                        setChanged();
                        notifyObservers(e2);
                    }

                    if (e.isDirectory()) {
                        File f = new File(dn);
                        f.mkdirs();
                        VNCViewer.pen.log(Level.FINEST, vc.logStr("Creating directory: [" + dn + "]."));
                    } else {
                        VNCViewer.pen.log(Level.FINEST, vc.logStr("Unzipping file [" + dn + "] from: [" + fnZip + "]."));
                        InputStream zipIs = zf.getInputStream(e);
                        FileOutputStream fileOs = new FileOutputStream(dn);

                        byte[] b = new byte[1024];
                        int len = 0;

                        while ((len = zipIs.read(b)) >= 0)
                            fileOs.write(b, 0, len);

                        zipIs.close();
                        fileOs.close();
                    }
                }

                zf.close();
                zippedDir.deleteOnExit();
                VNCViewer.pen.log(Level.FINEST, vc.logStr("Temporary zipfile: [" + fnZip + "] has been deleted."));
            } catch (IOException e) {
                success = false;
                VNCViewer.pen.log(Level.FINEST, vc.logStr("Failed to inflate compressed directory. " + e.getMessage()), e);
            }

            String hs = success ? "have been inflated" : "could not be inflated";
            String hsAll = "The contents of [" + fnZip + "] " + hs + " to directory [" + fnDstDir + "]";

            if (success) VNCViewer.pen.log(Level.FINEST, vc.logStr(hsAll));
            else VNCViewer.pen.log(Level.WARNING, vc.logStr(hsAll));

            vc.fireEvent(RFBHandlerEvent.zipend(vc, fnZip, fnDstDir, hsAll, true, true, false), success ? null : hsAll);

            zippedDir = null;
            unzip2Dir = null;
        }

        setBusy(false);
        String rs = fReceptionOk ? "succeeded." : "failed.";
        String tpe = fileIsAZippedDir ? "directory" : "file";
        String srcLoc = fileIsAZippedDir ? dn2bDownloaded : fnDownload;
        String dstLoc = fileIsAZippedDir ? fnDstDir : fnDownload;
        String msg = "Download of " + tpe + " [" + srcLoc + "] to [" + dstLoc + "] " + rs;

        if (success) VNCViewer.pen.log(Level.CONFIG, vc.logStr(msg));
        else VNCViewer.pen.log(Level.WARNING, vc.logStr(msg));

        // This is not a goof. Finer messages are relayed to the Web-Logger!
        VNCViewer.pen.log(Level.FINER, vc.logStr(msg));

        dn2bDownloaded = null;
        fnDownload = null;

        vc.fireEvent(RFBHandlerEvent.ftend(vc, fnZip, fnDstDir, true, fileIsAZippedDir, false), fReceptionOk ? null : msg);
    }

    /**
     * Refactored from readRfbFileTransferMsg()
     */
    private void onReceivingADriveOrDirectoryList(int contentParam) throws IOException {
        if (contentParam == RFB.rfbADrivesList) {
            readFTPMsgDriveList();
        } else if (contentParam == RFB.rfbADirectory && !inDirectory2) {
            inDirectory2 = true;
            readFTPMsgDirectoryList();
        } else if (contentParam == RFB.rfbADirectory && inDirectory2) {
            readFTPMsgDirectoryListContent();
        } else if (contentParam == 0) {
            readFTPMsgDirectoryListEndContent();
            inDirectory2 = false;
        } else
            VNCViewer.pen.log(Level.FINEST, vc.logStr("Received an unknown ContentParam: [" + contentParam + "]"));
    }

    /**
     * Compares the oldRDL to the remoteDrivesList.
     *
     * @return Returns true if they differ or if there is no oldRDL and false if they are logically the same.
     * Two lists will be considered logically the same, if they contain objects with the same characteristics
     * even if in a different order.
     */
    private boolean rdlChanged() {
        boolean found = false;

        if (oldRDL == null || remoteDrivesList == null) return true;

        if (remoteDrivesList.size() != oldRDL.size()) return true;

        Iterator<FTPath> i = oldRDL.iterator();
        FTPath drvOld = null;
        FTPath drvNew = null;

        // Iterate over the list of drives found last time
        while (i.hasNext() && !found) {
            drvOld = (FTPath) i.next();

            // Search the new list of drives for one with the same properties
            drvNew = getDrive(drvOld);

            // If no such drive exists, there must have been a change.
            found = drvNew == null;
        }

        return found;
    }

    /**
     * Internally used to receive list of drives available on the server
     */
    private void readFTPMsgDriveList() throws IOException {
        // skip over the padding
        for (int i = 0; i < 4; i++) is.readUnsignedByte();

        String str = "";
        int length = is.readInt();

        for (int i = 0; i < length; i++) {
            char temp = (char) is.readUnsignedByte();
            if (temp != '\0')
                str += temp;
        }

        VNCViewer.pen.log(Level.FINEST, vc.logStr("List of drives: [" + str + "]"));

        for (int i = 0; i < str.length(); i += 3) {
            FTPath d = new FTPath(str.substring(i, i + 3)/*, false*/);
            remoteDrivesList.add(d);
        }

        setBusy(false);

        if (rdlChanged()) vc.fireEvent(RFBHandlerEvent.drvlistchanged(vc), null);

        // Get rid of the old list
        oldRDL = null;
    }

    /**
     * Internally used to receive directory content from server.
     * Here, the server marks the start of the directory listing
     */
    private void readFTPMsgDirectoryList() throws IOException {
        is.readInt();
        int length = is.readInt();
        if (length == 0) {
            VNCViewer.pen.log(Level.WARNING, vc.logStr("Failed to change to [" + desiredDir + "]"));
            desiredDir = null;
            setBusy(false);
            setChanged();
            inDirectory2 = false;

            // Can't set report success false here. The application may fail 2 start! :(
            vc.fireEvent(RFBHandlerEvent.cdend(vc, currentDir, desiredDir), null);
        } else {
            // The new FT protocol will send the remote directory name
            // My observations:
            // Windumb UVNC Server sends number of bytes 2 read
            // 3 C:\<something definitely NOT a \0>
            // Linux VNC Server sends null terminated String
            // 4 C:\⇞ ⇞=\0

            byte b[] = new byte[length];
            is.readFully(b);
            currentDir = scrapNullTerminator(b);
            VNCViewer.pen.log(Level.FINEST, vc.logStr("The current remote directory is now [" + currentDir + "]"));

            // WORKINPROGRESS!!!
//            byte b[] = new byte[ length - 1];
//            is.readFully( b);
//            is.readByte(); // cheap trick to skip the null terminator
//
//            currentDir = new String( b, vc.getPresumedEncoding());

//            byte[] bErrMsg = new byte[ length];
//            is.read( bErrMsg);
//            dump( bErrMsg, "remote directory name");
//            String str = new String( bErrMsg, vc.getPresumedEncoding());

//            for ( int i = 0; i < length; i++) {
//                char temp = (char) is.readUnsignedByte();
//                if ( temp != '\0')
//                    str += temp;
//            }

//            String str = readNullTerminatedString();
//            VNCViewer.pen.log( Level.FINEST, vc.logStr( "New FT protocol sent the remote directory name [" + str + "]"));
//            currentDir = str;
        }
    }

    /**
     * Internally used to receive directory content from server.
     * Here, the server sends one file/directory with it's attributes
     */
    private void readFTPMsgDirectoryListContent() throws IOException {
        // The code is expecting a serialized
        // typedef struct _WIN32_FIND_DATA {
        //   DWORD dwFileAttributes;
        //   FILETIME ftCreationTime;
        //   FILETIME ftLastAccessTime;
        //   FILETIME ftLastWriteTime;
        //   DWORD nFileSizeHigh;
        //   DWORD nFileSizeLow;
        //   DWORD dwReserved0;
        //   DWORD dwReserved1;
        //   TCHAR cFileName[ MAX_PATH ];
        //   TCHAR cAlternateFileName[ 14 ];
        // } WIN32_FIND_DATA;

        String fName = "";
        String alternateFileName = "";
        int dwFileAttrib = 0;
        int nFileSizeHigh = 0;
        int nFileSizeLow = 0;
        //char cFileName, cAlternateFileName;
        is.readInt(); // Padding
        int length = is.readInt();

        dwFileAttrib = is.readInt();
        length -= 4;

        Date dCreat = readDate();
        Date dAccss = readDate();
        Date dWrite = readDate();
        length -= 24;

        nFileSizeHigh = is.readInt();
        length -= 4;
        nFileSizeLow = is.readInt();
        length -= 4;

        is.readInt();
        length -= 4; // unused bytes previously marked dwReserved0
        is.readInt();
        length -= 4; // unused bytes previously marked dwReserved1

        fName = readNullTerminatedString();
//        VNCViewer.pen.log( Level.FINEST, vc.logStr( "Read fName [" + fName + "]" + 
//                                                    "length [" + length + "]" +
//        		                                    "npBytes [" + npBytesRead + "]"));

        if (length > npBytesRead) {
            byte[] bArr = new byte[length - npBytesRead];
            is.readFully(bArr);
            alternateFileName = new String(bArr, vc.getPresumedEncoding());
        }

// Het werkt ....
//        char cFileName = 0;
//        char cAlternateFileName = 0;
//
//        cFileName = (char) is.readUnsignedByte();
//        length--;
//       
//        while (cFileName != '\0') {
//            fName += cFileName;
//            cFileName = (char) is.readUnsignedByte();
//            length--;
//        }
//
//        cAlternateFileName = (char) is.readByte();
//        length--;
//        while (length != 0) {
//            alternateFileName += cAlternateFileName;
//            cAlternateFileName = (char) is.readUnsignedByte();
//            length--;
//        }

        //VncViewer.pen.log( Level.WARNING, vc.logStr( "fName=[" + fName + "] cAlternateFileName=[" + alternateFileName + "]"));

        // OK ... This code assumes the above File attributes where sent by a
        // windumb box and represent a memcopied "serialization" of a
        // WIN32_FIND_DATA structure. ( If it can get any more outrageous please
        // be sure to tell me). This code is highly OS dependent and will likely
        // fail if the server should use a different representation.

        long fSizeH = Integer.reverseBytes(nFileSizeHigh) & 0xffffffffL;
        long fSizeL = Integer.reverseBytes(nFileSizeLow) & 0xffffffffL;
        long fSize = (long) (fSizeH * (MAXWORD + 1) + fSizeL);

//         VNCViewer.pen.log( Level.FINEST, "FileName                    : " + fName);
//         VNCViewer.pen.log( Level.FINEST, "FileName (alt)              : " + alternateFileName);
//         VNCViewer.pen.log( Level.FINEST, "dwFileAttrib                : " + dwFileAttrib);
//         VNCViewer.pen.log( Level.FINEST, "nFileSizeHigh/ReversedBytes : " + nFileSizeHigh + " / " + fSizeH);
//         VNCViewer.pen.log( Level.FINEST, "nFileSizeLow/ReversedBytes  : " + nFileSizeLow + " / " + fSizeL);
//         VNCViewer.pen.log( Level.FINEST, "fSize                       : " + fSize);
//         VNCViewer.pen.log( Level.FINEST, "RfSize                      : " + Integer.toBinaryString( nFileSizeLow));
//         VNCViewer.pen.log( Level.FINEST, "fSize/1024                  : " + fSize / 1024);
//         VNCViewer.pen.log( Level.FINEST, "--------------------------------------------------------------");
        int mdw = Integer.reverseBytes(dwFileAttrib);
        FTFile f = new FTFile(fName, alternateFileName, mdw, fSize, dCreat, dAccss, dWrite);
        rfs.add(f);

        vc.fireEvent(RFBHandlerEvent.rfsgrew(vc, fName, f.isDirectory()), null);
    }

    /**
     * Used internally to read directory content of server.
     * Here, the server signals the end of the directories list of contents.
     */
    private void readFTPMsgDirectoryListEndContent() throws IOException {
        is.readInt();
        int b = is.readInt(); // padding? (formerly labeled "length") I guess it's an error flag! 

        //currentDir = future_Dir;
        desiredDir = null;
        setBusy(false);

        vc.fireEvent(RFBHandlerEvent.cdend(vc, currentDir, desiredDir), null);
        VNCViewer.pen.log(Level.FINEST, vc.logStr("CD [" + (b != 0 ? "failed" : "succeeded") + "] CurrentDir [" + currentDir + "]"));
    }

    // sf@2004 - Read the destination file checksums data. (Not used for now)
    @SuppressWarnings("unused")
    private void ReceiveDestinationFileChecksums() throws IOException {
        is.readInt(); // padding? (formerly labeled "size")
        int length = is.readInt();

        byte[] ReceptionBuffer = new byte[length + 32];

        // Read the incoming file data
        is.readFully(ReceptionBuffer, 0, length);
    }

    /**
     * Zips the contents of \a dir2zip to the ZipOutputStream \a zos, cropping
     * the \a cp first letters of all filenames written to said stream. The parameters
     * \a fcount and \a zipped are used to compute a progress percentage. \a fcount
     * should hold the total number of files to be zipped. \A zipped th number of files
     * processed so far.
     */
    @SuppressWarnings("unused")
    private int zip(File dir2zip, ZipOutputStream zos, int cp, int fcount, int zipped) throws IOException {
        //interrupt();

        if (dir2zip.isDirectory()) {
            File[] dirList = dir2zip.listFiles();

            if (dirList == null) return zipped;

            for (int i = 0; i < dirList.length; i++)
                zipped = zip(dirList[i], zos, cp, fcount, zipped);
        } else {
            String hs = dir2zip.getAbsolutePath().substring(cp);
            zos.putNextEntry(new ZipEntry(hs));

            FileInputStream is = new FileInputStream(dir2zip);

            byte[] b = new byte[1024];
            int len = 0;

            while ((len = is.read(b)) >= 0)
                zos.write(b, 0, len);

            zos.closeEntry();
            is.close();

            RFBHandlerEvent e1 = RFBHandlerEvent.zipprogress(vc, dir2zip.getAbsolutePath(),
                    ""/*targetPath*/, false, dir2zip.isDirectory(), true);

            if (countObservers() > 0) {
                int pcnt = fcount != 0 ? (int) ((zipped++ * 100) / fcount) : 0;
                e1.progress = pcnt;
                setChanged();
                notifyObservers(e1);
            }
        }
        return zipped;
    }

    @SuppressWarnings("unused")
    private void dump(byte[] b, String s) {
        String ms = "";
        for (int i = 0; i < b.length; i++)
            ms += "[" + b[i] + "]";

        VNCViewer.pen.log(Level.WARNING, vc.logStr(s + "(" + b.length + "): " + ms));
    }

    private String scrapNullTerminator(byte[] b) {
        int i = b != null ? b.length - 1 : 0;
        if (i <= 0) return "";

        while (i >= 0 && b[i] == 0) i--;

        byte[] x = new byte[i + 1];
        System.arraycopy(b, 0, x, 0, i + 1);
        return new String(x, vc.getPresumedEncoding());
    }

    private String readNullTerminatedString() throws IOException {
        int read = 0;
        byte b = 0;
        byte[] buff = new byte[30];

        do {
            b = is.readByte();
            if (read + 1 > buff.length)
                buff = Arrays.copyOf(buff, buff.length * 2);
            buff[read++] = b;
        } while ((char) b != '\0');

        byte[] trunc = new byte[read - 1];
        System.arraycopy(buff, 0, trunc, 0, read - 1);
        npBytesRead = read;
        // WORKINPROGRESS!!!
        return new String(trunc, vc.getPresumedEncoding());
    }

    private void setBusy(boolean b) {
        if (ftActionInProgress == b) {
            return;
        }
        synchronized (ftActionInProgressLock) {
            if (ftActionInProgress != b) {
                ftActionInProgress = b;
            }
        }
    }

    /**
     * Will read two DWords from the stream and interpret them as a Windumb Filetime struct.
     *
     * @return Returns the corresponding Date object. It is known to be off by as much as an
     * hour because of the way Windumb handles DLST.
     */
    private Date readDate() throws IOException {
        int iLow = Integer.reverseBytes(is.readInt());
        int iHigh = Integer.reverseBytes(is.readInt());

        final long EPOCH_DIFF = 11644473600000L;

        final long filetime = ((long) iHigh) << 32 | (iLow & 0xffffffffL);
        final long microdumbTS = filetime / (1000 * 10);
        final long sinceEpoch = microdumbTS - EPOCH_DIFF;
        return new Date(sinceEpoch);
    }

    public void update(Observable o, Object arg) {
        // Here we forward VNCConnectionEvents coming from
        // DirectoryTransferAssistants.
        if (o instanceof DirectoryTransferAssistant) {
            setChanged();
            notifyObservers(arg);
        }
    }
}
