package com.wincornixdorf.vnc.rfb.ft;

import java.util.ListIterator;
import java.util.Vector;

public class RemoteFileSystem extends Vector<FTFile>
{
    private static final long serialVersionUID = -2612581489774386310L;

    private String dbgName = "unknown";

    public RemoteFileSystem( String name) {
        dbgName = name;

        // Insert a "placeholder" FTFile object to indicate it has yet to be
        // populated. The dummy File objects representation will "prompt" the
        // user to refresh the FS view if desired.
        add( FTFile.PLACEHOLDER);
    }

    /**Test if this Containers contents have changed since it was created.
    Disclaimer for the paranoid: The Method will not be able to detect if
    the container has been "restored" to its original state as to make it
    look as if it had never been modified.
    @return Returns true if this container is still untouched.*/
    public boolean isInMintCondition() {
        return size() == 1 && contains( FTFile.PLACEHOLDER);
    }

    public boolean add( FTFile f) {
        boolean b = super.add( f);
        // setChanged( new TableModelEvent( this, size() - 1, size() - 1, TableModelEvent.ALL_COLUMNS, TableModelEvent.INSERT));
        return b;
    }

    public void insertElementAt( FTFile obj, int index) {
        super.insertElementAt( obj, index);
        // setChanged( new TableModelEvent( this, index, index, TableModelEvent.ALL_COLUMNS, TableModelEvent.INSERT));
        // setChanged( new TableModelEvent( this, index + 1, size() - 1, TableModelEvent.ALL_COLUMNS, TableModelEvent.UPDATE));
    }

    /**Originally changes to the vector would be broadcasted to potential
    listeners as they occurred. This however spelled problems with several GUI
    components which would then cause null pointer exceptions. Though not fatal,
    they where a pain in the neck. Call this method to force displays of this
    vector to be updated completely.*/
//    public void forceUpdate() {
//        setChanged( new TableModelEvent( this));
//    }

    /**@return True, if this container holds a file or directory like the one
     * described by the argument. Please take note this method does <b>not</b>
     * search by object identity, but by the {@link FTFile}'s name and
     * {@link FTFile.dwFileAttrib}.
     * @param f The {@link FTFile} to look for.*/
    private boolean knows( FTFile f) {
        String fn = f.getFilename();
        return f.isDirectory() ? knowsDirectory( fn) : knowsFile( fn);
    }

    /**@param name The name of the file to look for of course.
     * @return True, if the current Directory contains the specified directory.*/
    public boolean knowsDirectory( String name)
    {
        boolean found = false;
        ListIterator<FTFile> it = listIterator();
        while ( !found && it.hasNext()) {
            FTFile f = it.next();
            found = f.isDirectory() && name.equalsIgnoreCase( f.getFilename());
        }
        return found;
    }

    /**@param name The name of the file to look for of course. 
     * @return True, if the current Directory contains the respective file.*/
    public boolean knowsFile( String name)
    {
        boolean found = false;
        ListIterator<FTFile> it = listIterator();
        while ( !found && it.hasNext()) {
            FTFile f = it.next();
            found = f.isFile() && name.equalsIgnoreCase( f.getFilename());
        }
        return found;
    }

    public void inSort( FTFile f) {
        if ( isEmpty()) { add( f); return; }

        if ( knows( f)) return;

        try {
            int i = 0;
            int h = 0;
            boolean found = false;
            FTFile g = null;

            while ( i < size() && !found) {
                g = get( i++);
                h = f.compareTo( g);
                found = ( h == Integer.MIN_VALUE) || ( h < 0 && h != Integer.MAX_VALUE);
            }

            if ( found) insertElementAt( f, i - 1);
            else add( f);

        } catch ( ArrayIndexOutOfBoundsException ea) {
            ea.printStackTrace();
        }
    }

    public int getDirectoriesCount() {
        int i = 0;
        ListIterator<FTFile> it = listIterator();
        while ( it.hasNext()) {
            i = it.next().isDirectory() ? i + 1 : i;
        }
        return i;
    }

    public int getFilesCount() {
        int i = 0;
        ListIterator<FTFile> it = listIterator();
        while ( it.hasNext()) {
            i = it.next().isFile() ? i + 1 : i;
        }
        return i;
    }

    public int getColumnCount() {
        return 5;
    }

    public String getColumnName( int columnIndex) {
        switch ( columnIndex) {
            case 0 : return( "Filename");
            case 1 : return( "Size (B)");
            case 2 : return( "Write");
            case 3 : return( "Access");
            case 4 : return( "Created");

            default: return( "Error");
        }
    }

    public int getRowCount() {
        return size();
    }

    public Class<?> getColumnClass( int columnIndex) {
        switch ( columnIndex) {
            case 0 : return FTFile.class;
            case 1 : return Long.class;
            case 2 : return String.class;
            case 3 : return String.class;
            case 4 : return String.class;
            default: return String.class;
        }
    }

    public String toString() {
        return getClass().getName() + " " +  dbgName;
    }

    public void setValueAt( Object aValue, int rowIndex, int columnIndex) {
    }
}
