package com.wincornixdorf.vnc.rfb.stream;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;

import com.wincornixdorf.vnc.global.BufferedRecorder;
import com.wincornixdorf.vnc.global.VNCViewer;

public class RecordableInputStream extends FilterInputStream {

    private boolean recording       = false;
    private BufferedRecorder bos    = null;

    public RecordableInputStream( InputStream isIn, File fRec) {
        super( isIn);
        try {
            //VncViewer.pen.log( Level.SEVERE, "rec to [" + fRec + "].");
            recordTo( fRec);
        } catch ( FileNotFoundException e) {
            VNCViewer.pen.log( Level.SEVERE, "Could not record session. File not found.", e);
        } catch ( IOException ioe) {
            VNCViewer.pen.log( Level.SEVERE, "Could not record session. Can't create File.", ioe);
        } catch ( SecurityException se) {
            VNCViewer.pen.log( Level.SEVERE, "Could not record session. No permission to create file.", se);
        }
    }

//     public int available() { return owner.blocked ? -1 : owner.buf.size(); }
    public int available() throws IOException { return in.available(); }

    public void close() throws IOException { clear(); in.close(); }

    public void mark( int readlimit) { in.mark( readlimit); }

    public boolean markSupported() { return in.markSupported(); }

    public int read() throws IOException {
        int i = in.read();

        if ( recording) {
            try { bos.write( i); } catch ( Exception e) {}
        }

        return i;
    }

    public int read( byte[] b) throws IOException {
        int h = in.read( b, 0, b.length);

        if ( recording) {
            try { bos.write( b, 0, h); } catch ( IOException e) {}
        }

        return h;
    }

    public int read( byte[] b, int off, int len) throws IOException {
        int read = in.read( b, off, len);

        if ( recording) {
            try { bos.write( b, 0, read); } catch ( IOException ioe) {}
        }

       return read;
    }

    public void reset() throws IOException { in.reset(); }

    public long skip( long n) throws IOException {
        return in.skip( n);
    }

    public void clear() {
        if ( bos == null) return;
        bos.clear();
    }

    public void flush() throws IOException {
        if ( bos == null) return;
        bos.flush();
    }

    public boolean recordTo( File f/*, boolean append*/)  throws IOException, SecurityException {
        if ( f == null || recording) {
            //VncViewer.pen.log( Level.WARNING, "Could not record session. Illegal File or already recording.");
            return false;
        }

        VNCViewer.pen.log( Level.SEVERE, "Will try recording session to [" + f + "].");
        f.createNewFile();
        VNCViewer.pen.log( Level.SEVERE, "Will record session to [" + f + "].");
        bos = new BufferedRecorder( new FileOutputStream( f));
        recording = true;
        return recording;
    }

    public boolean isRecording() {
        return recording;
    }

    public boolean stopRecording()  throws IOException {
        recording = false;

        if ( bos == null) return false;

        //bos.flush();
        bos.close();
        bos = null;
        return true;
    }

//     public boolean pauseRecording()  throws IOException {
//     }
//
//     public boolean resumeRecording()  throws IOException {
//     }

}
